Package: clustMixType
Version: 0.5-2
Date: 2026-06-25
Title: k-Prototypes Clustering for Mixed Variable-Type Data
Authors@R: c(person(given = "Gero", 
 		    family = "Szepannek", 
                    role = c("aut", "cre"), 
                    email = "gero.szepannek@web.de"),
             person(given = "Rabea", 
                    family = "Aschenbruck", 
                    role = "aut")
             )
Imports: RColorBrewer, tibble, combinat, dplyr, rlang, cluster, mclust
Suggests: testthat
Description: Functions to perform k-prototypes partitioning clustering for
    mixed variable-type data according to Z.Huang (1998): Extensions to the k-Means
    Algorithm for Clustering Large Data Sets with Categorical Variables, Data Mining
    and Knowledge Discovery 2, 283-304 and G.Szepannek et al. (2025): Clustering 
    Large Mixed-Type Data with Ordinal Variables, Adv Data Anal Classif 19, 749–767.
License: GPL (>= 2)
RoxygenNote: 7.3.3
NeedsCompilation: no
Packaged: 2026-06-28 10:19:21 UTC; szepannek
Encoding: UTF-8
Author: Gero Szepannek [aut, cre],
  Rabea Aschenbruck [aut]
Maintainer: Gero Szepannek <gero.szepannek@web.de>
Repository: CRAN
Date/Publication: 2026-06-29 06:50:13 UTC
Built: R 4.6.0; ; 2026-06-29 08:29:54 UTC; unix

# Generated by roxygen2: do not edit by hand

S3method(kproto,default)
S3method(plot,kproto)
S3method(predict,kproto)
S3method(print,kproto)
S3method(summary,kproto)
export(clprofiles)
export(importance_kproto)
export(kproto)
export(kproto2silhouette)
export(lambdaest)
export(stability_kproto)
export(validation_kproto)
importFrom(RColorBrewer,brewer.pal)
importFrom(cluster,silhouette)
importFrom(combinat,permn)
importFrom(dplyr,"%>%")
importFrom(dplyr,arrange)
importFrom(dplyr,lag)
importFrom(dplyr,lead)
importFrom(dplyr,mutate)
importFrom(graphics,barplot)
importFrom(graphics,boxplot)
importFrom(graphics,legend)
importFrom(graphics,par)
importFrom(mclust,adjustedRandIndex)
importFrom(rlang,.data)
importFrom(stats,complete.cases)
importFrom(stats,median)
importFrom(stats,na.omit)
importFrom(stats,predict)
importFrom(stats,qgeom)
importFrom(stats,sd)
importFrom(stats,var)
importFrom(tibble,is_tibble)
importFrom(utils,combn)
importFrom(utils,head)
importFrom(utils,tail)

# Development version of the R Package clustMixType

<img src="man/figures/logo.png" align="right" height="139" />

k-Prototypes Clustering for Mixed Variable-Type Data. For details cf. the [R Journal paper](https://journal.r-project.org/archive/2018/RJ-2018-048/index.html).

Corresponding papers on [valdiation indices](https://publikationen.bibliothek.kit.edu/1000120412) 
and [missing data](https://link.springer.com/article/10.1007/s00357-022-09422-y) for k prototypes.


## Installation

Install the current CRAN release of the package in R:

``` r
install.packages(clustMixType)
```

Install the latest development version of the package in R from this repository.
``` r
remotes::install_github("https://github.com/g-rho/clustMixType.git")
```

# Development version of the R Package clustMixType

<img src="man/figures/logo.png" align="right" height="139" />

k-Prototypes Clustering for Mixed Variable-Type Data. For details cf. the [R Journal paper](https://journal.r-project.org/articles/RJ-2018-048/index.html).

Corresponding papers on [valdiation indices](https://publikationen.bibliothek.kit.edu/1000120412) 
and [missing data](https://link.springer.com/article/10.1007/s00357-022-09422-y) and an [extension for ordinal variables using Gower's distance](https://link.springer.com/article/10.1007/s11634-024-00595-5) of the k-prototypes algorithm.


## Installation

Install the current CRAN release of the package in R:

``` r
install.packages(clustMixType)
```

Install the latest development version of the package in R from this repository.
``` r
remotes::install_github("https://github.com/g-rho/clustMixType.git")
```

calc.dist <- function(type, lambda, data1, data2 = NULL, all.dists = FALSE, data_plus = NULL){
  
  if(type == "huang"){# calculation with standard distance of k-prototypes
    
    # if(is.null(lambda)){
    #   lambda <- lambdaest(data, num.method = 1, fac.method = 1, outtype = "numeric", verbose = TRUE)
    # }
    
    if(!is.null(data2)){#contains the input two datasets?
      
      #tbd: check if structure of data1 and data2 are equal
      data <- rbind(data1, data2)
      
      if(nrow(data1) == 1 & nrow(data2) == 1){#calculation the distance between two observations (due to saving time)
        
        #numerical variables
        d1 <- (data[1, data_plus$numvars, drop = FALSE]-data[2, data_plus$numvars, drop = FALSE])^2
        d1[is.na(d1)] <- 0
        if(length(lambda) == 1){
          d1 <- sum(d1)
        }else{ #length(lambda) > 1
          d1 <- as.matrix(d1, nrow = 1) %*% matrix(lambda[data_plus$numvars], ncol = 1)
        }
        
        #categorical variables
        d2 <- sapply(which(data_plus$catvars), function(j) return(data[1,j] != data[2,j]))
        d2[is.na(d2)] <- FALSE
        if(length(lambda) == 1){
          d2 <- lambda * sum(d2)
        }else{#length(lambda) > 1
          d2 <- matrix(d2, nrow = 1) %*% matrix(lambda[data_plus$catvars], ncol = 1)
        }
        
        return(d1 + d2)
      }else{#calculate distances between objects of two datasets
        
        #numerical variables
        d1 <- (data1[rep(1:nrow(data1), each=nrow(data2)), data_plus$numvars, drop = FALSE] -
                 data2[rep(1:nrow(data2), times=nrow(data1)), data_plus$numvars, drop = FALSE])^2
        d1[is.na(d1)] <- 0
        if(length(lambda) == 1){
          d1 <- rowSums(d1)
        }else{ #length(lambda) > 1
          d1 <- as.matrix(d1) %*% lambda[data_plus$numvars]
        }
        
        #categorical variables
        d2 <- data1[rep(1:nrow(data1), each=nrow(data2)), data_plus$catvars, drop = FALSE] != 
          data2[rep(1:nrow(data2), times = nrow(data1)), data_plus$catvars, drop = FALSE]
        d2[is.na(d2)] <- FALSE
        if(length(lambda) == 1){
          d2 <- lambda * rowSums(d2)
        }else{ #length(lambda) > 1
          d2 <- as.matrix(d2) %*% lambda[data_plus$catvars]
        }
        
        return(sum(d1 + d2))
      }
    }else{ #calculate distances between the objects of one dataset
      
      d1 <- (data1[rep(1:(nrow(data1)-1),times=(nrow(data1)-1):1, each=1), data_plus$numvars, drop = FALSE] -
               data1[unlist(lapply(2:nrow(data1), seq, to=nrow(data1))), data_plus$numvars, drop = FALSE])^2
      d1[is.na(d1)] <- 0
      if(length(lambda) == 1) d1 <- rowSums(d1)
      if(length(lambda) > 1) d1 <- as.matrix(d1) %*% lambda[data_plus$numvars]
      
      d2 <- data1[rep(1:(nrow(data1)-1),times=(nrow(data1)-1):1, each=1), data_plus$catvars, drop = FALSE] != 
        data1[unlist(lapply(2:nrow(data1), seq, to=nrow(data1))), data_plus$catvars, drop = FALSE]
      d2[is.na(d2)] <- FALSE
      if(length(lambda) == 1) d2 <- lambda * rowSums(d2)
      if(length(lambda) > 1) d2 <- as.matrix(d2) %*% lambda[data_plus$catvars]
      
      if(all.dists == TRUE){
        m <- matrix(rep(0,nrow(data1)^2), ncol = nrow(data1))
        
        m[lower.tri(m)] <- d1+d2
        m <- t(m)
        m[lower.tri(m)] <- d1+d2
        return(m)
      }else{
        return(sum(d1 + d2))
      }
    }
  }else{ #calculation with gower distance
    
    if(!is.null(data2)){#contains the input two datasets?
      
      #tbd: check if structure of data1 and data2 are equal
      #data <- rbind(data1, data2)
      
      if(nrow(data1) == 1 & nrow(data2) == 1){#calculation the distance between two observations (due to saving time)
        
        d1 <- d2 <- d3 <- 0
        
        if(any(data_plus$numvars)){
          d1 <- abs(data1[1, data_plus$numvars, drop = FALSE] - data2[1, data_plus$numvars, drop = FALSE])
          for(jnum in 1:ncol(d1)) d1[,jnum] <- d1[,jnum] / data_plus$rgnums[jnum] 
          d1[is.na(d1)] <- 0
          if(length(lambda) > 1) d1 <- as.matrix(d1) %*% lambda[data_plus$numvars]
          if(is.null(lambda)) d1 <- rowSums(d1)
        }
        
        if(any(data_plus$catvars)){
          d2 <- data1[1, data_plus$catvars, drop = FALSE] != data2[1, data_plus$catvars, drop = FALSE]
          d2[is.na(d2)] <- FALSE
          if(length(lambda) > 1) d2 <- as.matrix(d2) %*% lambda[data_plus$catvars]
          if(is.null(lambda)) d2 <- rowSums(d2)
        }
        
        if(any(data_plus$ordvars)){
          d3 <- abs(data1[1, data_plus$ordvars, drop = FALSE] - data2[1, data_plus$ordvars, drop = FALSE])
          for(jord in 1:ncol(d3)) d3[,jord] <- d3[,jord] / data_plus$rgords[jord] 
          d3[is.na(d3)] <- 0
          if(length(lambda) > 1) d3 <- as.matrix(d3) %*% lambda[data_plus$ordvars]
          if(is.null(lambda)) d3 <- rowSums(d3)
        }
        
        return(d1 + d2 + d3)
        
      }else{#calculate distances between objects of two datasets
        
        # in case of no numeric / factor / ordinal variables set:
        d1 <- d2 <- d3 <- rep(0, nrow(data1)*nrow(data2))
        
        if(any(data_plus$numvars)){
          d1 <- abs(data1[rep(1:nrow(data1), each=nrow(data2)), data_plus$numvars, drop = FALSE] - 
                      data2[rep(1:nrow(data2), times=nrow(data1)), data_plus$numvars, drop = FALSE])
          for(jnum in 1:ncol(d1)) d1[,jnum] <- d1[,jnum] / data_plus$rgnums[jnum] 
          #time efficient alternative for codeline before:
          # d1 <- d1 / rep(data_plus$rgnums, rep(nrow(d1), ncol(d1)))
          d1[is.na(d1)] <- 0
          if(length(lambda) > 1) d1 <- as.matrix(d1) %*% lambda[data_plus$numvars]
          if(is.null(lambda)) d1 <- rowSums(d1)
        }
        
        if(any(data_plus$catvars)){
          d2 <- data1[rep(1:nrow(data1), each=nrow(data2)), data_plus$catvars, drop = FALSE] != 
            data2[rep(1:nrow(data2), times = nrow(data1)), data_plus$catvars, drop = FALSE]
          d2[is.na(d2)] <- FALSE
          if(length(lambda) > 1) d2 <- as.matrix(d2) %*% lambda[data_plus$catvars]
          if(is.null(lambda)) d2 <- rowSums(d2)
        }
        
        if(any(data_plus$ordvars)){
          d3 <- abs(data1[rep(1:nrow(data1), each=nrow(data2)), data_plus$ordvars, drop = FALSE] - 
                      data2[rep(1:nrow(data2), times=nrow(data1)), data_plus$ordvars, drop = FALSE])
          for(jord in 1:ncol(d3)) d3[,jord] <- d3[,jord] / data_plus$rgords[jord] 
          #time efficient alternative for codeline before:
          # d3 <- d3 / rep(data_plus$rgords, rep(nrow(d3), ncol(d3)))
          d3[is.na(d3)] <- 0
          if(length(lambda) > 1) d3 <- as.matrix(d3) %*% lambda[data_plus$ordvars]
          if(is.null(lambda)) d3 <- rowSums(d3)
        }
        
        return(sum(d1 + d2 + d3))
      }
    }else{ #calculate distances between the objects of one dataset
      
      d1 <- d2 <- d3 <- rep(0,length(rep(1:(nrow(data1)-1),times=(nrow(data1)-1):1, each=1)))
      
      if(any(data_plus$numvars)){
        d1 <- abs(data1[rep(1:(nrow(data1)-1),times=(nrow(data1)-1):1, each=1), data_plus$numvars, drop = FALSE] - 
                    data1[unlist(lapply(2:nrow(data1), seq, to=nrow(data1))), data_plus$numvars, drop = FALSE])
        for(jnum in 1:ncol(d1)) d1[,jnum] <- d1[,jnum] / data_plus$rgnums[jnum] 
        #time efficient alternative for codeline before:
        # d1 <- d1 / rep(rgnums, rep(nrow(d1), ncol(d1)))
        d1[is.na(d1)] <- 0
        if(length(lambda) > 1) d1 <- as.matrix(d1) %*% lambda[data_plus$numvars]
        if(is.null(lambda)) d1 <- rowSums(d1)
      }
      
      if(any(data_plus$catvars)){
        d2 <- data1[rep(1:(nrow(data1)-1),times=(nrow(data1)-1):1, each=1), data_plus$catvars, drop = FALSE] != 
          data1[unlist(lapply(2:nrow(data1), seq, to=nrow(data1))), data_plus$catvars, drop = FALSE]
        d2[is.na(d2)] <- FALSE
        if(length(lambda) > 1) d2 <- as.matrix(d2) %*% lambda[data_plus$catvars]
        if(is.null(lambda)) d2 <- rowSums(d2)
      }
      
      if(any(data_plus$ordvars)){
        d3 <- abs(data1[rep(1:(nrow(data1)-1),times=(nrow(data1)-1):1, each=1), data_plus$ordvars, drop = FALSE] - 
                    data1[unlist(lapply(2:nrow(data1), seq, to=nrow(data1))), data_plus$ordvars, drop = FALSE])
        for(jord in 1:ncol(d3)) d3[,jord] <- d3[,jord] / data_plus$rgords[jord] 
        d3[is.na(d3)] <- 0
        if(length(lambda) > 1) d3 <- as.matrix(d3) %*% lambda[data_plus$ordvars]
        if(is.null(lambda)) d3 <- rowSums(d3)
      }
      
      if(all.dists == TRUE){ #return distance matrix
        m <- matrix(rep(0,nrow(data1)^2), ncol = nrow(data1))
        
        m[lower.tri(m)] <- d1+d2+d3
        m <- t(m)
        m[lower.tri(m)] <- d1+d2+d3
        return(m)
      }else{ #return the sum of all dists between objects
        return(sum(d1 + d2 + d3))
      }
      
    }
  }
}


create.S_w <- function(object, data_plus){
  S_w <- 0
  for(k in seq_along(object$size)){
    x_k <- object$data[which(object$cluster==k),]
    if(object$size[k]>1){
      S_w <- S_w + calc.dist(lambda = object$lambda, type = object$type, 
                             data1 = x_k, 
                             data_plus = data_plus)
    }
  }
  return(S_w)
}


create.S_b <- function(object, data_plus){
  S_b <- 0
  for(k in 1:(length(object$size)-1)){
    x_k <- object$data[which(object$cluster==k),]
    for(l in (k+1):length(object$size)){
      S_b <- S_b + calc.dist(type = object$type, lambda = object$lambda,
                             data1 = x_k, 
                             data2 = object$data[which(object$cluster==l),], 
                             data_plus = data_plus)
    }
  }
  return(S_b)
}


create.N_w <- function(object){
  sum((object$size*(object$size-1))/2)
}




cindex_kproto <- function(object = NULL, type = NULL, data = NULL, k = NULL, S_sort = NULL, kp_obj = "optimal", lambda = NULL, data_plus = NULL, verbose = FALSE, ...){
  
  if(!is.null(object)){# it follows: determine the index value of the given cluster partition
    
    if(is.null(S_sort)){
      n <- length(object$cluster)
      S_all <- matrix(numeric(), ncol = n, nrow = n)
      for(i in 1:(n - 1)){
        for(j in (i + 1):n){
          S_all[i,j] <- calc.dist(type = object$type, lambda = object$lambda,
                                  data1 = object$data[i,], data2 = object$data[j,],
                                  data_plus = data_plus)
          
        }
      }
      S_sort <- sort(S_all)
    }
    
    S_w <- create.S_w(object, data_plus = data_plus)
    
    N_w <- create.N_w(object)
    S_min <- sum(head(S_sort,n = N_w))
    S_max <- sum(tail(S_sort,n = N_w))
    
    if(S_min != S_max){
      index <- (S_w - S_min)/(S_max - S_min)
    }
    
    return(index)
  }else{# determine index-optimal cluster partition
    
    n <- nrow(data)
    S_all <- matrix(numeric(), ncol=n, nrow=n)
    for(i in 1:(n - 1)){
      for(j in (i + 1):n){
        S_all[i,j] <- calc.dist(type = type, lambda = lambda, 
                                data1 = data[i,], data2 = data[j,],
                                data_plus = data_plus)
      }
    }
    S_sort <- sort(S_all)
    
    #calculate all kproto objects for k
    if(type == "gower"){
      # ...replace ranks with original ordered values
      data[,which(data_plus$ordvars)] <- data_plus$data_ord
    }
    object <- kproto(x = data, k = k[1], type = type, keep.data = TRUE, lambda = lambda, verbose = verbose, ...)
    if(type == "gower"){
      # ...replace ranks with original ordered values
      for(jord in which(data_plus$ordvars)) object$data[,jord] <- rank(data[,jord])
    }
    trace_kp <- list(list("index" = cindex_kproto(object = object, S_sort = S_sort, data_plus = data_plus), 
                          "k" = length(object$size), "object" = object))
    for(q in k[-1]){
      if(type == "gower"){
        # ...replace ranks with original ordered values
        data[,which(data_plus$ordvars)] <- data_plus$data_ord
      }
      object <- kproto(x = data, k = q, type = type, keep.data = TRUE, lambda = lambda, verbose = verbose)#, ...)
      if(type == "gower"){
        # ...replace ranks with original ordered values
        for(jord in which(data_plus$ordvars)) object$data[,jord] <- rank(data[,jord])
      }
      #save kproto object, if there isn't an object for this number of cluster
      if(!any(lapply(trace_kp, `[[`, 2) == length(object$size))){
        trace_kp <- c(trace_kp, list(list("index" = cindex_kproto(object = object, S_sort = S_sort, data_plus = data_plus), 
                                          "k" = length(object$size), "object" = object)))
      }else{
        # save kproto object if there is a cluster partition with same number of cluster but different validation index
        index_value <- cindex_kproto(object = object, S_sort = S_sort, data_plus = data_plus)
        if(!(index_value %in% unlist(lapply(trace_kp, `[[`, 1))[which(unlist(lapply(trace_kp, `[[`, 2)) == length(object$size))])){
          trace_kp <- c(trace_kp, list(list("index" = index_value, "k" = length(object$size), "object" = object)))
        }
      }
    }
    
    # save all index values for comparison and as summary for output
    indices <- unlist(lapply(trace_kp, `[[`, 1))
    names(indices) <- lapply(trace_kp, `[[`, 2)
    
    # find the optimal k, if it is ambiguously: sample
    k_m <- which(indices == min(indices))
    if(length(k_m)>1){k_m <- sample(k_m,1)}
    k_opt <- as.integer(names(indices[k_m]))
    index_opt <- indices[k_m]
    
    output <- switch(kp_obj,
                     "all" = list("k_opt" = k_opt, "index_opt" = index_opt, "indices" = indices, "kp_obj" = trace_kp, "type" = type),
                     "optimal" = list("k_opt" = k_opt, "index_opt" = index_opt, "indices" = indices, "kp_obj" = trace_kp[[k_m]]$object, "type" = type))
    return(output)
  }
}


dunn_kproto <- function(object = NULL, type = NULL, data = NULL, k = NULL, kp_obj = "optimal", lambda = NULL, data_plus = NULL, verbose = FALSE, ...){
  
  cond <- FALSE
  if(!is.null(object)){
    if(!is.null(object$lambda)){
      cond <- all(!as.logical(data_plus$numvars*object$lambda))
    }
  }else{
    if(!is.null(lambda)){
      cond <- all(!as.logical(data_plus$numvars*lambda))
    }
  }
  if(cond) message("As a result of the choice of lambda: No numeric variables in x! Index calculation might result in NA...\n")
  
  if(!is.null(object)){
    
    k <- length(object$size)
    
    #determine d(C_i,C_j)
    min_CiCj <- matrix(numeric(k*k), ncol = k, nrow = k)
    for(i in 1:(k-1)){
      xi <- object$data[which(object$cluster == i),]
      for(j in (i+1):k){
        xj <- object$data[which(object$cluster == j),]
        min_ij <- calc.dist(type = object$type, lambda = object$lambda, 
                            data1 = xi[1,], data2 = xj[1,],
                            data_plus = data_plus)
        for(l in 1:object$size[i]){
          for(m in 1:object$size[j]){
            min_neu <- calc.dist(type = object$type, lambda = object$lambda, 
                                 data1 = xi[l,], data2 = xj[m,],
                                 data_plus = data_plus)
            if(min_neu < min_ij){
              min_ij <- min_neu
            }
          }
        }
        min_CiCj[i,j] <- min_ij
      }
    }
    
    if(length(min_CiCj[min_CiCj > 0]) > 0){
      numerator <- min(min_CiCj[min_CiCj > 0])
    }else{
      return(NA)
    }
    
    
    #determine diam(C_k)
    max_diam <- numeric(k)
    for(p in 1:k){
      xi <- object$data[which(object$cluster == p),]
      if(object$size[p] > 1){
        max_ij <- calc.dist(type = object$type, lambda = object$lambda, 
                            data1 = xi[1,], data2 = xi[2,],
                            data_plus = data_plus)
        for(l in 1:(object$size[p]-1)){
          for(m in (l+1):(object$size[p])){
            max_neu <- calc.dist(type = object$type, lambda = object$lambda, 
                                 data1 = xi[l,], data2 = xi[m,],
                                 data_plus = data_plus)
            if(max_neu > max_ij){
              max_ij <- max_neu
            }
          }
        }
        max_diam[p] <- max_ij
      }
    }
    denominator <- max(max_diam)
    
    if(is.finite(numerator/denominator)){
      return(numerator/denominator)
    }else{
      return(NA)
    }
    
    
  }else{
    n <- nrow(data)
    
    #calculate all kproto objects for k
    object <- kproto(x = data, k = k[1], type = type, keep.data = TRUE, lambda = lambda, verbose = verbose, ...)
    trace_kp <- list(list("index" = dunn_kproto(object = object, data_plus = data_plus), 
                          "k" = length(object$size), "object" = object))
    for(q in k[-1]){
      object <- kproto(x = data, k = q, type = type, keep.data = TRUE, lambda = lambda, verbose = verbose, ...)
      #save kproto object, if there isn't an object for this number of cluster
      if(!any(lapply(trace_kp, `[[`, 2) == length(object$size))){
        trace_kp <- c(trace_kp, list(list("index" = dunn_kproto(object = object, data_plus = data_plus), 
                                          "k" = length(object$size), "object" = object)))
      }else{
        # save kproto object if there is a clusterpartition with same number of cluster but different validation index
        index_value <- dunn_kproto(object = object, data_plus = data_plus)
        if(!(index_value %in% unlist(lapply(trace_kp, `[[`, 1))[which(unlist(lapply(trace_kp, `[[`, 2)) == length(object$size))])){
          trace_kp <- c(trace_kp, list(list("index" = index_value, "k" = length(object$size), "object" = object)))
        }
      }
    }
    
    # save all index values for comparison and as summary for output
    indices <- unlist(lapply(trace_kp, `[[`, 1))
    names(indices) <- lapply(trace_kp, `[[`, 2)
    
    # find the optimal k, if it is ambiguously: sample
    if(all(is.na(indices))){return(NA)} # returning NA if dunn index couldn't be calculated (for all k), otherwise choose partition with highest index value
    k_m <- which(indices == max(indices, na.rm = TRUE))
    if(length(k_m)>1){k_m <- sample(k_m, 1)}
    k_opt <- as.integer(names(indices[k_m]))
    index_opt <- indices[k_m]
    
    output <- switch(kp_obj,
                     "all" = list("k_opt" = k_opt, "index_opt" = index_opt, "indices" = indices, "kp_obj" = trace_kp, "type" = type),
                     "optimal" = list("k_opt" = k_opt, "index_opt" = index_opt, "indices" = indices, "kp_obj" = trace_kp[[k_m]]$object, "type" = type))
    return(output)
    
  }
}


gamma_kproto <- function(object = NULL, type = NULL, data = NULL, k = NULL, dists = NULL, kp_obj = "optimal", lambda = NULL, data_plus = NULL, verbose = FALSE, ...){
  
  if(!is.null(object)){
    
    if(is.null(dists)){
      n <- nrow(object$data)
      dists <- matrix(numeric(), nrow = n, ncol = n)
      for(i in 1:(n - 1)){
        for(j in (i + 1):n){
          dists[i,j] <- calc.dist(type = object$type, lambda = object$lambda, 
                                  data1 = object$data[i,], data2 = object$data[j,],
                                  data_plus = data_plus)
        }
      }
    }
    
    s_plus <- 0
    s_minus <- 0
    dist_within <- 0
    dist_between <- 0
    for(k in as.numeric(names(object$size))){
      dist_within <- c(dist_within, na.omit(as.vector(dists[as.numeric(which(object$cluster == k)), as.numeric(which(object$cluster == k))])))
      dist_between <- c(dist_between, na.omit(as.vector(dists[as.numeric(which(object$cluster == k)), as.numeric(which(object$cluster != k))])))
    }
    dist_within <- dist_within[-1]
    for(m in 1:length(dist_within)){
      s_plus <- s_plus + sum(dist_within[m] < dist_between)
      s_minus <- s_minus + sum(dist_within[m] > dist_between)
    }
    index <- (s_plus - s_minus)/(s_plus + s_minus)
    
    return(index)
  }else{
    n <- nrow(data)
    
    if(is.null(dists)){
      dists <- matrix(numeric(), nrow = n, ncol = n)
      for(i in 1:(n - 1)){
        for(j in (i + 1):n){
          dists[i,j] <- calc.dist(type = type, lambda = lambda, 
                                  data1 = data[i,], data2 = data[j,],
                                  data_plus = data_plus)
        }
      }
    }
    
    index <- numeric(n)
    
    #calculate all kproto objects for k
    object <- kproto(x = data, k = k[1], type = type, keep.data = TRUE,  lambda = lambda, verbose = verbose, ...)
    trace_kp <- list(list("index" = gamma_kproto(object = object, dists = dists, data_plus = data_plus), 
                          "k" = length(object$size), "object" = object))
    for(q in k[-1]){
      object <- kproto(x = data, k = q, type = type, keep.data = TRUE, lambda = lambda, verbose = verbose, ...)
      #save kproto object, if there isn't an object for this number of cluster
      if(!any(lapply(trace_kp, `[[`, 2) == length(object$size))){
        trace_kp <- c(trace_kp, list(list("index" = gamma_kproto(object = object, dists = dists, data_plus = data_plus), 
                                          "k" = length(object$size), "object" = object)))
      }else{
        # save kproto object if there is a clusterpartition with same number of cluster but different validation index
        index_value <- gamma_kproto(object = object, dists = dists, data_plus = data_plus)
        if(!(index_value %in% unlist(lapply(trace_kp, `[[`, 1))[which(unlist(lapply(trace_kp, `[[`, 2)) == length(object$size))])){
          trace_kp <- c(trace_kp, list(list("index" = index_value, "k" = length(object$size), "object" = object)))
        }
      }
    }
    
    # save all index values for comparison and as summary for output
    indices <- unlist(lapply(trace_kp, `[[`, 1))
    names(indices) <- lapply(trace_kp, `[[`, 2)
    
    # find the optimal k, if it is ambiguously: sample
    k_m <- which(indices == max(indices))
    if(length(k_m)>1){k_m <- sample(k_m,1)}
    k_opt <- as.integer(names(indices[k_m]))
    index_opt <- indices[k_m]
    
    output <- switch(kp_obj,
                     "all" = list("k_opt" = k_opt, "index_opt" = index_opt, "indices" = indices, "kp_obj" = trace_kp, "type" = type),
                     "optimal" = list("k_opt" = k_opt, "index_opt" = index_opt, "indices" = indices, "kp_obj" = trace_kp[[k_m]]$object, "type" = type))
    return(output)
    
  }
}


gplus_kproto <- function(object = NULL, type = NULL, data = NULL, k = NULL, dists = NULL, kp_obj = "optimal", lambda = NULL, data_plus = NULL, verbose = FALSE, ...){
  
  if(!is.null(object)){
    
    n <- nrow(object$data)
    if(is.null(dists)){
      dists <- matrix(numeric(), nrow=n, ncol=n)
      for(i in 1:(n - 1)){
        for(j in (i + 1):n){
          dists[i,j] <- calc.dist(type = object$type, lambda = object$lambda, 
                                  data1 = object$data[i,], data2 = object$data[j,],
                                  data_plus = data_plus)
        }
      }
    }
    
    s_minus <- 0
    dist_within <- 0
    dist_between <- 0
    for(k in as.numeric(names(object$size))){
      dist_within <- c(dist_within, na.omit(as.vector(dists[as.numeric(which(object$cluster == k)),as.numeric(which(object$cluster == k))])))
      dist_between <- c(dist_between, na.omit(as.vector(dists[as.numeric(which(object$cluster == k)),as.numeric(which(object$cluster != k))])))
    }
    dist_within <- dist_within[-1]
    for(m in 1:length(dist_within)){
      s_minus <- s_minus + sum(dist_within[m] > dist_between)
    }
    N_t <- n*(n - 1)/2
    
    index <- (2 * s_minus)/(N_t * (N_t - 1))
    
    return(index)
  }else{
    n <- nrow(data)
    
    if(is.null(dists)){
      dists <- matrix(numeric(), nrow = n, ncol = n)
      for(i in 1:(n - 1)){
        for(j in (i + 1):n){
          dists[i,j] <- calc.dist(type = type, lambda = lambda, 
                                  data1 = data[i,], data2 = data[j,],
                                  data_plus = data_plus)
        }
      }
    }
    
    if(is.null(k)){k <- 2:sqrt(n)}
    
    #calculate all kproto objects for k
    object <- kproto(x = data, k = k[1], type = type, keep.data = TRUE, lambda = lambda, verbose = verbose, ...)
    trace_kp <- list(list("index" = gplus_kproto(object = object, dists = dists, data_plus = data_plus), 
                          "k" = length(object$size), "object" = object))
    for(q in k[-1]){
      object <- kproto(x = data, k = q, type = type, keep.data = TRUE, lambda = lambda, verbose = verbose, ...)
      #save kproto object, if there isn't an object for this number of cluster
      if(!any(lapply(trace_kp, `[[`, 2) == length(object$size))){
        trace_kp <- c(trace_kp, list(list("index" = gplus_kproto(object = object, dists = dists, data_plus = data_plus), 
                                          "k" = length(object$size), "object" = object)))
      }else{
        # save kproto object if there is a clusterpartition with same number of cluster but different validation index
        index_value <- gplus_kproto(object = object, dists = dists, data_plus = data_plus)
        if(!(index_value %in% unlist(lapply(trace_kp, `[[`, 1))[which(unlist(lapply(trace_kp, `[[`, 2)) == length(object$size))])){
          trace_kp <- c(trace_kp, list(list("index" = index_value, "k" = length(object$size), "object" = object)))
        }
      }
    }
    
    # save all index values for comparison and as summary for output
    indices <- unlist(lapply(trace_kp, `[[`, 1))
    names(indices) <- lapply(trace_kp, `[[`, 2)
    
    # find the optimal k, if it is ambiguously: sample
    k_m <- which(indices == min(indices))
    if(length(k_m)>1){k_m <- sample(k_m,1)}
    k_opt <- as.integer(names(indices[k_m]))
    index_opt <- indices[k_m]
    
    output <- switch(kp_obj,
                     "all" = list("k_opt" = k_opt, "index_opt" = index_opt, "indices" = indices, "kp_obj" = trace_kp, "type" = type),
                     "optimal" = list("k_opt" = k_opt, "index_opt" = index_opt, "indices" = indices, "kp_obj" = trace_kp[[k_m]]$object, "type" = type))
    return(output)
  }
}


mcclain_kproto <- function(object = NULL, type = NULL, data = NULL, k = NULL, kp_obj = "optimal", lambda = NULL, data_plus = NULL, verbose = FALSE, ...){
  
  if(!is.null(object)){
    n <- nrow(object$data)
    S_w_kpa <- create.S_w(object, data_plus = data_plus)
    S_b_kpa <- create.S_b(object, data_plus = data_plus)
    N_w <- create.N_w(object)
    N_t <- n*(n - 1)/2
    N_b <- N_t-N_w
    
    index <- (S_w_kpa/N_w)/(S_b_kpa/N_b)
    
    return(index)
  }else{
    #calculate all kproto objects for k
    object <- kproto(x = data, k = k[1], type = type, keep.data = TRUE, lambda = lambda, verbose = verbose, ...)
    trace_kp <- list(list("index" = mcclain_kproto(object = object, data_plus = data_plus), 
                          "k" = length(object$size), "object" = object))
    for(q in k[-1]){
      object <- kproto(x = data, k = q, type = type, keep.data = TRUE, lambda = lambda, verbose = verbose, ...)
      #save kproto object, if there isn't an object for this number of cluster
      if(!any(lapply(trace_kp, `[[`, 2) == length(object$size))){
        trace_kp <- c(trace_kp, list(list("index" = mcclain_kproto(object = object, data_plus = data_plus), 
                                          "k" = length(object$size), "object" = object)))
      }else{
        # save kproto object if there is a cluster partition with same number of cluster but different validation index
        index_value <- mcclain_kproto(object = object, data_plus = data_plus)
        if(!(index_value %in% unlist(lapply(trace_kp, `[[`, 1))[which(unlist(lapply(trace_kp, `[[`, 2)) == length(object$size))])){
          trace_kp <- c(trace_kp, list(list("index" = index_value, "k" = length(object$size), "object" = object)))
        }
      }
    }
    
    # save all index values for comparison and as summary for output
    indices <- unlist(lapply(trace_kp, `[[`, 1))
    names(indices) <- lapply(trace_kp, `[[`, 2)
    
    # find the optimal k, if it is ambiguously: sample
    k_m <- which(indices == min(indices))
    if(length(k_m)>1){k_m <- sample(k_m,1)}
    k_opt <- as.integer(names(indices[k_m]))
    index_opt <- indices[k_m]
    
    output <- switch(kp_obj,
                     "all" = list("k_opt" = k_opt, "index_opt" = index_opt, "indices" = indices, "kp_obj" = trace_kp, "type" = type),
                     "optimal" = list("k_opt" = k_opt, "index_opt" = index_opt, "indices" = indices, "kp_obj" = trace_kp[[k_m]]$object, "type" = type))
    return(output)
  }
}


ptbiserial_kproto <- function(object = NULL, type = NULL, data = NULL, k = NULL, s_d = NULL, kp_obj = "optimal", lambda = NULL, data_plus = NULL, verbose = FALSE, ...){
  
  if(!is.null(object)){
    
    n <- nrow(object$data)
    if(is.null(s_d)){
      S_all <- matrix(numeric(), ncol = n, nrow = n)
      for(i in 1:(n - 1)){
        for(j in (i + 1):n){
          S_all[i,j] <- calc.dist(type = object$type, lambda = object$lambda, 
                                  data1 = object$data[i,], data2 = object$data[j,],
                                  data_plus = data_plus)
        }
      }
      s_d <- sd(S_all, na.rm = TRUE)
    }
    
    S_w_kpa <- create.S_w(object, data_plus = data_plus)
    S_b_kpa <- create.S_b(object, data_plus = data_plus)
    N_w <- create.N_w(object)
    N_t <- n*(n-1)/2
    N_b <- N_t-N_w
    
    index <- ((S_b_kpa/N_b) - (S_w_kpa/N_w)) * sqrt(N_w * N_b/N_t^2) / s_d
    
    return(index)
  }else{
    
    n <- nrow(data)
    if(is.null(s_d)){
      S_all <- matrix(numeric(), ncol = n, nrow = n)
      for(i in 1:(n-1)){
        for(j in (i+1):n){
          S_all[i,j] <- calc.dist(type = type, lambda = lambda,
                                  data1 = data[i,], data2 = data[j,],
                                  data_plus = data_plus)
        }
      }
      s_d <- sd(S_all, na.rm = TRUE)
    }
    
    #calculate all kproto objects for k
    object <- kproto(x = data, k = k[1], type = type, keep.data = TRUE, lambda = lambda, verbose = verbose, ...)
    trace_kp <- list(list("index" = ptbiserial_kproto(object = object, s_d = s_d, data_plus = data_plus), 
                          "k" = length(object$size), "object" = object))
    for(q in k[-1]){
      object <- kproto(x = data, k = q, type = type, keep.data = TRUE, lambda = lambda, verbose = verbose, ...)
      #save kproto object, if there isn't an object for this number of cluster
      if(!any(lapply(trace_kp, `[[`, 2) == length(object$size))){
        trace_kp <- c(trace_kp, list(list("index" = ptbiserial_kproto(object = object, s_d = s_d, data_plus = data_plus), 
                                          "k" = length(object$size), "object" = object)))
      }else{
        # save kproto object if there is a cluster partition with same number of cluster but different validation index
        index_value <- ptbiserial_kproto(object = object, s_d = s_d, data_plus = data_plus)
        if(!(index_value %in% unlist(lapply(trace_kp, `[[`, 1))[which(unlist(lapply(trace_kp, `[[`, 2)) == length(object$size))])){
          trace_kp <- c(trace_kp, list(list("index" = index_value, "k" = length(object$size), "object" = object)))
        }
      }
    }
    
    # save all index values for comparison and as summary for output
    indices <- unlist(lapply(trace_kp, `[[`, 1))
    names(indices) <- lapply(trace_kp, `[[`, 2)
    
    # find the optimal k, if it is ambiguously: sample
    k_m <- which(indices == max(indices))
    if(length(k_m)>1){k_m <- sample(k_m,1)}
    k_opt <- as.integer(names(indices[k_m]))
    index_opt <- indices[k_m]
    
    output <- switch(kp_obj,
                     "all" = list("k_opt" = k_opt, "index_opt" = index_opt, "indices" = indices, "kp_obj" = trace_kp, "type" = type),
                     "optimal" = list("k_opt" = k_opt, "index_opt" = index_opt, "indices" = indices, "kp_obj" = trace_kp[[k_m]]$object, "type" = type))
    return(output)
  }
}


silhouette_kproto <- function(object = NULL, type = NULL, data = NULL, k = NULL, kp_obj = "optimal", lambda = NULL, data_plus = NULL, verbose = FALSE, ...){
  
  if(!is.null(object)){
    
    dists <- calc.dist(type = object$type, lambda = object$lambda, 
                       data1 = object$data, all.dists = TRUE,
                       data_plus = data_plus)
    
    cluster_dists <- matrix(numeric(nrow(object$data)*length(table(object$cluster))), 
                            nrow = nrow(object$data), ncol = length(table(object$cluster)))
    for(i in 1:length(table(object$cluster))){
      if(!(length(which(object$cluster == i)) == 1)){
        cluster_dists[,i] <- rowMeans(dists[,which(object$cluster == i)])
      }else{
        cluster_dists[,i] <- dists[,which(object$cluster == i)]
      }
    }
    
    #determine ai, bi and si
    a <- numeric(nrow(object$data))
    b <- numeric(nrow(object$data))
    s <- numeric(nrow(object$data))
    for(i in 1:nrow(object$data)){
      if(is.na(object$cluster[i])){
        # special case: no cluster assignment cluster[i] (usually resulting of all variables NA in x[i])
        s[i] <- NA
      }else{
        a[i] <- cluster_dists[i, object$cluster[i]]
        b[i] <- min(cluster_dists[i, -object$cluster[i]])
        
        if(max(a[i], b[i], na.rm = TRUE) == 0){
          # special case: a[i]=0 and b[i]=0 => s = 0, since x[i] lies equally far away (distance = 0) from both the clusters
          s[i] <- 0
        }else{
          s[i] <- (b[i] - a[i])/max(a[i],b[i], na.rm = TRUE)
        }
      }
      
    }
    if(any(table(object$cluster) == 1)){
      for(i in which(object$cluster %in% as.integer(which(table(object$cluster) == 1)))){
        s[i] <- 0
      }
      cat(length(which(object$cluster %in% as.integer(which(table(object$cluster) == 1))))," cluster with only one observation\n")
    }
    
    index <- mean(s, na.rm = TRUE)
    
    return(index)
  }else{
    #n <- nrow(data)
    
    #calculate all kproto objects for k
    object <- kproto(x = data, type = type, k = k[1], keep.data = TRUE, lambda = lambda, verbose = verbose, ...)
    trace_kp <- list(list("index" = silhouette_kproto(object = object, data_plus = data_plus), 
                          "k" = length(object$size), "object" = object))
    for(q in k[-1]){
      object <- kproto(x = data, type = type, k = q, keep.data = TRUE, lambda = lambda, verbose = verbose, ...)
      #save kproto object, if there isn't an object for this number of cluster
      if(!any(lapply(trace_kp, `[[`, 2) == length(object$size))){
        trace_kp <- c(trace_kp, list(list("index" = silhouette_kproto(object = object, data_plus = data_plus), 
                                          "k" = length(object$size), "object" = object)))
      }else{
        # save kproto object if there is a cluster partition with same number of cluster but different validation index
        index_value <- silhouette_kproto(object = object, data_plus = data_plus)
        if(!(index_value %in% unlist(lapply(trace_kp, `[[`, 1))[which(unlist(lapply(trace_kp, `[[`, 2)) == length(object$size))])){
          trace_kp <- c(trace_kp, list(list("index" = index_value, "k" = length(object$size), "object" = object)))
        }
      }
    }
    
    # save all index values for comparison and as summary for output
    indices <- unlist(lapply(trace_kp, `[[`, 1))
    names(indices) <- lapply(trace_kp, `[[`, 2)
    
    # find the optimal k, if it is ambiguously: sample
    k_m <- which(indices == max(indices, na.rm = TRUE))
    if(length(k_m)>1){k_m <- sample(k_m,1)}
    k_opt <- as.integer(names(indices[k_m]))
    index_opt <- indices[k_m]
    
    output <- switch(kp_obj,
                     "all" = list("k_opt" = k_opt, "index_opt" = index_opt, "indices" = indices, "kp_obj" = trace_kp, "type" = type),
                     "optimal" = list("k_opt" = k_opt, "index_opt" = index_opt, "indices" = indices, "kp_obj" = trace_kp[[k_m]]$object, "type" = type))
    return(output)
  }
}


tau_kproto <- function(object = NULL, type = NULL, data = NULL, k = NULL, dists = NULL, kp_obj = "optimal", lambda = NULL, data_plus = NULL, verbose = FALSE, ...){
  
  if(!is.null(object)){
    
    if(is.null(dists)){
      n <- nrow(object$data)
      dists <- matrix(numeric(), nrow = n, ncol = n)
      for(i in 1:(n - 1)){
        for(j in (i + 1):n){
          dists[i,j] <- calc.dist(type = object$type, lambda = object$lambda, 
                                  data1 = object$data[i,], data2 = object$data[j,],
                                  data_plus = data_plus)
        }
      }
    }
    
    s_plus <- 0
    s_minus <- 0
    dist_within <- 0
    dist_between <- 0
    for(k in as.numeric(names(object$size))){
      dist_within <- c(dist_within, na.omit(as.vector(dists[as.numeric(which(object$cluster == k)),as.numeric(which(object$cluster == k))])))
      dist_between <- c(dist_between, na.omit(as.vector(dists[as.numeric(which(object$cluster == k)),as.numeric(which(object$cluster != k))])))
    }
    dist_within <- dist_within[-1]
    for(m in 1:length(dist_within)){
      s_plus <- s_plus + sum(dist_within[m] < dist_between)
      s_minus <- s_minus + sum(dist_within[m] > dist_between)
    }
    
    n <- nrow(object$data)
    N_t <- n * (n - 1)/2
    M <- combn(1:n, m = 2)
    M <- rbind(M, apply(X = M, MARGIN = 2, function(x){
      if(any(is.na(c(object$cluster[x[1]], object$cluster[x[2]])))){return(-1)}else{
        if(object$cluster[x[1]] == object$cluster[x[2]]){return(1)}else{return(-1)}
      }
    }))
    t <- 0
    for(i in 1:(ncol(M)-1)){
      t <- t + length(which(M[3,i]*M[3,-(1:i)] > 0))
    }
    
    index <- (s_plus - s_minus)/sqrt((N_t * (N_t - 1) * 0.5 - t) * N_t * (N_t-1) * 0.5)
    
    return(index)
  }else{
    n <- nrow(data)
    
    if(is.null(dists)){
      dists <- matrix(numeric(), nrow = n, ncol = n)
      for(i in 1:(n-1)){
        for(j in (i+1):n){
          dists[i,j] <- calc.dist(type = type, lambda = lambda, 
                                  data1 = data[i,], data2 = data[j,],
                                  data_plus = data_plus)
        }
      }
    }
    
    if(is.null(k)){k <- 2:sqrt(n)}
    
    #calculate all kproto objects for k
    object <- kproto(x = data, k = k[1], type = type, keep.data = TRUE, lambda = lambda, verbose = verbose, ...)
    trace_kp <- list(list("index" = tau_kproto(object = object, dists = dists, data_plus = data_plus), 
                          "k" = length(object$size), "object" = object))
    for(q in k[-1]){
      object <- kproto(x = data, k = q, type = type, keep.data = TRUE, lambda = lambda, verbose = verbose, ...)
      #save kproto object, if there isn't an object for this number of cluster
      if(!any(lapply(trace_kp, `[[`, 2) == length(object$size))){
        trace_kp <- c(trace_kp, list(list("index" = tau_kproto(object = object, dists = dists, data_plus = data_plus), 
                                          "k" = length(object$size), "object" = object)))
      }else{
        # save kproto object if there is a clusterpartition with same number of cluster but different validation index
        index_value <- tau_kproto(object = object, dists = dists, data_plus = data_plus)
        if(!(index_value %in% unlist(lapply(trace_kp, `[[`, 1))[which(unlist(lapply(trace_kp, `[[`, 2)) == length(object$size))])){
          trace_kp <- c(trace_kp, list(list("index" = index_value, "k" = length(object$size), "object" = object)))
        }
      }
    }
    
    # save all index values for comparison and as summary for output
    indices <- unlist(lapply(trace_kp, `[[`, 1))
    names(indices) <- lapply(trace_kp, `[[`, 2)
    
    # find the optimal k, if it is ambiguously: sample
    k_m <- which(indices == max(indices))
    if(length(k_m)>1){k_m <- sample(k_m,1)}
    k_opt <- as.integer(names(indices[k_m]))
    index_opt <- indices[k_m]
    
    output <- switch(kp_obj,
                     "all" = list("k_opt" = k_opt, "index_opt" = index_opt, "indices" = indices, "kp_obj" = trace_kp, "type" = type),
                     "optimal" = list("k_opt" = k_opt, "index_opt" = index_opt, "indices" = indices, "kp_obj" = trace_kp[[k_m]]$object, "type" = type))
    return(output)
  }
}






#' @title Validating k Prototypes Clustering
#'
#' @description Calculating the preferred validation index for a k-Prototypes clustering with k clusters or computing the optimal number of clusters based on the choosen index for k-Prototype clustering. Possible validation indices are: \code{cindex}, \code{dunn}, \code{gamma}, \code{gplus}, \code{mcclain}, \code{ptbiserial}, \code{silhouette} and \code{tau}.
#' 
#' @param method Character specifying the validation index: \code{cindex}, \code{dunn}, \code{gamma}, \code{gplus}, \code{mcclain}, \code{ptbiserial}, \code{silhouette} (default) or \code{tau}.
#' @param object Object of class \code{kproto} resulting from a call with \code{kproto(..., keep.data=TRUE)}.
#' @param data Original data; only required if \code{object == NULL} and neglected if \code{object != NULL}.
#' @param type Character, to specify the distance for clustering; either \code{"huang"} or \code{"gower"}.
#' @param k Vector specifying the search range for optimum number of clusters; if \code{NULL} the range will set as \code{2:sqrt(n)}. Only required if \code{object == NULL} and neglected if \code{object != NULL}.
#' @param lambda Factor to trade off between Euclidean distance of numeric variables and simple matching coefficient between categorical variables.
#' @param kp_obj character either "optimal" or "all": Output of the index-optimal clustering (kp_obj == "optimal") or all computed cluster partitions (kp_obj == "all"); only required if \code{object != NULL}.
#' @param verbose Logical, whether additional information about process should be printed.
#' @param ... Further arguments passed to \code{\link[clustMixType]{kproto}}, like:
#'   \itemize{
#'     \item \code{nstart}: If > 1 repetitive computations of \code{kproto} with random initializations are computed.
#'     \item \code{na.rm}: Character, either \code{"yes"} to strip \code{NA} values for complete case analysis, \code{"no"} to keep and ignore \code{NA} values, \code{"imp.internal"} to impute the \code{NAs} within the algorithm or \code{"imp.onestep"} to apply the algorithm ignoring the \code{NAs} and impute them after the partition is determined.
#'   }
#' 
#' @details More information about the implemented validation indices:
#'   \itemize{
#'     \item \code{cindex} \deqn{Cindex = \frac{S_w-S_{min}}{S_{max}-S_{min}}} \cr
#' For \eqn{S_{min}} and \eqn{S_{max}} it is necessary to calculate the distances between all pairs of points in the entire data set (\eqn{\frac{n(n-1)}{2}}). 
#' \eqn{S_{min}} is the sum of the "total number of pairs of objects belonging to the same cluster" smallest distances and 
#' \eqn{S_{max}} is the sum of the "total number of pairs of objects belonging to the same cluster" largest distances. \eqn{S_w} is the sum of the within-cluster distances. \cr
#' The minimum value of the index is used to indicate the optimal number of clusters.
#' 
#'     \item \code{dunn} \deqn{Dunn = \frac{\min_{1 \leq i < j \leq q} d(C_i, C_j)}{\max_{1 \leq k \leq q} diam(C_k)}} \cr
#' The following applies: The dissimilarity between the two clusters \eqn{C_i} and \eqn{C_j} is defined as \eqn{d(C_i, C_j)=\min_{x \in C_i, y \in C_j} d(x,y)} and
#' the diameter of a cluster is defined as \eqn{diam(C_k)=\max_{x,y \in C} d(x,y)}. \cr
#' The maximum value of the index is used to indicate the optimal number of clusters.
#' 
#'     \item \code{gamma} \deqn{Gamma = \frac{s(+)-s(-)}{s(+)+s(-)}} \cr 
#' Comparisons are made between all within-cluster dissimilarities and all between-cluster dissimilarities. 
#' \eqn{s(+)} is the number of concordant comparisons and \eqn{s(-)} is the number of discordant comparisons.
#' A comparison is named concordant (resp. discordant) if a within-cluster dissimilarity is strictly less (resp. strictly greater) than a between-cluster dissimilarity.\cr
#' The maximum value of the index is used to indicate the optimal number of clusters.
#' 
#'     \item \code{gplus} \deqn{Gplus = \frac{2 \cdot s(-)}{\frac{n(n-1)}{2} \cdot (\frac{n(n-1)}{2}-1)}} \cr 
#' Comparisons are made between all within-cluster dissimilarities and all between-cluster dissimilarities. 
#' \eqn{s(-)} is the number of discordant comparisons and a comparison is named discordant if a within-cluster 
#' dissimilarity is strictly greater than a between-cluster dissimilarity. \cr
#' The minimum value of the index is used to indicate the optimal number of clusters.
#' 
#'     \item \code{mcclain} \deqn{McClain = \frac{\bar{S}_w}{\bar{S}_b}} \cr 
#' \eqn{\bar{S}_w} is the sum of within-cluster distances divided by the number of within-cluster distances and 
#' \eqn{\bar{S}_b} is the sum of between-cluster distances divided by the number of between-cluster distances.\cr
#' The minimum value of the index is used to indicate the optimal number of clusters.
#' 
#'     \item\code{ptbiserial} \deqn{Ptbiserial = \frac{(\bar{S}_b-\bar{S}_w) \cdot (\frac{N_w \cdot N_b}{N_t^2})^{0.5}}{s_d}} \cr 
#' \eqn{\bar{S}_w} is the sum of within-cluster distances divided by the number of within-cluster distances and 
#' \eqn{\bar{S}_b} is the sum of between-cluster distances divided by the number of between-cluster distances.\cr
#' \eqn{N_t} is the total number of pairs of objects in the data, \eqn{N_w} is the total number of pairs of 
#' objects belonging to the same cluster and \eqn{N_b} is the total number of pairs of objects belonging to different clusters.
#' \eqn{s_d} is the standard deviation of all distances.\cr
#' The maximum value of the index is used to indicate the optimal number of clusters.
#' 
#'     \item \code{silhouette} \deqn{Silhouette = \frac{1}{n} \sum_{i=1}^n \frac{b(i)-a(i)}{max(a(i),b(i))}} \cr 
#' \eqn{a(i)} is the average dissimilarity of the i\emph{th} object to all other objects of the same/own cluster.
#' \eqn{b(i)=min(d(i,C))}, where \eqn{d(i,C)} is the average dissimilarity of the i\emph{th} object to all the other clusters except the own/same cluster.\cr
#' The maximum value of the index is used to indicate the optimal number of clusters.
#'     
#'     \item \code{tau} \deqn{Tau = \frac{s(+) - s(-)}{((\frac{N_t(N_t-1)}{2}-t)\frac{N_t(N_t-1)}{2})^{0.5}}} \cr 
#' Comparisons are made between all within-cluster dissimilarities and all between-cluster dissimilarities. 
#' \eqn{s(+)} is the number of concordant comparisons and \eqn{s(-)} is the number of discordant comparisons.
#' A comparison is named concordant (resp. discordant) if a within-cluster dissimilarity is strictly less 
#' (resp. strictly greater) than a between-cluster dissimilarity.\cr
#' \eqn{N_t} is the total number of distances \eqn{\frac{n(n-1)}{2}} and \eqn{t} is the number of comparisons 
#' of two pairs of objects where both pairs represent within-cluster comparisons or both pairs are between-cluster
#' comparisons. \cr
#' The maximum value of the index is used to indicate the optimal number of clusters.
#'    
#'   }
#' 
#' @return For computing the optimal number of clusters based on the choosen validation index for k-Prototype clustering the output contains:
#' @return \item{k_opt}{optimal number of clusters (sampled in case of ambiguity)}
#' @return \item{index_opt}{index value of the index optimal clustering}
#' @return \item{indices}{calculated indices for \eqn{k=2,...,k_{max}}}
#' @return \item{kp_obj}{if(kp_obj == "optimal") the kproto object of the index optimal clustering and if(kp_obj == "all") all kproto which were calculated}
#' @return For computing the index-value for a given k-Prototype clustering the output contains:
#' @return \item{index}{calculated index-value}
#'
#' @author Rabea Aschenbruck
#' 
#' @references \itemize{
#'     \item Aschenbruck, R., Szepannek, G. (2020): 
#'     Cluster Validation for Mixed-Type Data. 
#'     \emph{Archives of Data Science, Series A, Vol 6, Issue 1}.
#'     \doi{10.5445/KSP/1000098011/02}.
#'     
#'     \item Charrad, M., Ghazzali, N., Boiteau, V., Niknafs, A. (2014): 
#'     NbClust: An R Package for Determining the Relevant Number of Clusters in a Data Set. 
#'     \emph{Journal of Statistical Software, Vol 61, Issue 6}.
#'     \doi{10.18637/jss.v061.i06}.
#'   }
#'
#' @examples
#' \dontrun{
#' # generate toy data with factors and numerics
#' n   <- 10
#' prb <- 0.99
#' muk <- 2.5 
#' 
#' x1 <- sample(c("A","B"), 2*n, replace = TRUE, prob = c(prb, 1-prb))
#' x1 <- c(x1, sample(c("A","B"), 2*n, replace = TRUE, prob = c(1-prb, prb)))
#' x1 <- as.factor(x1)
#' x2 <- sample(c("A","B"), 2*n, replace = TRUE, prob = c(prb, 1-prb))
#' x2 <- c(x2, sample(c("A","B"), 2*n, replace = TRUE, prob = c(1-prb, prb)))
#' x2 <- as.factor(x2)
#' x3 <- c(rnorm(n, mean = -muk), rnorm(n, mean = muk), rnorm(n, mean = -muk), rnorm(n, mean = muk))
#' x4 <- c(rnorm(n, mean = -muk), rnorm(n, mean = muk), rnorm(n, mean = -muk), rnorm(n, mean = muk))
#' x <- data.frame(x1,x2,x3,x4)
#' 
#' 
#' # calculate optimal number of cluster, index values and clusterpartition with Silhouette-index
#' val <- validation_kproto(method = "silhouette", data = x, k = 3:5, nstart = 5)
#' 
#' 
#' # apply k-prototypes
#' kpres <- kproto(x, 4, keep.data = TRUE)
#' 
#' # calculate cindex-value for the given clusterpartition
#' cindex_value <- validation_kproto(method = "cindex", object = kpres)
#' }
#' 
#' @rdname validation_kproto
#' 
#' @importFrom stats na.omit
#' @importFrom utils combn
#' @importFrom utils head
#' @importFrom utils tail
#' 
#' 
#' @export
validation_kproto <- function(method = "silhouette", object = NULL, data = NULL, type = "huang", k = NULL, lambda = NULL, kp_obj = "optimal", verbose = FALSE, ...){

  if(is.null(method)) stop("validation methode must be choosen!")
  if(!(method %in% c("cindex", "dunn", "gamma", "gplus", "mcclain", "ptbiserial", "silhouette", "tau"))) stop("choose one of these methods: cindex, dunn, gamma, gplus, mcclain, ptbiserial, silhouette, tau")
  
  if(is.null(data) & is.null(object)) stop("'data' or 'object' must be given!")
  
  if(!is.null(object) & !inherits(object, "kproto")) stop("object must be type of 'kproto'")
  if(!is.null(object) & is.null(object$data)) stop("kproto_object should have the original data included (kproto(..., keep.data = TRUE))")
  if(!is.null(object)){
    
    if(object$type == "gower"){
      data_plus <- list()
      
      data_plus$numvars <- sapply(object$data, is.numeric)
      data_plus$ordvars <- sapply(object$data, is.ordered)
      data_plus$catvars <- sapply(object$data, is.factor) & !data_plus$ordvars
      
      # ranges for normalization and transformed data for gower's distance calculation
      if(any(data_plus$numvars)) data_plus$rgnums <- sapply(object$data[, data_plus$numvars, drop = FALSE], function(z) diff(range(z)))
      if(any(data_plus$ordvars)){
        data_plus$data_ord <- object$data[, data_plus$ordvars, drop = FALSE] # store original variables 
        # ...and replace ordered variables by their ranks
        for(jord in which(data_plus$ordvars)) object$data[,jord] <- rank(object$data[,jord])
        data_plus$rgords <- sapply(object$data[, data_plus$ordvars, drop = FALSE], function(z) diff(range(z)))
      }
    }
    if(object$type == "huang"){
      data_plus <- list()
      data_plus$numvars <- sapply(object$data, is.numeric)
      data_plus$catvars <- sapply(object$data, is.factor)
    }
  }
  
  if(is.null(object)){
    if(!is.data.frame(data)) stop("data should be a data frame!")
    if(ncol(data) < 2) stop("for clustering data should contain at least two variables!")
    if(nrow(data) < 4) stop("for clustering data should contain at least four objects!")
    
    if(type %in% c("Huang", "standard")) type <- "huang"
    if(type == "Gower") type <- "gower"
    if(!type %in% c("huang", "gower")) stop("Argument type must be either 'huang' or 'gower'!")
    
    if(type == "gower"){
      type <- "gower"
      data_plus <- list()
      
      data_plus$numvars <- sapply(data, is.numeric)
      data_plus$ordvars <- sapply(data, is.ordered)
      data_plus$catvars <- sapply(data, is.factor) & !data_plus$ordvars
      
      # ranges for normalization and transformed data for gower's distance calculation
      if(any(data_plus$numvars)) data_plus$rgnums <- sapply(data[, data_plus$numvars, drop = FALSE], function(z) diff(range(z)))
      if(any(data_plus$ordvars)){
        data_plus$data_ord <- data[, data_plus$ordvars, drop = FALSE] # store original variables 
        # ...and replace ordered variables by their ranks
        for(jord in which(data_plus$ordvars)) data[,jord] <- rank(data[,jord])
        data_plus$rgords <- sapply(data[, data_plus$ordvars, drop = FALSE], function(z) diff(range(z)))
      }
      
      if(length(lambda) > 0 & length(lambda) != sum(c(data_plus$numvars, data_plus$catvars, data_plus$ordvars))) {
        warning("For gower distance if lambda is specified, its length must be the sum of numeric and factor variables in the data frame!")
        lambda <- NULL
      }
    }
    
    if(type == "huang"){
      data_plus <- list()
      
      data_plus$numvars <- sapply(data, is.numeric)
      data_plus$catvars <- sapply(data, is.factor)
      
      if(!is.null(lambda)){
        if(length(lambda) > 1){
          if(length(lambda) != sum(c(data_plus$numvars, data_plus$catvars))) stop("If lambda is a vector, its length should be the sum of numeric and factor variables in the data frame!")
        }else{
          if(length(lambda) == 1) {if(lambda == 0) stop("lambda has to be a value != 0. For automatic calculation use lambda = NULL (default setting)!")}
        }
      }else{
        lambda <- lambdaest(x = data, num.method = 1, fac.method = 1, outtype = "numeric", verbose = FALSE)
        if(verbose) cat("Estimated lambda:", lambda, "\n\n")
      }
    }
    
    if(!is.null(k) & length(k) == 1){
      stop("k should be the search range for optimum number of clusters, e.g. c(2:sqrt(n))")
    }
    if(length(k) > 1){
      if(nrow(data) < max(k)) stop("Data frame has less observations than clusters!")
      if(any(k <= 1) | any(k >= nrow(data))) stop("Elements of k must be greater than 1 and strictly less than n!")
      if(all(!as.integer(k)==k)) stop("Elements of k must be type of integer")
    }
    if(1 %in% k){
      if(method %in% c("cindex", "mcclain", "ptbiserial")) stop(paste("Index calculation not possible for k = 1. As the", method, "index is based on the relation of within-cluster distances and between-cluster distances, choose k > 1."))
      if(method %in% c("gamma", "gplus", "tau")) stop(paste("Index calculation not possible for k = 1. Since the", method, "index insists on comparing within-cluster distances and between-cluster distances, choose k > 1."))
      if(method == "dunn") stop("Index calculation not possible for k = 1. Calculation of the dunn index requires the determination of the distance between two clusters, so you have to set k > 1.")
      if(method == "silhouette") stop("Index calculation not possible for k = 1. The Silhouette index rates the average within-cluster distance for the own cluster and for the best alternative cluster for every object, so you have to set k > 1.")
    }
    if(length(object$size) == 1){
      if(method %in% c("cindex", "mcclain", "ptbiserial")) stop(paste("Index calculation not possible for kproto-objects with only one cluster. As the", method, "index is based on the relation of within-cluster distances and between-cluster distances a kproto-object which has been computed with k > 1 is required."))
      if(method %in% c("gamma", "gplus", "tau")) stop(paste("Index calculation not possible for kproto-objects with only one cluster. Since the", method, "index insists on comparing within-cluster distances and between-cluster distances a kproto-object which has been computed with k > 1 is required."))
      if(method == "dunn") stop("Index calculation not possible for kproto-objects with only one cluster. Calculation of the dunn index requires the determination of the distance between two clusters. You have to insert a kproto-object which has been computed with k > 1.")
      if(method == "silhouette") stop("Index calculation not possible for kproto-objects with only one cluster. The Silhouette index rates the average within-cluster distance for the own cluster and for the best alternative cluster for every object. You have to insert a kproto-object which has been computed with k > 1.")
    }
    if(is.null(k)){
      k <- 2:sqrt(nrow(data))
    }
  }
  
  if(!(kp_obj %in% c("optimal","all"))) stop("kp_obj must either be 'optimal' or 'all'!")
  
  output <- switch(method,
                   "cindex" = cindex_kproto(object = object, type = type, data = data, k = k, kp_obj = kp_obj, lambda = lambda, data_plus = data_plus, ...),
                   "dunn" = dunn_kproto(object = object, type = type, data = data, k = k, kp_obj = kp_obj, lambda = lambda, data_plus = data_plus, ...),
                   "gamma" = gamma_kproto(object = object, type = type, data = data, k = k, kp_obj = kp_obj, lambda = lambda, data_plus = data_plus, ...),
                   "gplus" = gplus_kproto(object = object, type = type, data = data, k = k, kp_obj = kp_obj, lambda = lambda, data_plus = data_plus, ...),
                   "mcclain" = mcclain_kproto(object = object, type = type, data = data, k = k, kp_obj = kp_obj, lambda = lambda, data_plus = data_plus, ...),
                   "ptbiserial" = ptbiserial_kproto(object = object, type = type, data = data, k = k, kp_obj = kp_obj, lambda = lambda, data_plus = data_plus, ...),
                   "silhouette" = silhouette_kproto(object = object, type = type, data = data, k = k, kp_obj = kp_obj, lambda = lambda, data_plus = data_plus, ...),
                   "tau" = tau_kproto(object = object, type = type, data = data, k = k, kp_obj = kp_obj, lambda = lambda, data_plus = data_plus, ...))
  
  return(output)
}

context("clustervalidation - basic tests.\n")

#library(clustMixType)

# generate test data
set.seed(42)
n   <- 10
prb <- 0.98
muk <- 3
clusid <- rep(1:4, each = n)

x1 <- sample(c("A","B"), 4*n, replace = TRUE, prob = c(prb, 1-prb))
x1 <- as.factor(x1)
x2 <- sample(c("A","B"), n, replace = TRUE, prob = c(1-prb, prb))
x2 <- c(x2, sample(c("A","B"), 3*n, replace = TRUE, prob = c(prb, 1-prb)))
x2 <- as.factor(x2)

x3 <- rnorm(4*n, mean = -muk)
x4 <- c(rnorm(n, mean = -muk), rnorm(n, mean = muk), rnorm(2*n, mean = -muk))

x5 <- rnorm(4*n) - muk
x6 <- rnorm(4*n) + c(rep(-muk, n), rep(-muk, n), rep(muk, n), rep(-muk, n))
x5 <- as.ordered(cut(x5, c(-Inf, -2, 1, Inf)))
x6 <- as.ordered(cut(x6, c(-Inf, -2, 1, Inf)))


x_num_cat <- data.frame(x1,x2,x3,x4)
x_num_cat_ord <- data.frame(x1,x2,x3,x4,x5,x6)
remove(n, prb, muk, clusid, x1, x2, x3, x4, x5, x6)

kpres_nc_huang <- clustMixType::kproto(x_num_cat, 4, verbose = FALSE, type = "huang")
kpres_nc_huang_wo <- clustMixType::kproto(x_num_cat, 4, verbose = FALSE, type = "huang", keep.data = FALSE)
kpres_nc_huang_wo_1 <- clustMixType::kproto(x_num_cat, 1, verbose = FALSE, type = "huang", keep.data = FALSE)
kpres_nc_gower <- clustMixType::kproto(x_num_cat, 4, verbose = FALSE, type = "gower")
kpres_nco_gower <- clustMixType::kproto(x_num_cat_ord, 4, verbose = FALSE, type = "gower")
kpres_nco_gower_wo <- clustMixType::kproto(x_num_cat_ord, 4, verbose = FALSE, type = "gower", keep.data = FALSE)
kpres_nco_gower_wo_1 <- clustMixType::kproto(x_num_cat_ord, 1, verbose = FALSE, type = "gower", keep.data = FALSE)



test_that("checking input objects",{
  #input method
  expect_error(validation_kproto(method = "other index", object = 2, verbose = FALSE))
  expect_error(validation_kproto(method = "cindex", object = 2, verbose = FALSE))

  # input data
  expect_error(validation_kproto(method = "cindex", data = NULL, verbose = FALSE))
  expect_error(validation_kproto(method = "cindex", data = data.frame("a" = 2,"b" = 2), verbose = FALSE))
  expect_error(validation_kproto(method = "cindex", data = data.frame("a" = c(2, 2)), verbose = FALSE))

  # input k
  expect_error(validation_kproto(method = "cindex", data = x_num_cat, k = 4, verbose = FALSE))
  expect_error(validation_kproto(method = "cindex", data = x_num_cat, k = -1:3, verbose = FALSE))
  expect_error(validation_kproto(method = "cindex", data = x_num_cat, k = 5:43, verbose = FALSE))
  expect_error(validation_kproto(method = "cindex", data = x_num_cat, k = 1, verbose = FALSE))
  expect_error(validation_kproto(method = "cindex", data = x_num_cat, k = 1:3, verbose = FALSE))

  #input object
  expect_error(validation_kproto(method = "cindex", data = NULL, object = kpres_nc_huang_wo, verbose = FALSE))
  expect_error(validation_kproto(method = "cindex", data = NULL, object = kpres_nco_gower_wo, verbose = FALSE))
  expect_error(validation_kproto(method = "cindex", data = NULL, object = kpres_nc_huang_wo_1, verbose = FALSE))
  expect_error(validation_kproto(method = "cindex", data = NULL, object = kpres_nco_gower_wo_1, verbose = FALSE))
}
)


context("clustervalidation - cindex.\n")
iv_nc_huang  <- validation_kproto(method = "cindex", object = kpres_nc_huang, verbose = FALSE)
iv_nc_gower  <- validation_kproto(method = "cindex", object = kpres_nc_gower, verbose = FALSE)
iv_nco_gower  <- validation_kproto(method = "cindex", object = kpres_nco_gower, verbose = FALSE)

ivk_nc_huang <- validation_kproto(method = "cindex", data = x_num_cat, k = c(3:4), verbose = FALSE, type = "huang")
ivk_nc_gower <- validation_kproto(method = "cindex", data = x_num_cat, k = c(3:4), verbose = FALSE, type = "gower")
ivk_nco_gower <- validation_kproto(method = "cindex", data = x_num_cat_ord, k = c(3:4), verbose = FALSE, type = "gower")

ivk_l_nc <- validation_kproto(method = "cindex", object = NULL, data = x_num_cat, k = c(3:4), kp_obj = 'optimal',
                                   lambda = 2, verbose = FALSE, type = "huang")
ivk_l_nco <- validation_kproto(method = "cindex", object = NULL, data = x_num_cat_ord, k = c(3:4), kp_obj = 'optimal',
                                       lambda = 2, verbose = FALSE, type = "gower")
ivk_huang_c <- validation_kproto(method = "cindex", object = NULL, data = x_num_cat, k = c(3:4),
                                        lambda = c(1,1,0,0), kp_obj = 'optimal', verbose = FALSE, type = "huang")
ivk_huang_n <- validation_kproto(method = "cindex", object = NULL, data = x_num_cat, k = c(3:4),
                                        lambda = c(0,0,1,1), kp_obj = 'optimal', verbose = FALSE, type = "huang")


test_that("resulting objects based on cindex are as expected",{
  ###validation index determination
  #nc_huang
  expect_is(iv_nc_huang, "numeric")
  expect_true(length(iv_nc_huang) == 1)
  expect_true(iv_nc_huang > 0)
  expect_true(iv_nc_huang < 1)
  #nc_gower
  expect_is(iv_nc_gower, "numeric")
  expect_true(length(iv_nc_gower) == 1)
  expect_true(iv_nc_gower > 0)
  expect_true(iv_nc_gower < 1)
  #nco_gower
  expect_is(iv_nco_gower, "numeric")
  expect_true(length(iv_nco_gower) == 1)
  expect_true(iv_nco_gower > 0)
  expect_true(iv_nco_gower < 1)

  ###determination index-optimal k
  #nc_huang
  expect_is(ivk_nc_huang[[1]], "integer")
  expect_true(length(ivk_nc_huang[[1]]) == 1)
  expect_true(ivk_nc_huang[[1]] %in% c(3:4))
  expect_true(is.numeric(ivk_nc_huang[[2]]))
  expect_true(is.numeric(ivk_nc_huang[[3]]))
  expect_true(length(ivk_nc_huang[[3]]) <= 3)
  expect_equal(ivk_nc_huang[[1]], as.integer(names(which.min(ivk_nc_huang[[3]]))))
  #nc_gower
  expect_is(ivk_nc_gower[[1]], "integer")
  expect_true(length(ivk_nc_gower[[1]]) == 1)
  expect_true(ivk_nc_gower[[1]] %in% c(3:4))
  expect_true(is.numeric(ivk_nc_gower[[2]]))
  expect_true(is.numeric(ivk_nc_gower[[3]]))
  expect_true(length(ivk_nc_gower[[3]]) <= 3)
  expect_equal(ivk_nc_gower[[1]], as.integer(names(which.min(ivk_nc_gower[[3]]))))
  #nco_gower
  expect_is(ivk_nco_gower[[1]], "integer")
  expect_true(length(ivk_nco_gower[[1]]) == 1)
  expect_true(ivk_nco_gower[[1]] %in% c(3:4))
  expect_true(is.numeric(ivk_nco_gower[[2]]))
  expect_true(is.numeric(ivk_nco_gower[[3]]))
  expect_true(length(ivk_nco_gower[[3]]) <= 3)
  expect_equal(ivk_nco_gower[[1]], as.integer(names(which.min(ivk_nco_gower[[3]]))))

  #checking lambda
  expect_equal(ivk_l_nc$kp_obj$lambda, 2)
  expect_equal(ivk_l_nco$kp_obj$lambda, NULL)
  expect_true(length(ivk_huang_c) == 5)
  expect_true(length(ivk_huang_n) == 5)
}
)


context("clustervalidation - dunn.\n")
iv_nc_huang  <- validation_kproto(method = "dunn", object = kpres_nc_huang, verbose = FALSE)
iv_nc_gower  <- validation_kproto(method = "dunn", object = kpres_nc_gower, verbose = FALSE)
iv_nco_gower  <- validation_kproto(method = "dunn", object = kpres_nco_gower, verbose = FALSE)

ivk_nc_huang <- validation_kproto(method = "dunn", data = x_num_cat, k = c(3:4), verbose = FALSE, type = "huang")
ivk_nc_gower <- validation_kproto(method = "dunn", data = x_num_cat, k = c(3:4), verbose = FALSE, type = "gower")
ivk_nco_gower <- validation_kproto(method = "dunn", data = x_num_cat_ord, k = c(3:4), verbose = FALSE, type = "gower")

ivk_l_nc <- validation_kproto(method = "dunn", object = NULL, data = x_num_cat, k = c(3:4), kp_obj = 'optimal',
                              lambda = 2, verbose = FALSE, type = "huang")
ivk_l_nco <- validation_kproto(method = "dunn", object = NULL, data = x_num_cat_ord, k = c(3:4), kp_obj = 'optimal',
                               lambda = 2, verbose = FALSE, type = "gower")
ivk_huang_c <- validation_kproto(method = "dunn", object = NULL, data = x_num_cat, k = c(3:4),
                                 lambda = c(1,1,0,0), kp_obj = 'optimal', verbose = FALSE, type = "huang")
ivk_huang_n <- validation_kproto(method = "dunn", object = NULL, data = x_num_cat, k = c(3:4),
                                 lambda = c(0,0,1,1), kp_obj = 'optimal', verbose = FALSE, type = "huang")


test_that("resulting objects based on dunn index are as expected",{
  ###validation index determination
  #nc_huang
  expect_is(iv_nc_huang, "numeric")
  expect_true(length(iv_nc_huang) == 1)
  expect_true(iv_nc_huang >= 0)
  #nc_gower
  expect_is(iv_nc_gower, "numeric")
  expect_true(length(iv_nc_gower) == 1)
  expect_true(iv_nc_gower >= 0)
  #nco_gower
  expect_is(iv_nco_gower, "numeric")
  expect_true(length(iv_nco_gower) == 1)
  expect_true(iv_nco_gower >= 0)

  ###determination index-optimal k
  #nc_huang
  expect_is(ivk_nc_huang[[1]], "integer")
  expect_true(length(ivk_nc_huang[[1]]) == 1)
  expect_true(ivk_nc_huang[[1]] %in% c(3:4))
  expect_true(is.numeric(ivk_nc_huang[[2]]))
  expect_true(is.numeric(ivk_nc_huang[[3]]))
  expect_true(length(ivk_nc_huang[[3]]) <= 3)
  expect_equal(ivk_nc_huang[[1]], as.integer(names(which.max(ivk_nc_huang[[3]]))))
  #nc_gower
  expect_is(ivk_nc_gower[[1]], "integer")
  expect_true(length(ivk_nc_gower[[1]]) == 1)
  expect_true(ivk_nc_gower[[1]] %in% c(3:4))
  expect_true(is.numeric(ivk_nc_gower[[2]]))
  expect_true(is.numeric(ivk_nc_gower[[3]]))
  expect_true(length(ivk_nc_gower[[3]]) <= 3)
  expect_equal(ivk_nc_gower[[1]], as.integer(names(which.max(ivk_nc_gower[[3]]))))
  #nco_gower
  expect_is(ivk_nco_gower[[1]], "integer")
  expect_true(length(ivk_nco_gower[[1]]) == 1)
  expect_true(ivk_nco_gower[[1]] %in% c(3:4))
  expect_true(is.numeric(ivk_nco_gower[[2]]))
  expect_true(is.numeric(ivk_nco_gower[[3]]))
  expect_true(length(ivk_nco_gower[[3]]) <= 3)
  expect_equal(ivk_nco_gower[[1]], as.integer(names(which.max(ivk_nco_gower[[3]]))))

  #checking lambda
  expect_equal(ivk_l_nc$kp_obj$lambda, 2)
  expect_equal(ivk_l_nco$kp_obj$lambda, NULL)
  expect_true(length(ivk_huang_c) == 5)
  expect_true(length(ivk_huang_n) == 5)
}
)


context("clustervalidation - gamma.\n")
iv_nc_huang  <- validation_kproto(method = "gamma", object = kpres_nc_huang, verbose = FALSE)
iv_nc_gower  <- validation_kproto(method = "gamma", object = kpres_nc_gower, verbose = FALSE)
iv_nco_gower  <- validation_kproto(method = "gamma", object = kpres_nco_gower, verbose = FALSE)

ivk_nc_huang <- validation_kproto(method = "gamma", data = x_num_cat, k = c(3:4), verbose = FALSE, type = "huang")
ivk_nc_gower <- validation_kproto(method = "gamma", data = x_num_cat, k = c(3:4), verbose = FALSE, type = "gower")
ivk_nco_gower <- validation_kproto(method = "gamma", data = x_num_cat_ord, k = c(3:4), verbose = FALSE, type = "gower")

ivk_l_nc <- validation_kproto(method = "gamma", object = NULL, data = x_num_cat, k = c(3:4), kp_obj = 'optimal',
                              lambda = 2, verbose = FALSE, type = "huang")
ivk_l_nco <- validation_kproto(method = "gamma", object = NULL, data = x_num_cat_ord, k = c(3:4), kp_obj = 'optimal',
                               lambda = 2, verbose = FALSE, type = "gower")
ivk_huang_c <- validation_kproto(method = "gamma", object = NULL, data = x_num_cat, k = c(3:4),
                                 lambda = c(1,1,0,0), kp_obj = 'optimal', verbose = FALSE, type = "huang")
ivk_huang_n <- validation_kproto(method = "gamma", object = NULL, data = x_num_cat, k = c(3:4),
                                 lambda = c(0,0,1,1), kp_obj = 'optimal', verbose = FALSE, type = "huang")

test_that("resulting objects based on gamma index are as expected",{
  ###validation index determination
  #nc_huang
  expect_is(iv_nc_huang, "numeric")
  expect_true(length(iv_nc_huang) == 1)
  expect_true(iv_nc_huang > -1)
  expect_true(iv_nc_huang <= 1)
  #nc_gower
  expect_is(iv_nc_gower, "numeric")
  expect_true(length(iv_nc_gower) == 1)
  expect_true(iv_nc_gower > -1)
  expect_true(iv_nc_gower <= 1)
  #nco_gower
  expect_is(iv_nco_gower, "numeric")
  expect_true(length(iv_nco_gower) == 1)
  expect_true(iv_nco_gower > -1)
  expect_true(iv_nco_gower <= 1)

  ###determination index-optimal k
  #nc_huang
  expect_is(ivk_nc_huang[[1]], "integer")
  expect_true(length(ivk_nc_huang[[1]]) == 1)
  expect_true(ivk_nc_huang[[1]] %in% c(3:4))
  expect_true(is.numeric(ivk_nc_huang[[2]]))
  expect_true(is.numeric(ivk_nc_huang[[3]]))
  expect_true(length(ivk_nc_huang[[3]]) <= 3)
  expect_equal(ivk_nc_huang[[1]], as.integer(names(which.max(ivk_nc_huang[[3]]))))
  #nc_gower
  expect_is(ivk_nc_gower[[1]], "integer")
  expect_true(length(ivk_nc_gower[[1]]) == 1)
  expect_true(ivk_nc_gower[[1]] %in% c(3:4))
  expect_true(is.numeric(ivk_nc_gower[[2]]))
  expect_true(is.numeric(ivk_nc_gower[[3]]))
  expect_true(length(ivk_nc_gower[[3]]) <= 3)
  expect_equal(ivk_nc_gower[[1]], as.integer(names(which.max(ivk_nc_gower[[3]]))))
  #nco_gower
  expect_is(ivk_nco_gower[[1]], "integer")
  expect_true(length(ivk_nco_gower[[1]]) == 1)
  expect_true(ivk_nco_gower[[1]] %in% c(3:4))
  expect_true(is.numeric(ivk_nco_gower[[2]]))
  expect_true(is.numeric(ivk_nco_gower[[3]]))
  expect_true(length(ivk_nco_gower[[3]]) <= 3)
  expect_equal(ivk_nco_gower[[1]], as.integer(names(which.max(ivk_nco_gower[[3]]))))

  #checking lambda
  expect_equal(ivk_l_nc$kp_obj$lambda, 2)
  expect_equal(ivk_l_nco$kp_obj$lambda, NULL)
  expect_true(length(ivk_huang_c) == 5)
  expect_true(length(ivk_huang_n) == 5)
}
)


context("clustervalidation - gplus.\n")
iv_nc_huang  <- validation_kproto(method = "gplus", object = kpres_nc_huang, verbose = FALSE)
iv_nc_gower  <- validation_kproto(method = "gplus", object = kpres_nc_gower, verbose = FALSE)
iv_nco_gower  <- validation_kproto(method = "gplus", object = kpres_nco_gower, verbose = FALSE)

ivk_nc_huang <- validation_kproto(method = "gplus", data = x_num_cat, k = c(3:4), verbose = FALSE, type = "huang")
ivk_nc_gower <- validation_kproto(method = "gplus", data = x_num_cat, k = c(3:4), verbose = FALSE, type = "gower")
ivk_nco_gower <- validation_kproto(method = "gplus", data = x_num_cat_ord, k = c(3:4), verbose = FALSE, type = "gower")

ivk_l_nc <- validation_kproto(method = "gplus", object = NULL, data = x_num_cat, k = c(3:4), kp_obj = 'optimal',
                              lambda = 2, verbose = FALSE, type = "huang")
ivk_l_nco <- validation_kproto(method = "gplus", object = NULL, data = x_num_cat_ord, k = c(3:4), kp_obj = 'optimal',
                               lambda = 2, verbose = FALSE, type = "gower")
ivk_huang_c <- validation_kproto(method = "gplus", object = NULL, data = x_num_cat, k = c(3:4),
                                 lambda = c(1,1,0,0), kp_obj = 'optimal', verbose = FALSE, type = "huang")
ivk_huang_n <- validation_kproto(method = "gplus", object = NULL, data = x_num_cat, k = c(3:4),
                                 lambda = c(0,0,1,1), kp_obj = 'optimal', verbose = FALSE, type = "huang")

test_that("resulting objects based on gplus index are as expected",{
  ###validation index determination
  #nc_huang
  expect_is(iv_nc_huang, "numeric")
  expect_true(length(iv_nc_huang) == 1)
  expect_true(iv_nc_huang >= 0)
  expect_true(iv_nc_huang < 1)
  #nc_gower
  expect_is(iv_nc_gower, "numeric")
  expect_true(length(iv_nc_gower) == 1)
  expect_true(iv_nc_gower >= 0)
  expect_true(iv_nc_gower < 1)
  #nco_gower
  expect_is(iv_nco_gower, "numeric")
  expect_true(length(iv_nco_gower) == 1)
  expect_true(iv_nco_gower >= 0)
  expect_true(iv_nco_gower < 1)

  ###determination index-optimal k
  #nc_huang
  expect_is(ivk_nc_huang[[1]], "integer")
  expect_true(length(ivk_nc_huang[[1]]) == 1)
  expect_true(ivk_nc_huang[[1]] %in% c(3:4))
  expect_true(is.numeric(ivk_nc_huang[[2]]))
  expect_true(is.numeric(ivk_nc_huang[[3]]))
  expect_true(length(ivk_nc_huang[[3]]) <= 3)
  expect_equal(ivk_nc_huang[[1]], as.integer(names(which.min(ivk_nc_huang[[3]]))))
  #nc_gower
  expect_is(ivk_nc_gower[[1]], "integer")
  expect_true(length(ivk_nc_gower[[1]]) == 1)
  expect_true(ivk_nc_gower[[1]] %in% c(3:4))
  expect_true(is.numeric(ivk_nc_gower[[2]]))
  expect_true(is.numeric(ivk_nc_gower[[3]]))
  expect_true(length(ivk_nc_gower[[3]]) <= 3)
  expect_equal(ivk_nc_gower[[1]], as.integer(names(which.min(ivk_nc_gower[[3]]))))
  #nco_gower
  expect_is(ivk_nco_gower[[1]], "integer")
  expect_true(length(ivk_nco_gower[[1]]) == 1)
  expect_true(ivk_nco_gower[[1]] %in% c(3:4))
  expect_true(is.numeric(ivk_nco_gower[[2]]))
  expect_true(is.numeric(ivk_nco_gower[[3]]))
  expect_true(length(ivk_nco_gower[[3]]) <= 3)
  expect_equal(ivk_nco_gower[[1]], as.integer(names(which.min(ivk_nco_gower[[3]]))))

  #checking lambda
  expect_equal(ivk_l_nc$kp_obj$lambda, 2)
  expect_equal(ivk_l_nco$kp_obj$lambda, NULL)
  expect_true(length(ivk_huang_c) == 5)
  expect_true(length(ivk_huang_n) == 5)
}
)


context("clustervalidation - mcclain.\n")
iv_nc_huang  <- validation_kproto(method = "mcclain", object = kpres_nc_huang, verbose = FALSE)
iv_nc_gower  <- validation_kproto(method = "mcclain", object = kpres_nc_gower, verbose = FALSE)
iv_nco_gower  <- validation_kproto(method = "mcclain", object = kpres_nco_gower, verbose = FALSE)

ivk_nc_huang <- validation_kproto(method = "mcclain", data = x_num_cat, k = c(3:4), verbose = FALSE, type = "huang")
ivk_nc_gower <- validation_kproto(method = "mcclain", data = x_num_cat, k = c(3:4), verbose = FALSE, type = "gower")
ivk_nco_gower <- validation_kproto(method = "mcclain", data = x_num_cat_ord, k = c(3:4), verbose = FALSE, type = "gower")

set.seed(42)
ivk_l_nc <- validation_kproto(method = "mcclain", object = NULL, data = x_num_cat, k = c(3:4), kp_obj = 'optimal', 
                              lambda = 2, verbose = FALSE, type = "huang")
ivk_l_nco <- validation_kproto(method = "mcclain", object = NULL, data = x_num_cat_ord, k = c(3:4), kp_obj = 'optimal', 
                               lambda = 2, verbose = FALSE, type = "gower")
ivk_huang_c <- validation_kproto(method = "mcclain", object = NULL, data = x_num_cat, k = c(3:4), 
                                 lambda = c(1,1,0,0), kp_obj = 'optimal', verbose = FALSE, type = "huang")
ivk_huang_n <- validation_kproto(method = "mcclain", object = NULL, data = x_num_cat, k = c(3:4), 
                                 lambda = c(0,0,1,1), kp_obj = 'optimal', verbose = FALSE, type = "huang")

test_that("resulting objects based on mcclain index are as expected",{
  ###validation index determination
  #nc_huang
  expect_is(iv_nc_huang, "numeric")
  expect_true(length(iv_nc_huang) == 1)
  expect_true(iv_nc_huang >= 0)
  #nc_gower
  expect_is(iv_nc_gower, "numeric")
  expect_true(length(iv_nc_gower) == 1)
  expect_true(iv_nc_gower >= 0)
  #nco_gower
  expect_is(iv_nco_gower, "numeric")
  expect_true(length(iv_nco_gower) == 1)
  expect_true(iv_nco_gower >= 0)
  
  ###determination index-optimal k
  #nc_huang
  expect_is(ivk_nc_huang[[1]], "integer")
  expect_true(length(ivk_nc_huang[[1]]) == 1)
  expect_true(ivk_nc_huang[[1]] %in% c(3:4))
  expect_true(is.numeric(ivk_nc_huang[[2]]))
  expect_true(is.numeric(ivk_nc_huang[[3]]))
  expect_true(length(ivk_nc_huang[[3]]) <= 3)
  expect_equal(ivk_nc_huang[[1]], as.integer(names(which.min(ivk_nc_huang[[3]]))))
  #nc_gower
  expect_is(ivk_nc_gower[[1]], "integer")
  expect_true(length(ivk_nc_gower[[1]]) == 1)
  expect_true(ivk_nc_gower[[1]] %in% c(3:4))
  expect_true(is.numeric(ivk_nc_gower[[2]]))
  expect_true(is.numeric(ivk_nc_gower[[3]]))
  expect_true(length(ivk_nc_gower[[3]]) <= 3)
  expect_equal(ivk_nc_gower[[1]], as.integer(names(which.min(ivk_nc_gower[[3]]))))
  #nco_gower
  expect_is(ivk_nco_gower[[1]], "integer")
  expect_true(length(ivk_nco_gower[[1]]) == 1)
  expect_true(ivk_nco_gower[[1]] %in% c(3:4))
  expect_true(is.numeric(ivk_nco_gower[[2]]))
  expect_true(is.numeric(ivk_nco_gower[[3]]))
  expect_true(length(ivk_nco_gower[[3]]) <= 3)
  expect_equal(ivk_nco_gower[[1]], as.integer(names(which.min(ivk_nco_gower[[3]]))))
  
  #checking lambda
  expect_equal(ivk_l_nc$kp_obj$lambda, 2)
  expect_equal(ivk_l_nco$kp_obj$lambda, NULL)
  expect_true(length(ivk_huang_c) == 5)
  expect_true(length(ivk_huang_n) == 5)
}
)


context("clustervalidation - ptbiserial.\n")
iv_nc_huang  <- validation_kproto(method = "ptbiserial", object = kpres_nc_huang, verbose = FALSE)
iv_nc_gower  <- validation_kproto(method = "ptbiserial", object = kpres_nc_gower, verbose = FALSE)
iv_nco_gower  <- validation_kproto(method = "ptbiserial", object = kpres_nco_gower, verbose = FALSE)

ivk_nc_huang <- validation_kproto(method = "ptbiserial", data = x_num_cat, k = c(3:4), verbose = FALSE, type = "huang")
ivk_nc_gower <- validation_kproto(method = "ptbiserial", data = x_num_cat, k = c(3:4), verbose = FALSE, type = "gower")
ivk_nco_gower <- validation_kproto(method = "ptbiserial", data = x_num_cat_ord, k = c(3:4), verbose = FALSE, type = "gower")

ivk_l_nc <- validation_kproto(method = "ptbiserial", object = NULL, data = x_num_cat, k = c(3:4), kp_obj = 'optimal',
                              lambda = 2, verbose = FALSE, type = "huang")
ivk_l_nco <- validation_kproto(method = "ptbiserial", object = NULL, data = x_num_cat_ord, k = c(3:4), kp_obj = 'optimal',
                               lambda = 2, verbose = FALSE, type = "gower")
ivk_huang_c <- validation_kproto(method = "ptbiserial", object = NULL, data = x_num_cat, k = c(3:4),
                                 lambda = c(1,1,0,0), kp_obj = 'optimal', verbose = FALSE, type = "huang")
ivk_huang_n <- validation_kproto(method = "ptbiserial", object = NULL, data = x_num_cat, k = c(3:4),
                                 lambda = c(0,0,1,1), kp_obj = 'optimal', verbose = FALSE, type = "huang")

test_that("resulting objects based on ptbiserial index are as expected",{
  ###validation index determination
  #nc_huang
  expect_is(iv_nc_huang, "numeric")
  expect_true(length(iv_nc_huang) == 1)
  #nc_gower
  expect_is(iv_nc_gower, "numeric")
  expect_true(length(iv_nc_gower) == 1)
  #nco_gower
  expect_is(iv_nco_gower, "numeric")
  expect_true(length(iv_nco_gower) == 1)

  ###determination index-optimal k
  #nc_huang
  expect_is(ivk_nc_huang[[1]], "integer")
  expect_true(length(ivk_nc_huang[[1]]) == 1)
  expect_true(ivk_nc_huang[[1]] %in% c(3:4))
  expect_true(is.numeric(ivk_nc_huang[[2]]))
  expect_true(is.numeric(ivk_nc_huang[[3]]))
  expect_true(length(ivk_nc_huang[[3]]) <= 3)
  expect_equal(ivk_nc_huang[[1]], as.integer(names(which.max(ivk_nc_huang[[3]]))))
  #nc_gower
  expect_is(ivk_nc_gower[[1]], "integer")
  expect_true(length(ivk_nc_gower[[1]]) == 1)
  expect_true(ivk_nc_gower[[1]] %in% c(3:4))
  expect_true(is.numeric(ivk_nc_gower[[2]]))
  expect_true(is.numeric(ivk_nc_gower[[3]]))
  expect_true(length(ivk_nc_gower[[3]]) <= 3)
  expect_equal(ivk_nc_gower[[1]], as.integer(names(which.max(ivk_nc_gower[[3]]))))
  #nco_gower
  expect_is(ivk_nco_gower[[1]], "integer")
  expect_true(length(ivk_nco_gower[[1]]) == 1)
  expect_true(ivk_nco_gower[[1]] %in% c(3:4))
  expect_true(is.numeric(ivk_nco_gower[[2]]))
  expect_true(is.numeric(ivk_nco_gower[[3]]))
  expect_true(length(ivk_nco_gower[[3]]) <= 3)
  expect_equal(ivk_nco_gower[[1]], as.integer(names(which.max(ivk_nco_gower[[3]]))))

  #checking lambda
  expect_equal(ivk_l_nc$kp_obj$lambda, 2)
  expect_equal(ivk_l_nco$kp_obj$lambda, NULL)
  expect_true(length(ivk_huang_c) == 5)
  expect_true(length(ivk_huang_n) == 5)
}
)


context("clustervalidation - silhouette.\n")
iv_nc_huang  <- validation_kproto(method = "silhouette", object = kpres_nc_huang, verbose = FALSE)
iv_nc_gower  <- validation_kproto(method = "silhouette", object = kpres_nc_gower, verbose = FALSE)
iv_nco_gower  <- validation_kproto(method = "silhouette", object = kpres_nco_gower, verbose = FALSE)

ivk_nc_huang <- validation_kproto(method = "silhouette", data = x_num_cat, k = c(3:4), verbose = FALSE, type = "huang")
ivk_nc_gower <- validation_kproto(method = "silhouette", data = x_num_cat, k = c(3:4), verbose = FALSE, type = "gower")
ivk_nco_gower <- validation_kproto(method = "silhouette", data = x_num_cat_ord, k = c(3:4), verbose = FALSE, type = "gower")

ivk_l_nc <- validation_kproto(method = "silhouette", object = NULL, data = x_num_cat, k = c(3:4), kp_obj = 'optimal',
                              lambda = 2, verbose = FALSE, type = "huang")
ivk_l_nco <- validation_kproto(method = "silhouette", object = NULL, data = x_num_cat_ord, k = c(3:4), kp_obj = 'optimal',
                               lambda = 2, verbose = FALSE, type = "gower")
ivk_huang_c <- validation_kproto(method = "silhouette", object = NULL, data = x_num_cat, k = c(3:4),
                                 lambda = c(1,1,0,0), kp_obj = 'optimal', verbose = FALSE, type = "huang")
ivk_huang_n <- validation_kproto(method = "silhouette", object = NULL, data = x_num_cat, k = c(3:4),
                                 lambda = c(0,0,1,1), kp_obj = 'optimal', verbose = FALSE, type = "huang")

test_that("resulting objects based on silhouette index are as expected",{
  ###validation index determination
  #nc_huang
  expect_is(iv_nc_huang, "numeric")
  expect_true(length(iv_nc_huang) == 1)
  expect_true(iv_nc_huang >= -1)
  expect_true(iv_nc_huang <= 1)
  #nc_gower
  expect_is(iv_nc_gower, "numeric")
  expect_true(length(iv_nc_gower) == 1)
  expect_true(iv_nc_gower >= -1)
  expect_true(iv_nc_gower <= 1)
  #nco_gower
  expect_is(iv_nco_gower, "numeric")
  expect_true(length(iv_nco_gower) == 1)
  expect_true(iv_nco_gower >= -1)
  expect_true(iv_nco_gower <= 1)

  ###determination index-optimal k
  #nc_huang
  expect_is(ivk_nc_huang[[1]], "integer")
  expect_true(length(ivk_nc_huang[[1]]) == 1)
  expect_true(ivk_nc_huang[[1]] %in% c(3:4))
  expect_true(is.numeric(ivk_nc_huang[[2]]))
  expect_true(is.numeric(ivk_nc_huang[[3]]))
  expect_true(length(ivk_nc_huang[[3]]) <= 3)
  expect_equal(ivk_nc_huang[[1]], as.integer(names(which.max(ivk_nc_huang[[3]]))))
  #nc_gower
  expect_is(ivk_nc_gower[[1]], "integer")
  expect_true(length(ivk_nc_gower[[1]]) == 1)
  expect_true(ivk_nc_gower[[1]] %in% c(3:4))
  expect_true(is.numeric(ivk_nc_gower[[2]]))
  expect_true(is.numeric(ivk_nc_gower[[3]]))
  expect_true(length(ivk_nc_gower[[3]]) <= 3)
  expect_equal(ivk_nc_gower[[1]], as.integer(names(which.max(ivk_nc_gower[[3]]))))
  #nco_gower
  expect_is(ivk_nco_gower[[1]], "integer")
  expect_true(length(ivk_nco_gower[[1]]) == 1)
  expect_true(ivk_nco_gower[[1]] %in% c(3:4))
  expect_true(is.numeric(ivk_nco_gower[[2]]))
  expect_true(is.numeric(ivk_nco_gower[[3]]))
  expect_true(length(ivk_nco_gower[[3]]) <= 3)
  expect_equal(ivk_nco_gower[[1]], as.integer(names(which.max(ivk_nco_gower[[3]]))))

  #checking lambda
  expect_equal(ivk_l_nc$kp_obj$lambda, 2)
  expect_equal(ivk_l_nco$kp_obj$lambda, NULL)
  expect_true(length(ivk_huang_c) == 5)
  expect_true(length(ivk_huang_n) == 5)
}
)


context("clustervalidation - tau.\n")
iv_nc_huang  <- validation_kproto(method = "tau", object = kpres_nc_huang, verbose = FALSE)
iv_nc_gower  <- validation_kproto(method = "tau", object = kpres_nc_gower, verbose = FALSE)
iv_nco_gower  <- validation_kproto(method = "tau", object = kpres_nco_gower, verbose = FALSE)

ivk_nc_huang <- validation_kproto(method = "tau", data = x_num_cat, k = c(3:4), verbose = FALSE, type = "huang")
ivk_nc_gower <- validation_kproto(method = "tau", data = x_num_cat, k = c(3:4), verbose = FALSE, type = "gower")
ivk_nco_gower <- validation_kproto(method = "tau", data = x_num_cat_ord, k = c(3:4), verbose = FALSE, type = "gower")

ivk_l_nc <- validation_kproto(method = "tau", object = NULL, data = x_num_cat, k = c(3:4), kp_obj = 'optimal',
                              lambda = 2, verbose = FALSE, type = "huang")
ivk_l_nco <- validation_kproto(method = "tau", object = NULL, data = x_num_cat_ord, k = c(3:4), kp_obj = 'optimal',
                               lambda = 2, verbose = FALSE, type = "gower")
ivk_huang_c <- validation_kproto(method = "tau", object = NULL, data = x_num_cat, k = c(3:4),
                                 lambda = c(1,1,0,0), kp_obj = 'optimal', verbose = FALSE, type = "huang")
ivk_huang_n <- validation_kproto(method = "tau", object = NULL, data = x_num_cat, k = c(3:4),
                                 lambda = c(0,0,1,1), kp_obj = 'optimal', verbose = FALSE, type = "huang")

test_that("resulting objects based on tau index are as expected",{
  ###validation index determination
  #nc_huang
  expect_is(iv_nc_huang, "numeric")
  expect_true(length(iv_nc_huang) == 1)
  expect_true(iv_nc_huang >= -1)
  expect_true(iv_nc_huang <= 1)
  #nc_gower
  expect_is(iv_nc_gower, "numeric")
  expect_true(length(iv_nc_gower) == 1)
  expect_true(iv_nc_gower >= -1)
  expect_true(iv_nc_gower <= 1)
  #nco_gower
  expect_is(iv_nco_gower, "numeric")
  expect_true(length(iv_nco_gower) == 1)
  expect_true(iv_nco_gower >= -1)
  expect_true(iv_nco_gower <= 1)

  ###determination index-optimal k
  #nc_huang
  expect_is(ivk_nc_huang[[1]], "integer")
  expect_true(length(ivk_nc_huang[[1]]) == 1)
  expect_true(ivk_nc_huang[[1]] %in% c(3:4))
  expect_true(is.numeric(ivk_nc_huang[[2]]))
  expect_true(is.numeric(ivk_nc_huang[[3]]))
  expect_true(length(ivk_nc_huang[[3]]) <= 3)
  expect_equal(ivk_nc_huang[[1]], as.integer(names(which.max(ivk_nc_huang[[3]]))))
  #nc_gower
  expect_is(ivk_nc_gower[[1]], "integer")
  expect_true(length(ivk_nc_gower[[1]]) == 1)
  expect_true(ivk_nc_gower[[1]] %in% c(3:4))
  expect_true(is.numeric(ivk_nc_gower[[2]]))
  expect_true(is.numeric(ivk_nc_gower[[3]]))
  expect_true(length(ivk_nc_gower[[3]]) <= 3)
  expect_equal(ivk_nc_gower[[1]], as.integer(names(which.max(ivk_nc_gower[[3]]))))
  #nco_gower
  expect_is(ivk_nco_gower[[1]], "integer")
  expect_true(length(ivk_nco_gower[[1]]) == 1)
  expect_true(ivk_nco_gower[[1]] %in% c(3:4))
  expect_true(is.numeric(ivk_nco_gower[[2]]))
  expect_true(is.numeric(ivk_nco_gower[[3]]))
  expect_true(length(ivk_nco_gower[[3]]) <= 3)
  expect_equal(ivk_nco_gower[[1]], as.integer(names(which.max(ivk_nco_gower[[3]]))))

  #checking lambda
  expect_equal(ivk_l_nc$kp_obj$lambda, 2)
  expect_equal(ivk_l_nco$kp_obj$lambda, NULL)
  expect_true(length(ivk_huang_c) == 5)
  expect_true(length(ivk_huang_n) == 5)
}
)






context("checking stability_kproto.\n")

# stab_nc_huang <- stability_kproto(object = kpres_nc_huang, method = c("luxburg","fowlkesmallows"))
# stab_nc_gower <- stability_kproto(object = kpres_nc_gower, method = c("luxburg","fowlkesmallows"))
# stab_nco_gower <- stability_kproto(object = kpres_nco_gower, method = c("luxburg","fowlkesmallows"))

test_that("checking input objects of stability_kproto",{
  expect_error(stability_kproto(method = "stability", object = kpres_nc_huang, verbose = FALSE))
  expect_error(stability_kproto(method = c("luxburg", "jaccard"), object = 42, verbose = FALSE))
  expect_error(stability_kproto(method = "jaccard", object = kpres_nc_huang_wo, verbose = FALSE))
  expect_error(stability_kproto(method = "jaccard", object = kpres_nc_huang, verbose = FALSE, B = "forty-two"))

}
)

stab_1_nc_huang  <- stability_kproto(method = "jaccard", object = kpres_nc_huang, verbose = FALSE)
stab_1_nc_gower  <- stability_kproto(method = "jaccard", object = kpres_nc_gower, verbose = FALSE)
stab_1_nco_gower  <- stability_kproto(method = "jaccard", object = kpres_nco_gower, verbose = FALSE)
stab_2_nc_huang  <- stability_kproto(method = c("luxburg", "jaccard"), object = kpres_nc_huang, verbose = FALSE)
stab_2_nc_gower  <- stability_kproto(method = c("luxburg", "jaccard"), object = kpres_nc_gower, verbose = FALSE)
stab_2_nco_gower  <- stability_kproto(method = c("luxburg", "jaccard"), object = kpres_nco_gower, verbose = FALSE)
stab_4_nc_huang  <- stability_kproto(method = c("luxburg", "rand", "fowlkesmallows", "jaccard"), object = kpres_nc_huang, verbose = FALSE)
stab_4_nc_gower  <- stability_kproto(method = c("luxburg", "rand", "fowlkesmallows", "jaccard"), object = kpres_nc_gower, verbose = FALSE)
stab_4_nco_gower  <- stability_kproto(method = c("luxburg", "rand", "fowlkesmallows", "jaccard"), object = kpres_nco_gower, verbose = FALSE)

test_that("resulting stability objects are as expected",{

  #nc_huang
  expect_is(stab_1_nc_huang, "list")
  expect_true(length(stab_1_nc_huang) == 2)
  expect_true(length(stab_2_nc_huang) == 2)
  expect_true(length(stab_4_nc_huang) == 2)

  expect_is(stab_1_nc_huang[[1]], "numeric")
  expect_true(length(stab_1_nc_huang[[1]]) == 1)
  expect_true(length(stab_2_nc_huang[[1]]) == 2)
  expect_true(length(stab_4_nc_huang[[1]]) == 4)

  expect_true(dim(stab_1_nc_huang[[2]])[[2]] == 2)
  expect_true(dim(stab_2_nc_huang[[2]])[[2]] == 3)
  expect_true(dim(stab_4_nc_huang[[2]])[[2]] == 5)

  #nc_gower
  expect_is(stab_1_nc_gower, "list")
  expect_true(length(stab_1_nc_gower) == 2)
  expect_true(length(stab_2_nc_gower) == 2)
  expect_true(length(stab_4_nc_gower) == 2)

  expect_is(stab_1_nc_gower[[1]], "numeric")
  expect_true(length(stab_1_nc_gower[[1]]) == 1)
  expect_true(length(stab_2_nc_gower[[1]]) == 2)
  expect_true(length(stab_4_nc_gower[[1]]) == 4)

  expect_true(dim(stab_1_nc_gower[[2]])[[2]] == 2)
  expect_true(dim(stab_2_nc_gower[[2]])[[2]] == 3)
  expect_true(dim(stab_4_nc_gower[[2]])[[2]] == 5)

  #nco_gower
  expect_is(stab_1_nco_gower, "list")
  expect_true(length(stab_1_nco_gower) == 2)
  expect_true(length(stab_2_nco_gower) == 2)
  expect_true(length(stab_4_nco_gower) == 2)

  expect_is(stab_1_nco_gower[[1]], "numeric")
  expect_true(length(stab_1_nco_gower[[1]]) == 1)
  expect_true(length(stab_2_nco_gower[[1]]) == 2)
  expect_true(length(stab_4_nco_gower[[1]]) == 4)

  expect_true(dim(stab_1_nco_gower[[2]])[[2]] == 2)
  expect_true(dim(stab_2_nco_gower[[2]])[[2]] == 3)
  expect_true(dim(stab_4_nco_gower[[2]])[[2]] == 5)
}
)


context("checking kproto2silhouettes.\n")

set.seed(42)
n   <- 50
prb <- 0.9
muk <- 1.5 
clusid <- rep(1:4, each = n)

x1 <- sample(c("A","B"), 2*n, replace = TRUE, prob = c(prb, 1-prb))
x1 <- c(x1, sample(c("A","B"), 2*n, replace = TRUE, prob = c(1-prb, prb)))
x1 <- as.factor(x1)

x2 <- sample(c("A","B"), 2*n, replace = TRUE, prob = c(prb, 1-prb))
x2 <- c(x2, sample(c("A","B"), 2*n, replace = TRUE, prob = c(1-prb, prb)))
x2 <- as.factor(x2)

x3 <- c(rnorm(n, mean = -muk), rnorm(n, mean = muk), rnorm(n, mean = -muk), rnorm(n, mean = muk))
x4 <- c(rnorm(n, mean = -muk), rnorm(n, mean = muk), rnorm(n, mean = -muk), rnorm(n, mean = muk))

x <- data.frame(x1,x2,x3,x4)

# apply k-prototypes
kpres <- kproto(x, 4)

# compute silhouette widhts
# extract object of class sil
sils <- kproto2silhouette(kpres)

test_that("checking class of kproto2silhouette",{
  expect_true(inherits(sils, "silhouette"))
  }
)

cindex_value <- validation_kproto(method = "silhouette", object = kpres)

test_that("comparing average silhouette with validation_kproto",{
  expect_true(round(mean(sils[,3]), 6) == round(cindex_value, 6))
  }
)

# determination of the number of cluster based on stability values
# since the research showed no superior advantage over the internal validation indices,
#    there is no export of this function
# (cf. paper: Aschenbruck, Szepannek, Wilhelm: 
#             Stability of mixed-type cluster partitions for determination of the number of clusters,
#             proceedings of the IFCS 2022)
# 
# Input: - data    - data.frame
#        - k       - numeric vector, search range
#        - method  - character, one of these: luxburg, fowlkesmallows, rand, jaccard
#        - B       - numeric, number of bootstrap samples
#        - lambda  - numeric, weighting factor of k-Prototypes
#        - ...     - Further arguments passed to \code{\link[clustMixType]{kproto}}
# 
# Output: the stability based number of cluster and additionally the decison based on the adjusted values
#         - k_opt      - optimal number of clusters (sampled in case of ambiguity)
#         - index_opt  - stability value of the optimal clustering
#         - indices    - calculated stability for \eqn{k=2,...,k_{max}}
#         - kp_obj     - the kproto object of the optimal clustering

stability_det_k <- function(data, k, method, B = 100, lambda = NULL, ...){
  
  if(length(method) > 1)stop("input has to be exactly one of luxburg, fowlkesmallows, rand or jaccard")
  if(!(method %in% c("luxburg", "fowlkesmallows", "rand", "jaccard"))) stop("input has to be one of luxburg, fowlkesmallows, rand or jaccard")
  
  if(!is.null(k) && length(k) == 1){
    stop("k should be the search range for optimum number of clusters, e.g. c(2:sqrt(n))")
  }
  if(length(k) > 1){
    if(nrow(data) < max(k)) stop("Data frame has less observations than clusters!")
    if(any(k <= 1) | any(k >= nrow(data))) stop("Elements of k must be greater than 1 and strictly less than n!")
    if(all(!as.integer(k)==k)) stop("Elements of k must be type of integer")
  }
  
  n <- nrow(data)
  p <- ncol(data)
  
  if(is.null(lambda)){
    numvars <- sapply(data, is.numeric)
    anynum <- any(numvars)
    catvars <- sapply(data, is.factor)
    anyfact <- any(catvars)
    vnum <- mean(sapply(data[,numvars, drop = FALSE], var, na.rm = TRUE))
    vcat <- mean(sapply(data[,catvars, drop = FALSE], function(z) return(1-sum((table(z)/sum(!is.na(z)))^2))))
    lambda <- vnum/vcat
  }
  
  #calculate all kproto objects for k and determine the stability values based on given method
  object <- kproto(x = data, k = k[1], keep.data = TRUE, lambda = lambda, verbose = FALSE, ...)
  trace_kp <- list(list("index" = stability_kproto(method = method, object = object, B = B, verbose = FALSE,...)[[1]][[1]], 
                        "k" = length(object$size), "object" = object))
  for(q in k[-1]){
    object <- kproto(x = data, k = q, keep.data = TRUE, lambda = lambda, verbose = FALSE, ...)
    #save kproto object, if there isn't an object for this number of cluster
    if(!any(lapply(trace_kp, `[[`, 2) == length(object$size))){
      trace_kp <- c(trace_kp, list(list("index" = stability_kproto(method = method, object = object, B = B, verbose = FALSE,...)[[1]][[1]], 
                                        "k" = length(object$size), "object" = object)))
    }else{
      # save kproto object if there is a clusterpartition with same number of cluster but different validation index
      index_value <- stability_kproto(method = method, object = object, B = B, verbose = FALSE,...)[[1]][[1]]
      if(!(index_value %in% unlist(lapply(trace_kp, `[[`, 1))[which(unlist(lapply(trace_kp, `[[`, 2)) == length(object$size))])){
        trace_kp <- c(trace_kp, list(list("index" = index_value, "k" = length(object$size), "object" = object)))
      }
    }
  }
  
  # save all index values for comparison and as summary for output
  indices <- unlist(lapply(trace_kp, `[[`, 1))
  names(indices) <- lapply(trace_kp, `[[`, 2)
  
  # find the optimal k, if it is ambiguously: sample
  # distinguish between minimum and maximum stability values:
  if(method == "luxburg"){
    k_m <- which(indices == min(indices, na.rm = TRUE))
  }else{
    k_m <- which(indices == max(indices, na.rm = TRUE))
  }
  if(length(k_m)>1){k_m <- sample(k_m,1)}
  k_opt <- as.integer(names(indices[k_m]))
  index_opt <- indices[k_m]
  
  # apply adjustment of paper
  adjustment <- data.frame(k_eval = names(indices), imean = indices) %>%
    arrange(.data$k_eval) %>% 
    mutate(Nachbar_vor = ifelse(is.na(lag(.data$imean)), 1, lag(.data$imean)),
           Nachbar_nach = ifelse(is.na(lead(.data$imean)), .data$imean, lead(.data$imean)),
           ivalue = .data$imean * sqrt(.data$imean/.data$Nachbar_vor * .data$imean/.data$Nachbar_nach))
  indices_adj <- adjustment$ivalue
  names(indices_adj) <- adjustment$k_eval
  
  if(method == "luxburg"){
    k_m <- which(indices_adj == min(indices_adj, na.rm = TRUE))
  }else{
    k_m <- which(indices_adj == max(indices_adj, na.rm = TRUE))
  }
  if(length(k_m)>1){k_m <- sample(k_m,1)}
  k_opt_adj <- as.integer(names(indices_adj[k_m]))
  index_opt_adj <- indices_adj[k_m]
  
  
  output <- list("k_opt" = k_opt, "index_opt" = index_opt, "indices" = indices,
                 "k_opt_adj" = k_opt_adj, "index_opt_adj" = index_opt_adj, "indices_adj" = indices_adj,
                 "kp_obj" = trace_kp[[k_m]]$object)
  
  return(output)
}
# # not run:
# # generate toy data with factors and numerics
# n   <- 10
# prb <- 0.99
# muk <- 2.5
# 
# x1 <- sample(c("A","B"), 2*n, replace = TRUE, prob = c(prb, 1-prb))
# x1 <- c(x1, sample(c("A","B"), 2*n, replace = TRUE, prob = c(1-prb, prb)))
# x1 <- as.factor(x1)
# x2 <- sample(c("A","B"), 2*n, replace = TRUE, prob = c(prb, 1-prb))
# x2 <- c(x2, sample(c("A","B"), 2*n, replace = TRUE, prob = c(1-prb, prb)))
# x2 <- as.factor(x2)
# x3 <- c(rnorm(n, mean = -muk), rnorm(n, mean = muk), rnorm(n, mean = -muk), rnorm(n, mean = muk))
# x4 <- c(rnorm(n, mean = -muk), rnorm(n, mean = muk), rnorm(n, mean = -muk), rnorm(n, mean = muk))
# x <- data.frame(x1,x2,x3,x4)
# 
# stability_det_k(data = x, k = 2:6, B = 100, method = "fowlkesmallows")





#' @title Determination the stability of k Prototypes Clustering
#'
#' @description Calculating the stability for a k-Prototypes clustering with k clusters or computing the stability-based optimal number of clusters for k-Prototype clustering. Possible stability indices are: \code{Jaccard}, \code{Rand}, \code{Fowlkes \& Mallows} and \code{Luxburg}.
#' 
#' @param object Object of class \code{kproto} resulting from a call with \code{kproto(..., keep.data=TRUE)}
#' @param method character specifying the stability, either one or more of \code{luxburg}, \code{fowlkesmallows}, \code{rand} or/and \code{jaccard}.
#' @param B numeric, number of bootstrap samples
#' @param verbose Logical whether information about the bootstrap procedure should be given.
#' @param ... Further arguments passed to \code{\link[clustMixType]{kproto}}, like:
#'   \itemize{
#'     \item \code{nstart}: If > 1 repetitive computations of \code{kproto} with random initial prototypes are computed.
#'     \item \code{lambda}: Factor to trade off between Euclidean distance of numeric variables and simple matching coefficient between categorical variables.
#'   }
#' 
#' @return The output contains the stability for a given k-Prototype clustering in a list with two elements:
#' @return \item{kp_stab}{stability values for the given clustering}
#' @return \item{kp_bts_stab}{stability values for each bootstrap samples}
#'
#' @author Rabea Aschenbruck
#' 
#' @references \itemize{
#'     \item Aschenbruck, R., Szepannek, G., Wilhelm, A.F.X (2023): 
#'     Stability of mixed-type cluster partitions for determination of the number of clusters. 
#'     \emph{Submitted}.
#'     
#'     \item von Luxburg, U. (2010): 
#'      Clustering stability: an overview. 
#'     \emph{Foundations and Trends in Machine Learning, Vol 2, Issue 3}.
#'     \doi{10.1561/2200000008}.
#'     
#'     \item Ben-Hur, A., Elisseeff, A., Guyon, I. (2002): 
#'     A stability based method for discovering structure in clustered data. 
#'     \emph{Pacific Symposium on Biocomputing}.
#'     \doi{10/bhfxmf}.
#'   }
#'
#' @examples
#' \dontrun{
#' # generate toy data with factors and numerics
#' n   <- 10
#' prb <- 0.99
#' muk <- 2.5 
#' 
#' x1 <- sample(c("A","B"), 2*n, replace = TRUE, prob = c(prb, 1-prb))
#' x1 <- c(x1, sample(c("A","B"), 2*n, replace = TRUE, prob = c(1-prb, prb)))
#' x1 <- as.factor(x1)
#' x2 <- sample(c("A","B"), 2*n, replace = TRUE, prob = c(prb, 1-prb))
#' x2 <- c(x2, sample(c("A","B"), 2*n, replace = TRUE, prob = c(1-prb, prb)))
#' x2 <- as.factor(x2)
#' x3 <- c(rnorm(n, mean = -muk), rnorm(n, mean = muk), rnorm(n, mean = -muk), rnorm(n, mean = muk))
#' x4 <- c(rnorm(n, mean = -muk), rnorm(n, mean = muk), rnorm(n, mean = -muk), rnorm(n, mean = muk))
#' x <- data.frame(x1,x2,x3,x4)
#' 
#' #' # apply k-prototypes
#' kpres <- kproto(x, 4, keep.data = TRUE)
#' 
#' # calculate cluster stability
#' stab <- stability_kproto(method = c("luxburg","fowlkesmallows"), object = kpres)
#' 
#' }
#' 
#' @rdname stability_kproto
#' 
#' @importFrom stats na.omit
#' @importFrom combinat permn
#' @importFrom dplyr %>% 
#' @importFrom dplyr arrange
#' @importFrom dplyr mutate
#' @importFrom dplyr lag
#' @importFrom dplyr lead
#' @importFrom rlang .data
#' 
#' 
#' @export


stability_kproto <- function(object, 
                             method = c("rand", "jaccard", "luxburg", "fowlkesmallows"), B = 100, 
                             verbose=FALSE, ...){
  
  if(!inherits(object, "kproto")) stop("object must be type of 'kproto'")
  if(!is.numeric(B)) stop("B has to be numeric, e. g. B = 100")
  
  if(is.null(object$data)) stop("object must contain clustered data (keep.data = TRUE)!")
  n <- nrow(object$data)
  
  if(is.character(method)){
    if(!any(!(method %in% c("luxburg", "fowlkesmallows", "rand", "jaccard")))){
      
      ### usage of bootstrap samples
      # object for the stability value for the whole cluster partition with respect to bts_res
      kp_bts_stab <- matrix(numeric(), ncol = length(method)+1, nrow = B)
      colnames(kp_bts_stab) <- c("B", "L", "FM", "R", "J")[c(TRUE, c("luxburg", "fowlkesmallows", "rand", "jaccard") %in% method)]
      kp_bts_stab[,"B"] <- 1:B
      
    }else{
      stop("method must only contain one or more of: luxburg, fowlkesmallows, rand or/and jaccard")
    }
  }else{
    stop("method must only contain character(s)!")
  }
  
  
  
  for(b in 1:B){ 
    # b <- 1
    if (verbose) cat(paste("boot ",b,"\n", sep = ""))
    
    # generate bootstrap samples:
    # AnmR.: which method? now: same approach as in fpc.default
    #        UPDATE: all samples data will be clustered and afterwards the subset is reduced for unique data
    bsamp <- sample(n, n, replace = TRUE)
    
    # cluster of bootstrap sample
    bts_res <- kproto(x = object$data[bsamp, ], k = length(object$size), verbose=FALSE, ...) 
    
    ### NEW: choose unique data
    # which objects are in the bootstrap sample and should be compared with object?
    ind_compare <- as.character(unique(bsamp))
    
    # cluster assignments for object and bts_res only with same subsample
    kp_cl <- object$cluster[unique(bsamp)]
    bts_cl <- bts_res$cluster[ind_compare]
    
    
    # Luxburg (2009) - "not normalized stability:
    if("luxburg" %in% method){
      # create all permutations of cluster group names and count minimum number of differences
      # short minimal example: after <- c(3,4,1,2); after[match(c(2,3,3,4,1), c(1,2,3,4))]
      kp_bts_stab[b,"L"] <- min(unlist(lapply(combinat::permn(unique(bts_cl)), 
                                                    FUN = function(x) sum(kp_cl != x[match(bts_cl, sort(unique(bts_cl)))]))))/length(ind_compare)
    }
    
    if(any(c("fowlkesmallows", "rand", "jaccard") %in% method)){
      # determination of number of pairs, which are in the same cluster
      L_kp <- sapply(kp_cl, FUN = function(x) as.numeric(x == kp_cl)) - diag(length(ind_compare))
      L_bts <- sapply(bts_cl, FUN = function(x) as.numeric(x == bts_cl)) - diag(length(ind_compare))
      
      if("fowlkesmallows" %in% method){
        # eqn. 3: similarity measure introduced by Fowlkes and Mallows:
        kp_bts_stab[b,"FM"] <- sum(L_kp * L_bts) / sqrt(sum(L_kp * L_kp) * sum(L_bts * L_bts) )
      }
      
      if("rand" %in% method){
        # eqn. 4: matching coefficient aka Rand:
        kp_bts_stab[b,"R"] <- 1 - 1/length(ind_compare)^2 * (sum(abs(L_kp - L_bts)^2)^(1/2))^2
      }

      if("jaccard" %in% method){
        # eqn. 5: jaccard coefficient
        kp_bts_stab[b,"J"] <- sum(L_kp * L_bts)/(sum(L_kp * L_kp) + sum(L_bts * L_bts) - sum(L_kp * L_bts))
      }
    }
    
  }
  
  # aggregation of stability values with respect to bootstrap samples
  #   to one value for object
  kp_stab <- colMeans(kp_bts_stab[,-1, drop=FALSE])
  
  return(list(kp_stab, kp_bts_stab))
}

###############################################
# For their feedback and suggestions thanks to:
# - Bridget Ssendagala 
# - Ben Feng 
# - Galen Terziysky
###############################################

#' @title k-Prototypes Clustering
#' @description Computes k-prototypes clustering for mixed-type data.
#' 
#' @details Like k-means, the k-prototypes algorithm iteratively recomputes cluster prototypes and reassigns 
#' clusters, whereby with \code{type = "huang"} clusters are assigned using the distance
#' \eqn{d(x,y) =  d_{euclid}(x,y) + \lambda d_{simple\,matching}(x,y)}. Cluster prototypes are computed as 
#' cluster means for numeric variables and modes for factors (cf. Huang, 1998). Ordered factors variables 
#' are treated as categorical variables.\cr
#' For \code{type = "gower"} range-normalized absolute distances from the cluster median are computed for 
#' the numeric variables (and for the ranks of the ordered factors respectively). For factors simple matching 
#' distance is used as in the original k prototypes algorithm. The prototypes are given by the median for 
#' numeric variables, the mode for factors and the level with the closest rank to the median rank of the 
#' corresponding cluster (cf. Szepannek et al., 2024).\cr
#' In case of \code{na.rm = FALSE}: for each observation variables with missings are ignored (i.e. only the 
#' remaining variables are considered for distance computation). In consequence for observations with missings 
#' this might result in a change of variable's weighting compared to the one specified by \code{lambda}. For 
#' these observations distances to the prototypes will typically be smaller as they are based on fewer variables.\cr
#' The \code{type} argument also accepts input \code{"standard"}, but this naming convention is deprecated and 
#' has been renamed to \code{"huang"}. Please use \code{"huang"} instead.
#' 
#' @keywords classif 
#' @keywords cluster
#' @keywords multivariate
#' @importFrom stats median
#' @importFrom stats complete.cases
#' 
#' 
#' @rdname kproto
#' @export 
kproto <- function (x, ...) 
  UseMethod("kproto")

#' @aliases kproto kproto.default 
#' 
#' @param x Data frame with both numerics and factors (also ordered factors are possible).
#' @param k Either the number of clusters, a vector specifying indices of initial prototypes, or a data frame of 
#' prototypes of the same columns as \code{x}.
#' @param lambda Parameter > 0 to trade off between Euclidean distance of numeric variables and simple matching 
#' coefficient between categorical variables (if \code{type = "huang"}). Also a vector of variable specific factors 
#' is possible where the order must correspond to the order of the variables in the data. In this case all variables' 
#' distances will be multiplied by their corresponding lambda value.
#' @param type Character, to specify the distance for clustering. Either \code{"huang"} or \code{"gower"} (cf. details 
#' below).
#' @param iter.max Numeric; maximum number of iterations if no convergence before.
#' @param nstart Numeric; If > 1 repetitive computations with random initializations are computed and the result with 
#' minimum \code{tot.dist} is returned.
#' @param na.rm Character, either \code{"yes"} to strip \code{NA} values for complete case analysis, \code{"no"} to 
#' keep and ignore \code{NA} values, \code{"imp.internal"} to impute the \code{NAs} within the algorithm or 
#' \code{"imp.onestep"} to apply the algorithm ignoring the \code{NAs} and impute them after the partition is determined.
#' @param keep.data Logical, whether original should be included in the returned object.
#' @param verbose Logical, whether additional information about process should be printed. 
#' Caution: For \code{verbose=FALSE}, if the number of clusters is reduced during the iterations it will not mentioned.
#' @param init Character, to specify the initialization strategy. Either \code{"nbh.dens"}, \code{"sel.cen"} or 
#' \code{"nstart.m"}. Default is \code{"NULL"}, which results in nstart repetitive algorithm computations with random 
#' starting prototypes. Otherwise, \code{nstart} is not used. Argument \code{k} must be a number if a specific 
#' initialization strategy is choosen!
#' @param p_nstart.m Numeric, probability(=0.9 is default) for \code{init="nstart.m"}, where the strategy assures 
#' that with a probability of \code{p_nstart.m} at least one of the m sets of initial prototypes contains objects 
#' of every cluster group (cf. Aschenbruck et al. (2023): Random-based Initialization for clustering mixed-type data 
#' with the k-Prototypes algorithm. In: {\emph{Cladag 2023 Book of abstracts and short spapers}}, isbn: 9788891935632.).  
#' @param \dots Currently not used.
#
#' @return \code{\link{kmeans}} like object of class \code{kproto}:
#' @return \item{cluster}{Vector of cluster memberships.}
#' @return \item{centers}{Data frame of cluster prototypes.}
#' @return \item{lambda}{Distance parameter lambda.}
#' @return \item{size}{Vector of cluster sizes.}
#' @return \item{withinss}{Vector of within cluster distances for each cluster, i.e. summed distances of all observations belonging to a cluster to their respective prototype.}
#' @return \item{tot.withinss}{Target function: sum of all observations' distances to their corresponding cluster prototype.}
#' @return \item{dists}{Matrix with distances of observations to all cluster prototypes.}
#' @return \item{iter}{Prespecified maximum number of iterations.}
#' @return \item{trace}{List with two elements (vectors) tracing the iteration process: 
#' \code{tot.dists} and \code{moved} number of observations over all iterations.}
#' @return \item{inits}{Initial prototypes determined by specified initialization strategy, if init is either 'nbh.dens' or 'sel.cen'.}
#' @return \item{nstart.m}{only for 'init = nstart_m': determined number of randomly choosen sets.}
#' @return \item{data}{if 'keep.data = TRUE' than the original data will be added to the output list.}
#' @return \item{type}{Type argument of the function call.}
#' @return \item{stdization}{Only returned for \code{type = "gower"}: List of standardized ranks for ordinal variables 
#' and an additional element \code{num_ranges} with ranges of all numeric variables. Used by \code{\link{predict.kproto}}.}

#'   
#' @examples
#' # generate toy data with factors and numerics
#' 
#' n   <- 100
#' prb <- 0.9
#' muk <- 1.5 
#' clusid <- rep(1:4, each = n)
#' 
#' x1 <- sample(c("A","B"), 2*n, replace = TRUE, prob = c(prb, 1-prb))
#' x1 <- c(x1, sample(c("A","B"), 2*n, replace = TRUE, prob = c(1-prb, prb)))
#' x1 <- as.factor(x1)
#' 
#' x2 <- sample(c("A","B"), 2*n, replace = TRUE, prob = c(prb, 1-prb))
#' x2 <- c(x2, sample(c("A","B"), 2*n, replace = TRUE, prob = c(1-prb, prb)))
#' x2 <- as.factor(x2)
#' 
#' x3 <- c(rnorm(n, mean = -muk), rnorm(n, mean = muk), rnorm(n, mean = -muk), rnorm(n, mean = muk))
#' x4 <- c(rnorm(n, mean = -muk), rnorm(n, mean = muk), rnorm(n, mean = -muk), rnorm(n, mean = muk))
#' 
#' x <- data.frame(x1,x2,x3,x4)
#' 
#' # apply k-prototypes
#' kpres <- kproto(x, 4)
#' clprofiles(kpres, x)
#' 
#' # in real world clusters are often not as clear cut
#' # by variation of lambda the emphasize is shifted towards factor / numeric variables    
#' kpres <- kproto(x, 2)
#' clprofiles(kpres, x)
#' 
#' kpres <- kproto(x, 2, lambda = 0.1)
#' clprofiles(kpres, x)
#' 
#' kpres <- kproto(x, 2, lambda = 25)
#' clprofiles(kpres, x)
#' 
#' @author \email{gero.szepannek@@web.de}
#' 
#' @references \itemize{
#'     \item Szepannek, G. (2018): clustMixType: User-Friendly Clustering of Mixed-Type Data in R, 
#'     {\emph{The R Journal 10/2}}, 200-208, \doi{10.32614/RJ-2018-048}.
#'     \item Aschenbruck, R., Szepannek, G., Wilhelm, A. (2022): Imputation Strategies for Clustering Mixed‑Type Data with Missing Values, 
#'     {\emph{Journal of Classification}}, \doi{10.1007/s00357-022-09422-y}. 
#'     \item Szepannek, G., Aschenbruck, R., Wilhelm, A. (2024): Clustering Large Mixed-Type Data with Ordinal Variables,
#'     {\emph{Advances in Data Analysis and Classification}}, \doi{10.1007/s11634-024-00595-5}.
#'     \item Z.Huang (1998): Extensions to the k-Means Algorithm for Clustering Large Data Sets with Categorical Variables, 
#'           {\emph{Data Mining and Knowledge Discovery 2}}, 283-304.
#'   }
#' 
#' @rdname kproto
#' 
#' @importFrom stats complete.cases
#' @importFrom tibble is_tibble 
#' @importFrom stats qgeom
#' 
#' @method kproto default
#' @export 
kproto.default <- function(x, k, lambda = NULL, type = "huang", iter.max = 100, nstart = 1, na.rm = "yes", keep.data = TRUE, verbose = TRUE, init = NULL, p_nstart.m = 0.9, ...){
  
  # enable input of tibbles
  if(is_tibble(x) == TRUE){x <- as.data.frame(x)}
  
  if(nrow(x) == 1) stop("Only one observation clustering not meaningful.")
  k_input <- k # store input k for nstart > 1 as clusters can be merged 
  
  # treatment of missings
  NAcount <- apply(x, 2, function(z) sum(is.na(z)))
  if(verbose){
    cat("# NAs in variables:\n")
    print(NAcount)
  }
  if(any(NAcount == nrow(x))) stop(paste("Variable(s) have only NAs please remove them:",names(NAcount)[NAcount == nrow(x)],"!"))
  
  # Backward compatibility
  if(is.logical(na.rm)){
    if(na.rm){
      na.rm <- "yes"
    }else{
      na.rm <- "no"
    }
    message("Logical input for na.rm is deprecated. Please use either 'yes','no','imp.internal' or 'imp.onestep'.\n")
  }
  if(na.rm %in% c("yes","no","imp.internal", "imp.onestep")){
    if(na.rm == "yes") {
      miss <- apply(x, 1, function(z) any(is.na(z)))
      if(verbose){
        cat(sum(miss), "observation(s) with NAs.\n")
        if(sum(miss) > 0) message("Observations with NAs are removed.\n")
        cat("\n")
      } 
      x <- x[!miss,]
    } # remove missings
    
    if(na.rm != "yes"){
      allNAs <- apply(x,1,function(z) all(is.na(z)))
      if(sum(allNAs) > 0){
        if(verbose) cat(sum(allNAs), "observation(s) where all variables NA.\n")
        warning("No meaningful cluster assignment possible for observations where all variables NA.\n")
        if(verbose) cat("\n")
        
      }
    }
  }else{
    stop("Argument na.rm must be either 'yes','no','imp.internal' or 'imp.onestep'!")
  }
  
  
  if(type %in% c("Huang", "standard")) type <- "huang"
  if(type == "Gower") type <- "gower"
  if(!type %in% c("huang", "gower")) stop("Argument type must be either 'huang' or 'gower'!")
  if(type == "huang"){
    
    # save origin data, before starting the imputation process
    if(na.rm == "imp.internal"){origin <- x}
    
    # initial error checks
    if(!is.data.frame(x)) stop("x should be a data frame!")
    if(ncol(x) < 2) stop("For clustering x should contain at least two variables!")
    if(iter.max < 1 | nstart < 1) stop("iter.max and nstart must not be specified < 1!")
    if(!is.null(lambda)){
      if(any(lambda < 0)) stop("lambda must be specified >= 0!")
      if(!any(lambda > 0)) stop("lambda must be specified > 0 for at least two variable!")
    }
    # check for numeric and factor variables
    numvars <- sapply(x, is.numeric)
    anynum <- any(numvars)
    catvars <- sapply(x, is.factor)
    anyfact <- any(catvars)
    if(!anynum) stop("\n No numeric variables in x! Try using kmodes() from package klaR...\n\n")
    if(!anyfact) stop("\n No factor variables in x! Try using kmeans()...\n\n")
    
    # initialize prototypes
    if(!is.null(init)){ # determine prototypes with selected initialization strategy
      if(length(k) == 1){
        if(as.integer(k) != k){k <- as.integer(k); warning(paste("k has been set to", k,"!"))}
      }else{
        stop("Argument k must be a number if a specific initialization strategy is choosen!")
      }
      if(init == "nbh.dens"){
        k <- protos_init <- f_nbh_dens(dat = x, k = k)
      }else{
        if(init == "sel.cen"){
          k <- protos_init <- f_sel_cen(dat = x, k = k)
        }else{
          if(init == "nstart.m"){
            N <- sum(complete.cases(x))

            prob <- 1
            for(i in 0:(k-1)){prob <- prob * (N - i * N/k)/(N - i)}
            
            # pgeom describes x number of failures and x+1 number of failures until 1 success.. => therefore: qgeom + 1
            nstart <- init_nstart.m <- qgeom(p = p_nstart.m, prob = prob)+1
          }else{
            stop("Argument 'init' must be either 'nbh.dens','sel.cen','nstart.m' or NULL!")
          }
        }
      }
        
    }
    if(!is.data.frame(k)){
      if(length(k) == 1){
        if(as.integer(k) != k){k <- as.integer(k); warning(paste("k has been set to", k,"!"))}
        if(sum(complete.cases(x)) < k) stop("Data frame has less complete observations than clusters!")
        ids <- sample(row.names(x[complete.cases(x),]), k)
        protos <- x[ids,]
      }
      if(length(k) > 1){
        if(nrow(x) < length(k)) stop("Data frame has less observations than clusters!")
        ids <- k
        k <- length(ids)
        if(length(unique(ids)) != length(ids)) stop("If k is specified as a vector it should contain different indices!")
        if(any(ids<1)|any(ids>nrow(x))) stop("If k is specified as a vector all elements must be valid indices of x!")
        #check for integer
        protos <- x[ids,]
        if(any(!complete.cases(protos))) stop("Choose initial prototypes without missing values!")
      }
      rm(ids)
    }
    if(is.data.frame(k)){
      if(nrow(x) < nrow(k)) stop("Data frame has less observations than clusters!")
      if(length(names(k)) != length(names(x))) stop("k and x have different numbers of columns!")
      if(any(names(k) != names(x))) stop("k and x have different column names!")
      if(anynum) {if( any(sapply(k, is.numeric) != numvars)) stop("Numeric variables of k and x do not match!")}
      if(anyfact) {if( any(sapply(k, is.factor) != catvars)) stop("Factor variables of k and x do not match!")}
      protos <- k
      if(any(!complete.cases(protos))) stop("Prototypes with missing values. Choose initial prototypes without missing values!")
      k <- nrow(protos)
    }
    if(k < 1) stop("Number of clusters k must not be smaller than 1!")
    
    # automatic calculation of lambda
    if(length(lambda) > 1){
      if(length(lambda) != sum(c(numvars,catvars))) stop("If lambda is a vector, its length should be the sum of numeric and factor variables in the data frame!")
      # warning for variable selection via lambda (which results in no numvars or no catvars)
      if(all(!as.logical(numvars*lambda))) warning("As a result of the choice of lambda: No numeric variables in x! Better try using kmodes() from package klaR...\n")
      if(all(!as.logical(catvars*lambda))) warning("As a result of the choice of lambda: No factor variables in x! Better try using kmeans()...\n")
    }else{
      if(length(lambda) == 1) {if(lambda == 0) stop("lambda has to be a value != 0. For automatic calculation use lambda = NULL (default setting)!")}
    }
    if(is.null(lambda)){
      lambda <- lambdaest(x = x, num.method = 1, fac.method = 1, outtype = "numeric", verbose = FALSE)
      if(verbose) cat("Estimated lambda:", lambda, "\n\n")
    }
    
    # initialize clusters
    clusters  <- numeric(nrow(x)) 
    tot.dists <- NULL
    moved   <- NULL
    iter <- 1
    
    # check for any equal prototypes and reduce cluster number in case of occurence
    if(k > 1){
      keep.protos <- rep(TRUE,k)
      for(l in 1:(k-1)){
        for(m in (l+1):k){
          d1 <- sum((protos[l,numvars, drop = FALSE]-protos[m,numvars, drop = FALSE])^2) # euclidean for numerics
          d2 <- sum(protos[l,catvars, drop = FALSE] != protos[m,catvars, drop = FALSE]) # wtd simple matching for categorics 
          if((d1+d2) == 0) keep.protos[m] <- FALSE 
        }
      }
      if(!all(keep.protos)){
        protos <- protos[keep.protos,]
        k <- sum(keep.protos)
        if(verbose) message("Equal prototypes merged. Cluster number reduced to:", k, "\n\n")      
      }
    }
    
    # special case only one cluster
    if(k == 1){clusters <- rep(1, nrow(x)); size  <- table(clusters); iter <- iter.max} # REM: named vector size is needed later...
    
    # start iterations for standard case (i.e. k > 1)
    while(iter < iter.max){
      
      # compute distances 
      nrows <- nrow(x)
      dists <- matrix(NA, nrow=nrows, ncol = k)
      for(i in 1:k){
        #a0 <- proc.time()[3]      
        #d1 <- apply(x[,numvars],1, function(z) sum((z-protos[i,numvars])^2)) # euclidean for numerics
        d1 <- (x[,numvars, drop = FALSE] - matrix(rep(as.numeric(protos[i, numvars, drop = FALSE]), nrows), nrow=nrows, byrow=T))^2
        d1[is.na(d1)] <- 0
        if(length(lambda) == 1) d1 <- rowSums(d1, na.rm = TRUE)
        if(length(lambda) > 1) d1 <- as.matrix(d1) %*% lambda[numvars]
        #a1 <- proc.time()[3]      
        #d2 <- lambda * apply(x[,catvars],1, function(z) sum((z != protos[i,catvars]))) # wtd simple matching for categorics 
        d2 <- sapply(which(catvars), function(j) return(x[,j] != rep(protos[i,j], nrows)) )
        d2[is.na(d2)] <- FALSE
        if(length(lambda) == 1) d2 <- lambda * rowSums(d2)
        if(length(lambda) > 1) d2 <- as.matrix(d2) %*% lambda[catvars]
        #a2 <- proc.time()[3]      
        dists[,i] <- d1 + d2
        #cat(a1-a0, a2-a1, "\n")
      }
      
      # assign clusters 
      old.clusters  <- clusters
      # clusters      <- apply(dists, 1, function(z) which.min(z))
      clusters      <- apply(dists, 1, function(z) {a <- which(z == min(z)); if (length(a)>1) a <- sample(a,1); return(a)}) # sample in case of multiple minima
      size          <- table(clusters)  
      min.dists     <- apply(cbind(clusters, dists), 1, function(z) z[z[1]+1])
      within        <- as.numeric(by(min.dists, clusters, sum))
      tot.within    <- sum(within)
      # prevent from empty classes
      #tot.within    <- numeric(k)
      #totw.list     <- by(min.dists, clusters, sum) 
      #tot.within[names(totw.list)] <- as.numeric(totw.list)
      
      # ...check for empty clusters and eventually reduce number of prototypes    
      if (length(size) < k){
        k <- length(size)
        protos <- protos[1:length(size),]  
        if(verbose) cat("Empty clusters occur. Cluster number reduced to:", k, "\n\n")
      }
      
      # trace
      tot.dists <- c(tot.dists, sum(tot.within))      
      moved <- c(moved, sum(clusters != old.clusters))
      
      # compute new prototypes
      remids <- as.integer(names(size))
      for(i in remids){
        some_vals <- sapply(x[clusters == i, , drop = FALSE], function(z) !all(is.na(z))) # only update variables if not all values are NA
        if(any(some_vals & numvars)){
          protos[which(remids == i), some_vals & numvars] <- sapply(x[clusters == i, some_vals & numvars, drop = FALSE], mean, na.rm = TRUE)
        }
        if(any(some_vals & catvars)){
          protos[which(remids == i), some_vals & catvars] <- sapply(x[clusters == i, some_vals & catvars, drop = FALSE], function(z) levels(z)[which.max(table(z))])
        }
      }
      
      # update missing values, if NAs should be imputed during the algorithm
      if(na.rm == "imp.internal"){
        x_old <- x
        x <- origin
        
        if(any(is.na(x[,numvars]))){
          x[,numvars][is.na(x[,numvars])] <- protos[clusters,numvars][is.na(x[,numvars])]
        }
        if(any(is.na(x[,catvars]))){
          x[,catvars][is.na(x[,catvars])] <- protos[clusters,catvars][is.na(x[,catvars])]
        }
        
        # observation with all values NA won't be imputed
        x[allNAs,] <- NA
      }
      
      if(k == 1){clusters <- rep(1, length(clusters)); size <- table(clusters); iter <- iter.max; break}
      
      # check for any equal prototypes and reduce cluster number in case of occurence
      if(iter == (iter.max-1)){ # REM: for last iteration equal prototypes are allowed. otherwise less prototypes than assigned clusters.
        keep.protos <- rep(TRUE,k)
        for(l in 1:(k-1)){
          for(m in (l+1):k){
            d1 <- sum((protos[l,numvars, drop = FALSE]-protos[m,numvars, drop = FALSE])^2) # euclidean for numerics
            d2 <- sum(protos[l,catvars, drop = FALSE] != protos[m,catvars, drop = FALSE]) # wtd simple matching for categorics 
            if((d1+d2) == 0) keep.protos[m] <- FALSE 
          }
        }
        if(!all(keep.protos)){
          protos <- protos[keep.protos,]
          k <- sum(keep.protos)
          if(verbose) cat("Equal prototypes merged. Cluster number reduced to:", k, "\n\n")      
        }
      }
      
      # add stopping rules
      if(moved[length(moved)] ==  0){
        if(na.rm != "imp.internal"){
          break
        }else{
          if(sum(!(x == x_old), na.rm = TRUE) == 0){
            break
          }
        }
      }
      
      if(k == 1){clusters <- rep(1, length(clusters)); size <- table(clusters); iter <- iter.max; break}
      
      #cat("iter", iter, "moved", moved[length(moved)], "tot.dists",tot.dists[length(tot.dists)],"\n" )      
      iter <- iter+1
    }
    
    
    ### Final update of prototypes and dists
    if(iter == iter.max){ # otherwise there have been no moves anymore and prototypes correspond to cluster assignments 
      # compute new prototypes
      remids <- as.integer(names(size))
      for(i in remids){
        some_vals <- sapply(x[clusters == i, , drop = FALSE], function(z) !all(is.na(z))) # only update variables if not all values are NA
        if(any(some_vals & numvars)){
          protos[which(remids == i), some_vals & numvars] <- sapply(x[clusters == i, some_vals & numvars, drop = FALSE], mean, na.rm = TRUE)
        }
        if(any(some_vals & catvars)){
          protos[which(remids == i), some_vals & catvars] <- sapply(x[clusters == i, some_vals & catvars, drop = FALSE], function(z) levels(z)[which.max(table(z))])
        }
      }
      
      # compute distances 
      nrows <- nrow(x)
      dists <- matrix(NA, nrow=nrows, ncol = k)
      for(i in 1:k){
        d1 <- (x[,numvars, drop = FALSE] - matrix(rep(as.numeric(protos[i, numvars, drop = FALSE]), nrows), nrow=nrows, byrow=T))^2
        d1[is.na(d1)] <- 0
        if(length(lambda) == 1) d1 <- rowSums(d1, na.rm = TRUE)
        if(length(lambda) > 1) d1 <- as.matrix(d1) %*% lambda[numvars]
        d2 <- sapply(which(catvars), function(j) return(x[,j] != rep(protos[i,j], nrows)) )
        d2[is.na(d2)] <- FALSE
        if(length(lambda) == 1) d2 <- lambda * rowSums(d2)
        if(length(lambda) > 1) d2 <- as.matrix(d2) %*% lambda[catvars]
        dists[,i] <- d1 + d2
      }
      
      size          <- table(clusters)  
      min.dists     <- apply(cbind(clusters, dists), 1, function(z) z[z[1]+1])
      within        <- as.numeric(by(min.dists, clusters, sum))
      tot.within    <- sum(within)
    }
    
    # observations with all NA are not assigned to a cluster
    if(na.rm != "yes"){
      if(sum(allNAs) > 0){
        clusters[allNAs] <- NA
        dists[allNAs,] <- NA
      }
    }
    
    names(clusters) <- row.names(dists) <- row.names(x)
    rownames(protos) <- NULL
    # create result: 
    res <- list(cluster = clusters,  
                centers = protos, 
                lambda = lambda, 
                size = size,
                withinss = within,
                tot.withinss = tot.within,   
                dists = dists, 
                iter = iter, 
                trace = list(tot.dists = tot.dists, moved = moved))
    
    # add the determined initial prototypes
    if(!is.null(init)){
      if(init %in% c("nbh.dens", "sel.cen")){res[["inits"]] <- protos_init}
    }
    
  }
  
  if(type == "gower"){
    if(na.rm %in% c("imp.internal")){
      stop("Argument na.rm must be either 'yes','no' or 'imp-onestep', since imp.internal is not yet implemented for type = 'gower'!")
    }
    res <- kproto_gower(x=x, k=k_input, lambda = lambda, iter.max = iter.max, verbose=verbose, na.rm = na.rm)
  }
  
  # if keep.data == TRUE add data to list before nstart-loop otherwise the first determined x (inclusive imputed values)
  # is added and not necessarily the data of the x with the "best" imputed values (of clusterpartition with smallest tot.withinss)
  if(keep.data) res$data = x
  
  # loop: if nstart > 1:
  if(nstart > 1)
    for(j in 2:nstart){
      # start function kproto with the origin data in case of imputation
      if(na.rm == "imp.internal"){
        x <- origin
      }
      res.new <- kproto(x=x, k=k_input, lambda = lambda,  type = type, iter.max = iter.max, nstart=1, verbose=verbose, na.rm = na.rm, keep.data = keep.data)
      if(!is.null(init)){
        if(init == "nstart.m"){res.new[["nstart.m"]] <- init_nstart.m}
      } 
      if(res.new$tot.withinss < res$tot.withinss) res <- res.new
    }  
  
  if(na.rm == "imp.onestep"){
    x <- res$data
    
    if(type == "huang"){
      if(any(is.na(x[,numvars]))){
        x[,numvars][is.na(x[,numvars])] <- protos[clusters,numvars][is.na(x[,numvars])]
      }
      if(any(is.na(x[,catvars]))){
        x[,catvars][is.na(x[,catvars])] <- protos[clusters,catvars][is.na(x[,catvars])]
      }
    }else{#type=gower
      numvars <- sapply(x, is.numeric)
      ordvars <- sapply(x, is.ordered)
      catvars <- sapply(x, is.factor) & !ordvars
      
      if(any(is.na(x[,numvars]))){
        x[,numvars][is.na(x[,numvars])] <- protos[clusters,numvars][is.na(x[,numvars])]
      }
      if(any(is.na(x[,ordvars]))){
        x[,ordvars][is.na(x[,ordvars])] <- protos[clusters,ordvars][is.na(x[,ordvars])]
      }
      if(any(is.na(x[,catvars]))){
        x[,catvars][is.na(x[,catvars])] <- protos[clusters,catvars][is.na(x[,catvars])]
      }
    }
    
    res$data <- x
  }
  
  res$type <- type
  class(res) <- "kproto"
  return(res)
}



### determination of distances, required for initialization functions
dists_kproto <- function(x, y = NULL, lambda = NULL, verbose = FALSE){
  
  if(is.null(y)){
    dat_part1 <- x[rep(c(1:nrow(x)), each = nrow(x)),]
    dat_part2 <- x[rep(c(1:nrow(x)), times = nrow(x)),]
  }else{
    if(nrow(x) == 1 & nrow(y) == 1){
      return(cbind(x, y, dist = dist_kproto(x,y,lambda = lambda)))
    }else{
      dat_part1 <- x[rep(c(1:nrow(x)), each = nrow(y)),]
      dat_part2 <- y[rep(c(1:nrow(y)), times = nrow(x)),]
    }
  }
  
  # check for numeric and factor variables
  numvars <- sapply(x, is.numeric)
  anynum <- any(numvars)
  catvars <- sapply(x, is.factor)
  anyfact <- any(catvars)
  
  # determination of lambda
  if(length(lambda) > 1) {if(length(lambda) != sum(c(numvars,catvars))) stop("If lambda is a vector, its length should be the sum of numeric and factor variables in the data frame!")}
  if(is.null(lambda)){
    if(anynum & anyfact){
      vnum <- mean(sapply(x[,numvars, drop = FALSE], var, na.rm = TRUE))
      vcat <- mean(sapply(x[,catvars, drop = FALSE], function(z) return(1-sum((table(z)/sum(!is.na(z)))^2))))
      if (vnum == 0){
        if(verbose) warning("All numerical variables have zero variance.")
        anynum <- FALSE
      } 
      if (vcat == 0){
        if(verbose) warning("All categorical variables have zero variance.")
        anyfact <- FALSE
      } 
      if(anynum & anyfact){
        lambda <- vnum/vcat
        if(verbose) cat("Estimated lambda:", lambda, "\n\n")
      }else{
        lambda <- 1
      }
    }
  }
  
  # compute distances 
  nrows <- nrow(x)
  d1 <- (dat_part1[,numvars, drop = FALSE] - dat_part2[,numvars, drop = FALSE])^2
  if(length(lambda) == 1) d1 <- rowSums(d1, na.rm = TRUE)
  if(length(lambda) > 1) d1 <- as.matrix(d1) %*% lambda[numvars]
  d2 <- sapply(which(catvars), function(j) return(dat_part1[,j] != dat_part2[,j]))
  d2[is.na(d2)] <- FALSE
  if(length(lambda) == 1) d2 <- lambda * rowSums(d2)
  if(length(lambda) > 1) d2 <- as.matrix(d2) %*% lambda[catvars]
  
  return(cbind(dat_part1, dat_part2, dist = as.vector(d1 + d2)))
}



dist_kproto <- function(x, y, lambda, verbose = FALSE){
  
  x <- rbind(x,y)
  
  # check for numeric and factor variables
  numvars <- sapply(x, is.numeric)
  anynum <- any(numvars)
  catvars <- sapply(x, is.factor)
  anyfact <- any(catvars)
  
  # compute distances 
  nrows <- nrow(x)
  d1 <- (x[1, numvars, drop = FALSE] - x[2, numvars, drop = FALSE])^2
  if(length(lambda) == 1) d1 <- rowSums(d1, na.rm = TRUE)
  if(length(lambda) > 1) d1 <- as.matrix(d1) %*% lambda[numvars]
  d2 <- sapply(which(catvars), function(j) return(x[1,j] != x[2,j]))
  d2[is.na(d2)] <- FALSE
  if(length(lambda) == 1) d2 <- lambda * rowSums(matrix(d2, nrow = 1))
  if(length(lambda) > 1) d2 <- as.matrix(d2) %*% lambda[catvars]
  
  return(as.numeric(d1 + d2))
}



### initial prototypes selection based on neighborhood density and distance
f_nbh_dens <- function(dat, k){
  
  # determine object to save initial prototypes
  protos_initial <- dat[0,]
  
  # determine distances between all pairs of objects:
  lambda <- lambdaest(x = dat, verbose = FALSE)
  all_dists <- dists_kproto(x = dat, lambda = lambda)
  
  # determine \bar{d}/2 (where \bar{d} is the average distance between all pairs of objects)
  nbh_radius <- (sum(all_dists$dist)/(nrow(all_dists)-nrow(dat)))/2
  
  # determine N_e, which is the sum of h() for every object 
  # Ne_nbh says the number of objects with distance smaller than nbh_radius for every object
  Ne_nbh <- numeric()
  for(i in 1:nrow(dat)){
    # noteR: -1 because the distance between X_i and X_i is included and always < nbh_radius...
    Ne_nbh[i] <- sum(all_dists[which(rowSums(mapply("==", dat[i,], all_dists[, 1:ncol(dat)])) == ncol(dat)), ]$dist <= nbh_radius) - 1
  }
  
  # determine first initial prototype
  arrange_index <- order(Ne_nbh, decreasing = TRUE)
  protos_initial[1,] <- dat[arrange_index[1],]
  
  # determination of initial prototypes, lowering L until we have enough initis found
  L <- nbh_radius*4 *2
  while(nrow(protos_initial) < k){
    # determine the other initial prototypes by comparison of objects to yet known prototypes
    L <- L/2
    
    for(i in arrange_index[-1]){
      dists <- dists_kproto(x = dat[i,, drop = FALSE], y = protos_initial, lambda = lambda)
      ## if(any(dists$dist > L)){ 
      # noteR: paper: !"...determine whether any M_i \in M has dist(Y_i,M_i) > L holds.
      #        If satisfied, then Y_i is put into the set M as next prototype..
      #        NOT MEANINGFUL, THE DISTANCE TO ALL PROTOS MUST BE > L...
      if(all(dists$dist > L)){ #the same but other code: if(min(dist%dist) > L)
        protos_initial <- rbind(protos_initial, dat[i,])
      }
    }
  }
  
  return(protos_initial[1:k,])
}



# Katsavounidis (1994), He: 
f_sel_cen <- function(dat, k){
  
  # determine object which has the highest 
  # sum of distances to all other objects as first prototypes
  lambda <- lambdaest(x = dat, verbose = FALSE)
  all_dists <- dists_kproto(x = dat, lambda = lambda)
  dists_matrix <- matrix(all_dists[,"dist"], ncol = nrow(dat), nrow = nrow(dat)) - diag(NA, nrow = nrow(dat))
  protos_initial <- dat[which.max(colSums(dists_matrix, na.rm = TRUE)),]
  
  # determine minimal distance to initial prototypes
  while(nrow(protos_initial) < k){
    Dmin <- numeric()
    for(j in 1:nrow(dat)){
      Dmin[j] <- min(dists_kproto(x = dat[j,, drop = FALSE], y = protos_initial, lambda = lambda)$dist)
    }
    
    # choose next initial prototype by highest distance to already choosen protos
    protos_initial <- rbind(protos_initial, dat[which.max(Dmin),])
    
  }
  
  return(protos_initial)
}


#' @title Assign k-Prototypes Clusters
#'
#' @description Predicts k-prototypes cluster memberships and distances for new data.
#'
#' @param object Object resulting from a call of \code{kproto}.
#' @param newdata New data frame (of same structure) where cluster memberships are to be predicted.
#' @param \dots Currently not used.
#
#' @return \code{\link{kmeans}} like object of class \code{kproto}:
#' @return \item{cluster}{Vector of cluster memberships.}
#' @return \item{dists}{Matrix with distances of observations to all cluster prototypes.}
# 
#' @examples
#' # generate toy data with factors and numerics
#' 
#' n   <- 100
#' prb <- 0.9
#' muk <- 1.5 
#' clusid <- rep(1:4, each = n)
#' 
#' x1 <- sample(c("A","B"), 2*n, replace = TRUE, prob = c(prb, 1-prb))
#' x1 <- c(x1, sample(c("A","B"), 2*n, replace = TRUE, prob = c(1-prb, prb)))
#' x1 <- as.factor(x1)
#' 
#' x2 <- sample(c("A","B"), 2*n, replace = TRUE, prob = c(prb, 1-prb))
#' x2 <- c(x2, sample(c("A","B"), 2*n, replace = TRUE, prob = c(1-prb, prb)))
#' x2 <- as.factor(x2)
#' 
#' x3 <- c(rnorm(n, mean = -muk), rnorm(n, mean = muk), rnorm(n, mean = -muk), rnorm(n, mean = muk))
#' x4 <- c(rnorm(n, mean = -muk), rnorm(n, mean = muk), rnorm(n, mean = -muk), rnorm(n, mean = muk))
#' 
#' x <- data.frame(x1,x2,x3,x4)
#' 
#' # apply k-prototyps
#' kpres <- kproto(x, 4)
#' predicted.clusters <- predict(kpres, x) 
#' 
#' 
#' @author \email{gero.szepannek@@web.de}
#' 
#' @rdname predict.kproto
#' @export
predict.kproto <-function(object, newdata, ...){
  lambda <- object$lambda
  k <- length(object$size)
  protos <- object$centers
  x <- newdata
  
  if(object$type == "huang"){
    # check for numeric and factor variables  
    numvars <- sapply(x, is.numeric)
    catvars <- sapply(x, is.factor)
    
    if(!all(names(object$centers) == names(x))) warning("Variable names of newdata do not match!")
    if(ncol(object$centers) != ncol(x)) stop("Column number of newdata does not match object!")
    if(!all(sapply(object$centers[,numvars, drop = FALSE], is.numeric))) stop("Numeric variables of object and newdata do not match!")
    if(!all(sapply(object$centers[,catvars, drop = FALSE], is.factor))) stop("Factor variables of object and newdata do not match!")
    
    # compute distances 
    dists <- matrix(NA, nrow=nrow(x), ncol = k)
    nrows <- nrow(x)
    for(i in 1:k){
      #a0 <- proc.time()[3]      
      #d1 <- apply(x[,numvars],1, function(z) sum((z-protos[i,numvars])^2)) # euclidean for numerics
      d1 <- (x[,numvars, drop = FALSE] - matrix(rep(as.numeric(protos[i,numvars, drop = FALSE]), nrows), nrow=nrows, byrow=T))^2
      if(length(lambda) == 1) d1 <- rowSums(d1, na.rm = TRUE)
      if(length(lambda) > 1) d1<- as.matrix(d1) %*% lambda[numvars]
      #a1 <- proc.time()[3]      
      #d2 <- lambda * apply(x[,catvars],1, function(z) sum((z != protos[i,catvars]))) # wtd simple matching for categorics 
      d2 <- sapply(which(catvars), function(j) return(x[,j] != rep(protos[i,j], nrows)) )
      # fix for x data frame with only one observation according to G.Terziysky
      if (nrows == 1) d2 <- matrix(data = d2, nrow = 1, byrow = TRUE, dimnames = list(NULL, names(x)[catvars])) # <- FIX
      d2[is.na(d2)] <- FALSE
      if(length(lambda) == 1) d2 <- lambda * rowSums(d2)
      if(length(lambda) > 1) d2 <- as.matrix(d2) %*% lambda[catvars]
      #a2 <- proc.time()[3]      
      dists[,i] <- d1 + d2
      #cat(a1-a0, a2-a1, "\n")
    }
  }

  if(object$type == "gower"){
    # check for numeric, ordinal and factor variables
    numvars <- sapply(x, is.numeric)
    anynum <- any(numvars)
    ordvars <- sapply(x, is.ordered)
    anyord <- any(ordvars)
    catvars <- sapply(x, is.factor) & !ordvars
    anyfact <- any(catvars)

    # normalization numeric variables to unit range from lookup table 
    if(any(numvars)){
      for(jord in which(numvars)){
        n <- names(x)[jord]
        x[,jord] <- x[,jord] / object$stdization$num_ranges[[n]]
        protos[,jord] <- protos[,jord] / object$stdization$num_ranges[[n]]
      }
    } 
    # replace ordinal levels by standardized ranks from lookup table
    if(any(ordvars)){
      for(jord in which(ordvars)){
        val <- object$stdization[[names(x)[jord]]]$value
        names(val) <- object$stdization[[names(x)[jord]]]$level
        x[,jord] <- val[x[,jord]]
        protos[,jord] <- val[protos[,jord]]
      }  
    }

    # compute distances 
    nrows <- nrow(x)
    dists <- matrix(NA, nrow=nrows, ncol = k)
    for(i in 1:k){
      
      # in case of no numeric / factor / ordinale variables set:
      d1 <- d2 <- d3 <- rep(0, nrows)
      
      if(any(numvars)){
        d1 <- abs(x[, numvars, drop = FALSE] - matrix(rep(as.numeric(protos[i, numvars, drop = FALSE]), nrows), nrow=nrows, byrow=T))
        for(jnum in 1:ncol(d1)) d1[,jnum] <- d1[,jnum] / object$stdization$num_ranges[[names(d1)[jnum]]] 
        d1[is.na(d1)] <- 0
        if(length(lambda) > 1) d1 <- as.matrix(d1) %*% lambda[numvars]
        if(is.null(lambda)) d1 <- rowSums(d1)
      }
      
      if(any(catvars)){
        d2 <- sapply(which(catvars), function(j) return(x[,j] != rep(protos[i,j], nrows)) )
        d2[is.na(d2)] <- FALSE
        if(length(lambda) > 1) d2 <- as.matrix(d2) %*% lambda[catvars]
        if(is.null(lambda)) d2 <- rowSums(d2)
      }
      
      if(any(ordvars)){
        d3 <- abs(x[, ordvars, drop = FALSE] - matrix(rep(as.numeric(protos[i, ordvars, drop = FALSE]), nrows), nrow=nrows, byrow=T))
        #for(jord in 1:ncol(d3)) d3[,jord] <- d3[,jord] / rgords[jord] # different to kproto_gower() not needed -- already standardized via exported lookup table in ln. 518ff
        d3[is.na(d3)] <- 0
        if(length(lambda) > 1) d3 <- as.matrix(d3) %*% lambda[ordvars]
        if(is.null(lambda)) d3 <- rowSums(d3)
      }
      
      dists[,i] <- d1 + d2 + d3
    }
    
  }
  
  # assign clusters
  clusters      <- apply(dists, 1, function(z) {a <- which.min(z); if (length(a)>1) a <- sample(a,1); return(a)}) # sample in case of multiple minima
  
  res <- list(cluster = clusters, dists = dists)
  return(res)
}


#' @title Profiling k-Prototypes Clustering
#'
#' @description Visualization of a k-prototypes clustering result for cluster interpretation.
#' 
#' @details For numerical variables boxplots and for factor variables barplots of each cluster are generated.
#' 
#' @param object Object resulting from a call of resulting \code{kproto}. Also other \code{kmeans} like objects with \code{object$cluster} and \code{object$size} are possible. 
#' @param x Original data.
#' @param vars Optional vector of either column indices or variable names.
#' @param col Palette of cluster colours to be used for the plots. As a default RColorBrewer's 
#' \code{brewer.pal(max(unique(object$cluster)), "Set3")} is used for k > 2 clusters and lightblue and orange else.
#'
#' @examples
#' # generate toy data with factors and numerics
#' 
#' n   <- 100
#' prb <- 0.9
#' muk <- 1.5 
#' clusid <- rep(1:4, each = n)
#' 
#' x1 <- sample(c("A","B"), 2*n, replace = TRUE, prob = c(prb, 1-prb))
#' x1 <- c(x1, sample(c("A","B"), 2*n, replace = TRUE, prob = c(1-prb, prb)))
#' x1 <- as.factor(x1)
#' 
#' x2 <- sample(c("A","B"), 2*n, replace = TRUE, prob = c(prb, 1-prb))
#' x2 <- c(x2, sample(c("A","B"), 2*n, replace = TRUE, prob = c(1-prb, prb)))
#' x2 <- as.factor(x2)
#' 
#' x3 <- c(rnorm(n, mean = -muk), rnorm(n, mean = muk), rnorm(n, mean = -muk), rnorm(n, mean = muk))
#' x4 <- c(rnorm(n, mean = -muk), rnorm(n, mean = muk), rnorm(n, mean = -muk), rnorm(n, mean = muk))
#' 
#' x <- data.frame(x1,x2,x3,x4)
#' 
#' # apply k-prototyps
#' kpres <- kproto(x, 4)
#' clprofiles(kpres, x)
#' 
#' # in real world clusters are often not as clear cut
#' # by variation of lambda the emphasize is shifted towards factor / numeric variables    
#' kpres <- kproto(x, 2)
#' clprofiles(kpres, x)

#' 
#' kpres <- kproto(x, 2, lambda = 0.1)
#' clprofiles(kpres, x)
#' 
#' kpres <- kproto(x, 2, lambda = 25)
#' clprofiles(kpres, x)

#' 
#' @author \email{gero.szepannek@@web.de}
#' 
#' @rdname clprofiles
#' @importFrom graphics barplot
#' @importFrom graphics boxplot
#' @importFrom graphics legend
#' @importFrom graphics par
#' @importFrom RColorBrewer brewer.pal
#' 
#' @export
clprofiles <- function(object, x, vars = NULL, col = NULL){
  if(length(object$cluster) != nrow(x)) stop("Size of x does not match cluster result!")
  if(is.null(vars)) vars <- 1:ncol(x)
  if(!is.numeric(vars)) vars <- sapply(vars, function(z) return(which(colnames(x)==z)))
  if(length(vars) < 1) stop("Specified variable names do not match x!")
  if(is.null(col)){
    k <- max(unique(object$cluster))
    if(k > 2)  col <- brewer.pal(k, "Set3")
    if(k == 2) col <- c("lightblue","orange")
    if(k == 1) col <- "lightblue"
  }
  #clusids <- as.numeric(names(object$size)) # object size not named for kmeans
  clusids <- sort(unique(object$cluster)) 
  if(length(col) != max(clusids)) warning("Length of col should match number of clusters!")
  
  par(ask=TRUE)
  for(i in vars){
    if(is.numeric(x[,i])){
      boxplot(x[,i]~object$cluster, col = col, main = colnames(x)[i])
      legend("topright", legend=clusids, fill = col)
    } 
    if(is.factor(x[,i])){
      tab <- table(x[,i], object$cluster)
      for(j in 1:length(object$size)) tab[,j] <- tab[,j]/object$size[j]
      barplot(t(tab), beside = TRUE, main = colnames(x)[i], col = col)
    } 
  }
  par(ask=FALSE)
  invisible()
}


#' @title Compares Variability of Variables
#'
#' @description Investigation of the variables' variances/concentrations to support specification of lambda for k-prototypes clustering.
#' 
#' @details Variance (\code{num.method = 1}) or standard deviation (\code{num.method = 2}) of numeric variables 
#' and \eqn{1-\sum_i p_i^2} (\code{fac.method = 1}) or \eqn{1-\max_i p_i} (\code{fac.method = 2}) for factors is computed.
#' 
#' @param x Data.frame with both numerics and factors.
#' @param num.method Integer 1 or 2. Specifies the heuristic used for numeric variables.
#' @param fac.method Integer 1 or 2. Specifies the heuristic used for factor variables.
#' @param outtype Specifies the desired output: either 'numeric', 'vector' or 'variation'.
#' @param verbose Logical whether additional information about process should be printed. 
#' 
#' @return \item{lambda}{Ratio of averages over all numeric/factor variables is returned. 
#' In case of \code{outtype = "vector"} the separate lambda for all variables is returned as the inverse of the single variables' 
#' variation as specified by the \code{num.method} and \code{fac.method} argument. \code{outtype = "variation"} directly returns these quantities and is not meant to be 
#' passed directly to \code{kproto()}.}
#' 
#' @examples
#' # generate toy data with factors and numerics
#' 
#' n   <- 100
#' prb <- 0.9
#' muk <- 1.5 
#' clusid <- rep(1:4, each = n)
#' 
#' x1 <- sample(c("A","B"), 2*n, replace = TRUE, prob = c(prb, 1-prb))
#' x1 <- c(x1, sample(c("A","B"), 2*n, replace = TRUE, prob = c(1-prb, prb)))
#' x1 <- as.factor(x1)
#' 
#' x2 <- sample(c("A","B"), 2*n, replace = TRUE, prob = c(prb, 1-prb))
#' x2 <- c(x2, sample(c("A","B"), 2*n, replace = TRUE, prob = c(1-prb, prb)))
#' x2 <- as.factor(x2)
#' 
#' x3 <- c(rnorm(n, mean = -muk), rnorm(n, mean = muk), rnorm(n, mean = -muk), rnorm(n, mean = muk))
#' x4 <- c(rnorm(n, mean = -muk), rnorm(n, mean = muk), rnorm(n, mean = -muk), rnorm(n, mean = muk))
#' 
#' x <- data.frame(x1,x2,x3,x4)
#' 
#' lambdaest(x)
#' res <- kproto(x, 4, lambda = lambdaest(x))
#' 
#' @author \email{gero.szepannek@@web.de}
#' 
#' @rdname lambdaest
#' 
#' @importFrom stats var
#' @importFrom stats sd
#' @importFrom tibble is_tibble 
#' @export
lambdaest <- function(x, num.method = 1, fac.method = 1, outtype = "numeric", verbose = TRUE){
  
  # enable input of tibbles
  if(is_tibble(x) == TRUE){x <- as.data.frame(x)}
  
  # initial error checks
  if(!is.data.frame(x)) stop("x should be a data frame!")
  if(nrow(x) == 1) stop("Determination for only one observation is not meaningful.")
  if(ncol(x) < 2) stop("x should contain at least two variables!")
  
  if(!num.method %in% 1:2) stop("Argument 'num.method' must be either 1 or 2!")
  if(!fac.method %in% 1:2) stop("Argument 'fac.method' must be either 1 or 2!")
  if(!outtype %in% c("numeric","vector","variation")) stop("Wrong specification of argument 'outtype'!")
    
  # check for numeric and factor variables
  numvars <- sapply(x, is.numeric)
  anynum <- any(numvars)
  catvars <- sapply(x, is.factor)
  anyfact <- any(catvars)
  if(!anynum) stop("\n No numeric variables in x! \n\n")
  if(!anyfact) stop("\n No factor variables in x! \n\n")
  
  if(anynum & num.method == 1) vnum <- sapply(x[,numvars, drop = FALSE], var, na.rm =TRUE)
  if(anynum & num.method == 2) vnum <- sapply(x[,numvars, drop = FALSE], sd, na.rm = TRUE)
  
  if(anyfact & fac.method == 1) vcat <- sapply(x[,catvars, drop = FALSE], function(z) return(1-sum((table(z)/sum(!is.na(z)))^2)))
  if(anyfact & fac.method == 2) vcat <- sapply(x[,catvars, drop = FALSE], function(z) return(1-max(table(z)/sum(!is.na(z)))))
  if (mean(vnum) == 0){
    if(verbose) warning("All numerical variables have zero variance.")
    anynum <- FALSE
  } 
  if (mean(vcat) == 0){
    if(verbose) warning("All categorical variables have zero variance.")
    anyfact <- FALSE
  } 
  
  if(verbose){
    if(num.method == 1) cat("Numeric variances:\n")
    if(num.method == 2) cat("Numeric standard deviations:\n")
    print(vnum)
    if(num.method == 1) cat("Average numeric variance:", mean(vnum), "\n\n")
    if(num.method == 2) cat("Average numeric standard deviation:", mean(vnum), "\n\n")
    
    cat(paste("Heuristic for categorical variables: (method = ",fac.method,") \n", sep = ""))
    print(vcat)
    cat("Average categorical variation:", mean(vcat), "\n\n")
  }
  
  # output depending on argument 'outtype'
  if(outtype == "numeric"){
    lambda <- mean(vnum)/mean(vcat)
    if(verbose) cat("Estimated lambda:", lambda, "\n\n")
  }
  if(outtype != "numeric"){
    lambda <- rep(0, ncol(x))
    names(lambda) <- names(x)
    lambda[numvars] <- vnum
    lambda[catvars] <- vcat
  }
  if(outtype == "vector"){
    lambda <- 1/lambda
  }
  
  return(lambda)
}



#' @export
print.kproto <- function(x, ...){
  cat("Distance type:", x$type, "\n\n")
  
  cat("Numeric predictors:", sum(sapply(x$centers, is.numeric)), "\n")
  numord <- sum(sapply(x$centers, is.ordered))
  numcat <- sum(sapply(x$centers, is.factor))
  if(x$type == "gower"){
    cat("Ordinal predictors:", numord, "\n")
    cat("Categorical predictors:", numcat - numord, "\n")
  }
  if(x$type == "huang") cat("Categorical predictors:", numcat, "\n")
  if(!is.null(x$lambda)) cat("Lambda:", x$lambda, "\n\n")
  
  cat("Number of Clusters:", length(x$size), "\n")
  cat("Cluster sizes:", x$size, "\n")
  cat("Within cluster error:", x$withinss, "\n\n")
  
  cat("Cluster prototypes:\n")
  print(x$centers)
}

#' @title Summary Method for kproto Cluster Result
#'
#' @description Investigation of variances to specify lambda for k-prototypes clustering.
#' 
#' @details For numeric variables statistics are computed for each clusters using \code{summary()}. 
#' For categorical variables distribution percentages are computed.  
#' 
#' @param object Object of class \code{kproto}.
#' @param data Optional data set to be analyzed. If \code{!(is.null(data))} clusters for \code{data} are assigned by 
#' \code{predict(object, data)}. If not specified the clusters of the original data ara analyzed which is only possible if \code{kproto} 
#' has been called using \code{keep.data = TRUE}.
#' @param pct.dig Number of digits for rounding percentages of factor variables.
#' @param \dots Further arguments to be passed to internal call of \code{summary()} for numeric variables.
#'  
#' @return List where each element corresponds to one variable. Each row of any element corresponds to one cluster.
#' 
#' @examples
#' # generate toy data with factors and numerics
#' 
#' n   <- 100
#' prb <- 0.9
#' muk <- 1.5 
#' clusid <- rep(1:4, each = n)
#' 
#' x1 <- sample(c("A","B"), 2*n, replace = TRUE, prob = c(prb, 1-prb))
#' x1 <- c(x1, sample(c("A","B"), 2*n, replace = TRUE, prob = c(1-prb, prb)))
#' x1 <- as.factor(x1)
#' 
#' x2 <- sample(c("A","B"), 2*n, replace = TRUE, prob = c(prb, 1-prb))
#' x2 <- c(x2, sample(c("A","B"), 2*n, replace = TRUE, prob = c(1-prb, prb)))
#' x2 <- as.factor(x2)
#' 
#' x3 <- c(rnorm(n, mean = -muk), rnorm(n, mean = muk), rnorm(n, mean = -muk), rnorm(n, mean = muk))
#' x4 <- c(rnorm(n, mean = -muk), rnorm(n, mean = muk), rnorm(n, mean = -muk), rnorm(n, mean = muk))
#' 
#' x <- data.frame(x1,x2,x3,x4)
#' 
#' res <- kproto(x, 4)
#' summary(res)
#' 
#' @author \email{gero.szepannek@@web.de}
#' 
#' @rdname summary
#'
#' @importFrom stats predict 
#' @export
summary.kproto <- function(object, data = NULL, pct.dig = 3, ...){
  if(!inherits(object, "kproto")) stop("object must be of class kproto!")
  
  if(is.null(data)){
    data <- object$data
    cluster <- object$cluster
  }
  if(!is.null(data)) cluster <- predict(object, data)$cluster
  
  numvars <- sapply(data, is.numeric)
  #anynum <- any(numvars)
  catvars <- sapply(data, is.factor)
  #anyfact <- any(catvars)

  res <- NULL
  for(i in 1:ncol(data)){
    cat(names(data)[i],"\n")
    if(numvars[i]){
      resi <- by(data[,i], cluster, summary, ...)
      res[[i]] <- matrix(unlist(resi), nrow = length(unique(cluster)), byrow=TRUE)
      colnames(res[[i]]) <- names(resi[[1]])
      rownames(res[[i]]) <- sort(unique(cluster))
      }
    if(catvars[i])  res[[i]] <- round(prop.table(table(cluster, data[,i]),1), digits = pct.dig)
    print(res[[i]])
    cat("\n-----------------------------------------------------------------\n")
  }
  names(res) <- names(data)  

  #return(res)
  invisible(res)
}



# optilambda <- function(x, k, upper = 3*lambdaest(x), iter.max = 100, nstart=2, keep.data = FALSE){
#   
#   foo <- function(lamb, x = x, k = k, iter.max = iter.max, nstart=nstart, keep.data = keep.data){
#     res <- kproto(x = x, k = k, lambda = lamb, iter.max = iter.max, nstart=nstart, keep.data = keep.data)  
#     numvars <- sapply(x, is.numeric)
#     catvars <- sapply(x, is.factor)
#     nrows <- nrow(x)
#     protos <- res$centers
#     
#     dists <- ndists <- cdists <- matrix(NA, nrow=nrows, ncol = nrow(res$centers)) #dists <- rep(NA, nrows)
#     for(i in 1:nrow(res$centers)){
#       d1 <- (x[,numvars, drop = FALSE] - matrix(rep(as.numeric(protos[i,numvars, drop = FALSE]), nrows), nrow=nrows, byrow=T))^2
#       d2 <- sapply(which(catvars), function(j) return(x[,j] != rep(protos[i,j], nrows)) )
#       ndists[,i] <- rowSums(d1)
#       cdists[,i] <- lamb * rowSums(d2)
#       dists[,i]  <- ndists[,i] + cdists[,i]
#       }
#     
#     #min.dists     <- apply(cbind(res$cluster, dists), 1, function(z) z[z[1]+1])
#     min.ndists     <- apply(cbind(res$cluster, ndists), 1, function(z) z[z[1]+1])
#     min.cdists     <- apply(cbind(res$cluster, cdists), 1, function(z) z[z[1]+1])
#     crit           <- abs(sum(min.ndists) - sum(min.cdists))
#     return(crit)
#     }
#   
#     foo(lamb, x = x, k = k, iter.max = iter.max, nstart=nstart, keep.data = keep.data)
#     lambda <- optimize(foo, lower = 0, upper = upper, x = x, k = k, iter.max = iter.max, nstart=nstart, keep.data = keep.data)$minimum
#     return(lambda)  
#   }
# 
# optilambda(x=x, k = 4, nstart = 5)
# res <- kproto(x = x, k = k, lambda = 9.56, iter.max = iter.max, nstart=5, keep.data = keep.data) 


#' @title Assign k-Prototypes Clusters
#'
#' @description Plot distributions of the clusters across the variables.
#'
#' @details Wrapper around \code{\link{clprofiles}}. Only works for \code{kproto} object created with \code{keep.data = TRUE}. 
#' 
#' @param x Object resulting from a call of \code{kproto}.
#' @param \dots Additional arguments to be passet to \code{\link{clprofiles}} such as e.g. \code{vars}.
#'
#' @examples
#' # generate toy data with factors and numerics
#' 
#' n   <- 100
#' prb <- 0.9
#' muk <- 1.5 
#' clusid <- rep(1:4, each = n)
#' 
#' x1 <- sample(c("A","B"), 2*n, replace = TRUE, prob = c(prb, 1-prb))
#' x1 <- c(x1, sample(c("A","B"), 2*n, replace = TRUE, prob = c(1-prb, prb)))
#' x1 <- as.factor(x1)
#' 
#' x2 <- sample(c("A","B"), 2*n, replace = TRUE, prob = c(prb, 1-prb))
#' x2 <- c(x2, sample(c("A","B"), 2*n, replace = TRUE, prob = c(1-prb, prb)))
#' x2 <- as.factor(x2)
#' 
#' x3 <- c(rnorm(n, mean = -muk), rnorm(n, mean = muk), rnorm(n, mean = -muk), rnorm(n, mean = muk))
#' x4 <- c(rnorm(n, mean = -muk), rnorm(n, mean = muk), rnorm(n, mean = -muk), rnorm(n, mean = muk))
#' 
#' x <- data.frame(x1,x2,x3,x4)
#' 
#' # apply k-prototyps
#' kpres <- kproto(x, 4)
#' plot(kpres, vars = c("x1","x3")) 
#' 
#' 
#' @author \email{gero.szepannek@@web.de}
#' 
#' @rdname plot.kproto
#' @export
plot.kproto <- function(x, ...){
  if(!any("data" %in% names(x))) stop("plot method for kproto objects only works for keep.data = TRUE. Use clprofiles() instead.")
  clprofiles(x, x$data, ...)  
}

context("Basic tests.\n")

library(clustMixType)

# generate test data from example
set.seed(42)
n   <- 100
prb <- 0.9
muk <- 1.5 
clusid <- rep(1:4, each = n)

x1 <- sample(c("A","B"), 2*n, replace = TRUE, prob = c(prb, 1-prb))
x1 <- c(x1, sample(c("A","B"), 2*n, replace = TRUE, prob = c(1-prb, prb)))
x1 <- as.factor(x1)

x2 <- sample(c("A","B"), 2*n, replace = TRUE, prob = c(prb, 1-prb))
x2 <- c(x2, sample(c("A","B"), 2*n, replace = TRUE, prob = c(1-prb, prb)))
x2 <- as.factor(x2)

x3 <- c(rnorm(n, mean = -muk), rnorm(n, mean = muk), rnorm(n, mean = -muk), rnorm(n, mean = muk))
x4 <- c(rnorm(n, mean = -muk), rnorm(n, mean = muk), rnorm(n, mean = -muk), rnorm(n, mean = muk))

x <- data.frame(x1,x2,x3,x4)


kpres <- kproto(x, 4)
pred  <- predict(kpres, x)
test_that("resulting objects are as expected",{
  expect_is(kpres, "kproto")
  expect_is(pred$cluster, "integer")
  expect_equal(length(pred$cluster), nrow(x))
  expect_is(pred$dists, "matrix")
  expect_equal(nrow(pred$dists), nrow(x))
  expect_equal(ncol(pred$dists), length(kpres$size))
  expect_true(length(kpres$size) <= 4)
  expect_equal(sort(unique(kpres$cluster)), 1:length(kpres$size))
}
)

test_that("different initializations work properly",{
  expect_is(kproto(x, c(1,101,201,301)), "kproto")
  expect_error(kproto(x, c(-1,101,201,301)))            # initialization by vector of indizes
  expect_error(kproto(x, c(1,101,201,401)))
  expect_is(kproto(x, x[c(1,101,201,301),]), "kproto")  # initialization by df
  expect_message(kproto(x, x[c(1,1,2,3,3),]),"Equal prototypes merged. Cluster number reduced to:3")
  expect_is(kproto(x, x[1,]), "kproto")                 # one single observation
  expect_is(kproto(x, x[1,])$centers, "data.frame")
  expect_equal(nrow(kproto(x, x[1,])$centers), 1)
  expect_is(kproto(x, 1), "kproto")                 # k = 1
}
)

test_that("different ways of specifying lambda work properly",{
  expect_is(kproto(x, 4, lambda = 1), "kproto")
  expect_equal(kproto(x, 4, lambda = 0.834)$lambda, 0.834)
  expect_is(kproto(x, 4, lambda = 2:5), "kproto")           # vector
  expect_equal(kproto(x, 4, lambda = 2:5)$lambda, 2:5)
  expect_error(kproto(x, 4, lambda = 1:3))
  expect_error(kproto(x, 4, lambda = -1:2))
  expect_error(kproto(x, 4, lambda = c(0,0,0,0)))
  expect_error(kproto(x, 4, lambda = 0))
  expect_is(kproto(x, 4, lambda = c(0,0,0,1)), "kproto")
  expect_is(kproto(x, 4, lambda = c(1,0,0,0)), "kproto")
  expect_is(kproto(x, 4, lambda = c(0,0,1,1)), "kproto")
  expect_is(kproto(x, 4, lambda = c(1,1,0,0)), "kproto")
}
)


test_that("lambdaest does what it is supposed to do",{
  expect_is(lambdaest(x), "numeric")
  expect_is(lambdaest(x, outtype = "vector"), "numeric")
  expect_equal(length(lambdaest(x, outtype = "vector")), ncol(x))                         # correct dimension of lambda 
  expect_is(lambdaest(x, outtype = "variation"), "numeric")
  expect_equal(length(lambdaest(x, outtype = "variation")), ncol(x))
  expect_equal(lambdaest(x, outtype = "vector"), 1/lambdaest(x, outtype = "variation"))   # compare lambdas variations 
  expect_equal(kproto(x, 2, lambda = lambdaest(x))$lambda, lambdaest(x))                                          # estimated lambda used by kproto 
  expect_equal(kproto(x, 2, lambda = lambdaest(x, outtype = "vector"))$lambda, lambdaest(x, outtype = "vector"))  # estimated lambda used by kproto  
}
)

test_that("also works for 1D",{
    expect_is(kproto(x[,c(1,3)], 4), "kproto")
    expect_error(kproto(x[,1:2], 4))
    expect_error(kproto(x[,3:4], 4))
}
)

x2 <- data.frame(V1 = factor(c("A","A","B","B")), V2 = c(-3,-1,1,3)) 
kpres <- kproto(x2, x2[c(1,3),], lambda = 1)
test_that("distances are computed correctly",{
  expect_equal(as.numeric(kpres$dists[,1]), c(1,1,10,26))
}
)



# test missings and imputation

x3 <- x4 <- x
x3$x1[1] <- NA
x4$x1[1] <- x4$x2[1] <- x4$x3[1] <- x4$x4[1] <-  NA
test_that("message for NAs.",{
  expect_message(kproto(x3, 4),"Observations with NAs are removed.")
  expect_warning(kproto(x4, 4, na.rm = "no"),"No meaningful cluster assignment possible for observations where all variables NA.")
  expect_error(kproto(x3, 4, na.rm = "not"), "Argument na.rm must be either 'yes','no','imp.internal' or 'imp.onestep'!")
  expect_message(kproto(x3, 4, na.rm = TRUE),"Logical input for na.rm is deprecated. Please use either 'yes','no','imp.internal' or 'imp.onestep'.\n")
  expect_error(kproto(x3, 4, na.rm = "imp.internal", type = "gower"), "Argument na.rm must be either 'yes','no' or 'imp-onestep', since imp.internal is not yet implemented for type = 'gower'!")}
)

prototypes <- data.frame(V1 = factor(c("A","B")), V2 = c(-3,3)) 
x5 <- data.frame(V1 = factor(c(rep("A",10),rep("B",10))), V2 = c(rep(-3, 5), rep(-5, 5), rep(NA, 10))) 
kpres <- kproto(x = x5, k = prototypes, na.rm = "no")
test_that("handling all NAs in variable in a cluster.",{
  expect_equal(kpres$centers[2,2], prototypes[2,2])}
)

x6 <- data.frame(V1 = factor(c(rep("A",10),rep("B",10))), V2 = c(rep(-3, 5), rep(-5, 5), rep(5, 5), rep(NA, 5))) 
kpres_i <- kproto(x = x6, k = prototypes, na.rm = "imp.internal")
kpres_ios <- kproto(x = x6, k = prototypes, na.rm = "imp.onestep")
test_that("application of different imputation strategies",{
  expect_is(kpres_i, "kproto")
  expect_is(kpres_i$cluster, "integer")
  expect_is(kpres_i$dists, "matrix")
  expect_false(any(is.na(kpres_i$data)))
  expect_is(kpres_ios, "kproto")
  expect_is(kpres_ios$cluster, "integer")
  expect_is(kpres_ios$dists, "matrix")
  expect_false(any(is.na(kpres_ios$data)))}
)



# test initialization
kpres <- kproto(x, 4, init = "nbh.dens", verbose = FALSE)
test_that("testing kproto with selected initialization nbh.dens",{
  expect_is(kpres$nstart.m, "NULL")
  expect_is(kpres$inits, "data.frame")
  expect_equal(nrow(kpres$inits), 4)
  expect_equal(ncol(kpres$inits), ncol(x))
}
)
kpres <- kproto(x, 4, init = "nstart.m", verbose = FALSE)
test_that("testing kproto with selected initialization nstart.m",{
  expect_is(kpres$inits, "NULL")
  expect_is(kpres$nstart.m, "numeric")
}
)






# test gower extension
kpres <- kproto(x = x, k = 4, type = "gower")
test_that("Type = gower can be called instead of standard.",{
  expect_equal(kpres$type, "gower")}
)

clusid <- rep(1:4, each = n)
muk    <- 2
# numeric
mus <- c(rep(-muk, n), rep(-muk, n), rep(muk, n), rep(muk, n))
x1  <- rnorm(4*n) + mus
# ordered factor
mus <- c(rep(-muk, n),rep(muk, n),rep(-muk, n),rep(muk, n))
x2 <- rnorm(4*n) + mus
quants <- quantile(x2, seq(0, 1, length.out = (8+1)))
quants[1] <- -Inf
quants[length(quants)] <- Inf
x2 <- as.ordered(cut(x2, quants))
x <- data.frame(x1, x2)

kpres <- kproto(x = x, k = 4)
test_that("Standard still works if ordered factors.",{
  expect_equal(kpres$type, "huang")}
)

kpres <- kproto(x = x, k = 4, type = "standard")
test_that("Standard still works if ordered factors.",{
  expect_equal(kpres$type, "huang")}
)

kpres <- kproto(x = x, k = 4, type = "gower")
test_that("Type = gower will be called.",{
  expect_equal(kpres$type, "gower")}
)

xx <- x
xx[,3] <- factor(x[,2], ordered = F)
kpres <- kproto(x = xx, k = 4, type = "gower")
test_that("Prototypes for ordered factor variables using gower extension.",{
  expect_equal(class(kpres$centers[,2])[1], "ordered")}
)

test_that("Prototypes for (unordered) factor variables using gower extension.",{
  expect_equal(class(kpres$centers[,3]), "factor")}
)

data(iris)
model <- kproto(x = iris, k = 3)
pred  <- predict(model, iris[1, ])
test_that("Prediction for data frame with one single observation.",{
  expect_equal(class(pred$dists)[1], "matrix")}
)

iris2 <- iris
iris2$Species <- as.ordered(iris$Species)
model <- kproto(x = iris2, k = 3, type = "gower")
pred <- predict(model, iris2[1, ])
test_that("Prediction for data frame with one single observation for type = gower.",{
  expect_equal(class(pred$dists)[1], "matrix")}
)

kpres <- kproto(x = x, k = x[1:4,])
test_that("kproto_gower works for is.data.frame(k).",{
  expect_equal(class(kpres), "kproto")}
)

# #############################
# ### ...under construction...
# #############################
# 
# 
# # library(clustMixType)
# # set.seed(42)
# # n   <- 100; prb <- 0.9; muk <- 1.5
# # clusid <- rep(1:4, each = n)
# # x1 <- sample(c("A","B"), 2*n, replace = TRUE, prob = c(prb, 1-prb))
# # x1 <- c(x1, sample(c("A","B"), 2*n, replace = TRUE, prob = c(1-prb, prb)))
# # x1 <- as.factor(x1)
# # x3 <- c(rnorm(n, mean = -muk), rnorm(n, mean = muk), rnorm(n, mean = -muk), rnorm(n, mean = muk))
# # x <- data.frame(x1,x3)
# # 
# # k <- 4
# # f <- "kproto"
# # a <- paste(f, "(x, k = ", k,")", sep = "")
# # 
# # obj <- eval(str2expression(a))
# # names(obj)
# # vars = NULL
# # data = obj$data
# # f = "kproto"
# # boot = 5
# # args = list()
# 
# clustVarImp <- function(obj, data = obj$data, f = "kproto", vars = NULL, boot = 100, args = list()){
#   cl.orig <- obj$cluster
#   k       <- length(unique(cl.orig))
#   n       <- nrow(data)
#   
#   if (is.null(vars)) vars <- 1:ncol(data)
#   if(is.numeric(vars)){
#     if(!any((1:ncol(data)) %in% vars)) stop("Indices in vars are not in data!")
#     if(!all(vars  %in% 1:ncol(data))) warning("Some indices in vars are not in data!")
#   }
#   if(is.character(vars)){
#     if(!any(colnames(data) %in% vars)) stop("No variable name in vars matches any colnames in data!")
#     if(!all(vars  %in% colnames(data))) warning("Some variable name(s) without match in data!")
#     vars <- which.max(colnames(data) %in% vars)
#   }
#   
#   varimps <- matrix(NA, nrow = boot, ncol = length(vars))
#   colnames(varimps) <- colnames(data)[vars]
#   
#   fcall <- paste(f, "(x, k = ", k,")", sep = "")
#   if (length(args) > 0) {
#     for(i in 1: length(args)){
#       fcall <- paste(fcall, ",", args[[i]])
#     }
#   }
# 
#     
#   for(j in 1:length(vars)){
#     for(b in 1:boot){
#       v      <- vars[j]
#       x      <- data 
#       x[,v]  <- sample(x[,v])
#       obj.vb <- eval(str2expression(fcall)) 
#       varimps[b,j] <- mclust::adjustedRandIndex(obj$cluster, obj.vb$cluster)     
#     }
#   }
# 
#   mean.VIs <- apply(varimps, 2, mean)
#   all.VIs  <- varimps
#   return(list(mean.VIs, all.VIs))
# }

###############################################
# According to a suggestion from Sam Sievertsen
###############################################

# add unit test

#' @title Compute individual Silhouette Widths
#' @description Computes an object of class \code{silhouette} from k-prototypes clustering for mixed-type data.
#' 
#' @details Information on the individual silhouette width for all observations is computed. 
#' The k-prototypes result of class \code{kproto} is turned into an object of class \code{silhouette} 
#' that can be further used with the functions \code{\link[cluster]{summary.silhouette}} or \code{\link[cluster]{plot.silhouette}}. 
#' 
#' @param object Object of class \code{kproto}.
#' 
#' @return Object of class \code{silhouette} to be further used with \code{\link[cluster]{summary.silhouette}} 
#' or \code{\link[cluster]{plot.silhouette}}. It consists in a matrix where each row corresponds to one observation 
#' and the columns contain the assigned cluster as well as the neighbor cluster as well as the individual silhouette 
#' width ob the observation.
#' 
#' @examples
#' # generate toy data with factors and numerics
#' 
#' n   <- 50
#' prb <- 0.9
#' muk <- 1.5 
#' clusid <- rep(1:4, each = n)
#' 
#' x1 <- sample(c("A","B"), 2*n, replace = TRUE, prob = c(prb, 1-prb))
#' x1 <- c(x1, sample(c("A","B"), 2*n, replace = TRUE, prob = c(1-prb, prb)))
#' x1 <- as.factor(x1)
#' 
#' x2 <- sample(c("A","B"), 2*n, replace = TRUE, prob = c(prb, 1-prb))
#' x2 <- c(x2, sample(c("A","B"), 2*n, replace = TRUE, prob = c(1-prb, prb)))
#' x2 <- as.factor(x2)
#' 
#' x3 <- c(rnorm(n, mean = -muk), rnorm(n, mean = muk), rnorm(n, mean = -muk), rnorm(n, mean = muk))
#' x4 <- c(rnorm(n, mean = -muk), rnorm(n, mean = muk), rnorm(n, mean = -muk), rnorm(n, mean = muk))
#' 
#' x <- data.frame(x1,x2,x3,x4)
#' 
#' # apply k-prototypes
#' kpres <- kproto(x, 4)
#' 
#' # compute silhouette widhts
#' # extract object of class sil
#' sils <- kproto2silhouette(kpres)
#' 
#' # apply functions from package cluster
#' library(cluster)
#' summary(sils)
#' plot(sils)
#' 
#' # compute average silhouette width  
#' mean(sils[,3])
#' 
#' # ...alternatively, compute average silhouette width directly using validation_kproto() 
#' cindex_value <- validation_kproto(method = "silhouette", object = kpres)
#' cindex_value
#' 
#' @author \email{gero.szepannek@@web.de}
#' 
#' 
#' @references \itemize{
#'     \item Szepannek, G. (2018): 
#'     clustMixType: User-Friendly Clustering of Mixed-Type Data in R, 
#'     {\emph{The R Journal 10/2}}, 200-208. 
#'     \doi{10.32614/RJ-2018-048}.
#'     
#'     \item Aschenbruck, R., Szepannek, G. (2020): 
#'     Cluster Validation for Mixed-Type Data. 
#'     \emph{Archives of Data Science, Series A, Vol 6, Issue 1}.
#'     \doi{10.5445/KSP/1000098011/02}.
#'     
#'     \item Kaufman, L., Rousseeuw, P. (1990): 
#'     \emph{Finding Groups in Data: An Introduction to Cluster Analysis.} 
#'      Wiley.
#'   }
#' @rdname kproto2silhouette
#' 
#' @importFrom cluster silhouette
#' 
#' @export
kproto2silhouette <- function(object){ 
  
  if(!(inherits(object, "kproto"))) stop("Object must be of class kproto!")
  type = object$type
  if(!("data" %in% names(object))) stop("Computation is not possible. The kproto object has to be vreated using the argument keep.data = TRUE!")
  data = object$data 
  k = length(object$withinss) 
  lambda = object$lambda 
  # kp_obj = "optimal", 
  # verbose = FALSE, ...
  # data_plus = NULL, 
  
  
  if(object$type == "gower"){  #...same as from validation_kproto.R ln. 1095--1166  
    data_plus <- list()
    
    data_plus$numvars <- sapply(object$data, is.numeric)
    data_plus$ordvars <- sapply(object$data, is.ordered)
    data_plus$catvars <- sapply(object$data, is.factor) & !data_plus$ordvars
    
    # ranges for normalization and transformed data for gower's distance calculation
    if(any(data_plus$numvars)) data_plus$rgnums <- sapply(object$data[, data_plus$numvars, drop = FALSE], function(z) diff(range(z)))
    if(any(data_plus$ordvars)){
      data_plus$data_ord <- object$data[, data_plus$ordvars, drop = FALSE] # store original variables 
      # ...and replace ordered variables by their ranks
      for(jord in which(data_plus$ordvars)) object$data[,jord] <- rank(object$data[,jord])
      data_plus$rgords <- sapply(object$data[, data_plus$ordvars, drop = FALSE], function(z) diff(range(z)))
    }
  }
  if(object$type == "huang"){
    data_plus <- list()
    data_plus$numvars <- sapply(object$data, is.numeric)
    data_plus$catvars <- sapply(object$data, is.factor)
  }
  
  
  dists <- calc.dist(type = object$type, lambda = object$lambda, 
                     data1 = object$data, all.dists = TRUE,
                     data_plus = data_plus)
  
  cluster_dists <- matrix(numeric(nrow(object$data)*length(table(object$cluster))), 
                          nrow = nrow(object$data), ncol = length(table(object$cluster)))
  for(i in 1:length(table(object$cluster))){
    if(!(length(which(object$cluster == i)) == 1)){
      cluster_dists[,i] <- rowMeans(dists[,which(object$cluster == i)])
    }else{
      cluster_dists[,i] <- dists[,which(object$cluster == i)]
    }
  }
  
  
  #determine ai, bi and si as well as the neighbour cluster
  a <- numeric(nrow(object$data))
  b <- numeric(nrow(object$data))
  s <- numeric(nrow(object$data))
  neighbour <- numeric(nrow(object$data))
  
  for(i in 1:nrow(object$data)){
    if(is.na(object$cluster[i])){
      # special case: no cluster assignment cluster[i] (usually resulting of all variables NA in x[i])
      s[i] <- NA
    }else{
      a[i] <- cluster_dists[i, object$cluster[i]]
      b[i] <- min(cluster_dists[i, -object$cluster[i]])
      
      # computation of the neighbour is a bit complicated but hopefully also correct in case 
      # ...the cluster assignment has been sampled from two equally close clusters  
      neighbours.i <- order(cluster_dists[i,])
      neighbours.i <- neighbours.i[neighbours.i != object$cluster[i]]
      neighbour[i] <- neighbours.i[1]
      
      if(max(a[i], b[i], na.rm = TRUE) == 0){
        # special case: a[i]=0 and b[i]=0 => s = 0, since x[i] lies equally far away (distance = 0) from both the clusters
        s[i] <- 0
      }else{
        s[i] <- (b[i] - a[i])/max(a[i],b[i], na.rm = TRUE)
      }
    }
  }
  if(any(table(object$cluster) == 1)){
    for(i in which(object$cluster %in% as.integer(which(table(object$cluster) == 1)))){
      s[i] <- 0
    }
    cat(length(which(object$cluster %in% as.integer(which(table(object$cluster) == 1))))," cluster with only one observation\n")
  }
  
  result <- cbind(object$cluster, neighbour, s)
  colnames(result) <- c("cluster", "neighbor", "sil_width")
  class(result) <- 'silhouette'
  attr(result, "Ordered") <- FALSE
  return(result)
}



# calc.dist <- function(type, lambda, data1, data2 = NULL, all.dists = FALSE, data_plus = NULL){
#   
#   if(type == "huang"){# calculation with standard distance of k-prototypes
#     
#     # if(is.null(lambda)){
#     #   lambda <- lambdaest(data, num.method = 1, fac.method = 1, outtype = "numeric", verbose = TRUE)
#     # }
#     
#     if(!is.null(data2)){#contains the input two datasets?
#       
#       #tbd: check if structure of data1 and data2 are equal
#       data <- rbind(data1, data2)
#       
#       if(nrow(data1) == 1 & nrow(data2) == 1){#calculation the distance between two observations (due to saving time)
#         
#         #numerical variables
#         d1 <- (data[1, data_plus$numvars, drop = FALSE]-data[2, data_plus$numvars, drop = FALSE])^2
#         d1[is.na(d1)] <- 0
#         if(length(lambda) == 1){
#           d1 <- sum(d1)
#         }else{ #length(lambda) > 1
#           d1 <- as.matrix(d1, nrow = 1) %*% matrix(lambda[data_plus$numvars], ncol = 1)
#         }
#         
#         #categorical variables
#         d2 <- sapply(which(data_plus$catvars), function(j) return(data[1,j] != data[2,j]))
#         d2[is.na(d2)] <- FALSE
#         if(length(lambda) == 1){
#           d2 <- lambda * sum(d2)
#         }else{#length(lambda) > 1
#           d2 <- matrix(d2, nrow = 1) %*% matrix(lambda[data_plus$catvars], ncol = 1)
#         }
#         
#         return(d1 + d2)
#       }else{#calculate distances between objects of two datasets
#         
#         #numerical variables
#         d1 <- (data1[rep(1:nrow(data1), each=nrow(data2)), data_plus$numvars, drop = FALSE] -
#                  data2[rep(1:nrow(data2), times=nrow(data1)), data_plus$numvars, drop = FALSE])^2
#         d1[is.na(d1)] <- 0
#         if(length(lambda) == 1){
#           d1 <- rowSums(d1)
#         }else{ #length(lambda) > 1
#           d1 <- as.matrix(d1) %*% lambda[data_plus$numvars]
#         }
#         
#         #categorical variables
#         d2 <- data1[rep(1:nrow(data1), each=nrow(data2)), data_plus$catvars, drop = FALSE] != 
#           data2[rep(1:nrow(data2), times = nrow(data1)), data_plus$catvars, drop = FALSE]
#         d2[is.na(d2)] <- FALSE
#         if(length(lambda) == 1){
#           d2 <- lambda * rowSums(d2)
#         }else{ #length(lambda) > 1
#           d2 <- as.matrix(d2) %*% lambda[data_plus$catvars]
#         }
#         
#         return(sum(d1 + d2))
#       }
#     }else{ #calculate distances between the objects of one dataset
#       
#       d1 <- (data1[rep(1:(nrow(data1)-1),times=(nrow(data1)-1):1, each=1), data_plus$numvars, drop = FALSE] -
#                data1[unlist(lapply(2:nrow(data1), seq, to=nrow(data1))), data_plus$numvars, drop = FALSE])^2
#       d1[is.na(d1)] <- 0
#       if(length(lambda) == 1) d1 <- rowSums(d1)
#       if(length(lambda) > 1) d1 <- as.matrix(d1) %*% lambda[data_plus$numvars]
#       
#       d2 <- data1[rep(1:(nrow(data1)-1),times=(nrow(data1)-1):1, each=1), data_plus$catvars, drop = FALSE] != 
#         data1[unlist(lapply(2:nrow(data1), seq, to=nrow(data1))), data_plus$catvars, drop = FALSE]
#       d2[is.na(d2)] <- FALSE
#       if(length(lambda) == 1) d2 <- lambda * rowSums(d2)
#       if(length(lambda) > 1) d2 <- as.matrix(d2) %*% lambda[data_plus$catvars]
#       
#       if(all.dists == TRUE){
#         m <- matrix(rep(0,nrow(data1)^2), ncol = nrow(data1))
#         
#         m[lower.tri(m)] <- d1+d2
#         m <- t(m)
#         m[lower.tri(m)] <- d1+d2
#         return(m)
#       }else{
#         return(sum(d1 + d2))
#       }
#     }
#   }else{ #calculation with gower distance
#     
#     if(!is.null(data2)){#contains the input two datasets?
#       
#       #tbd: check if structure of data1 and data2 are equal
#       #data <- rbind(data1, data2)
#       
#       if(nrow(data1) == 1 & nrow(data2) == 1){#calculation the distance between two observations (due to saving time)
#         
#         d1 <- d2 <- d3 <- 0
#         
#         if(any(data_plus$numvars)){
#           d1 <- abs(data1[1, data_plus$numvars, drop = FALSE] - data2[1, data_plus$numvars, drop = FALSE])
#           for(jnum in 1:ncol(d1)) d1[,jnum] <- d1[,jnum] / data_plus$rgnums[jnum] 
#           d1[is.na(d1)] <- 0
#           if(length(lambda) > 1) d1 <- as.matrix(d1) %*% lambda[data_plus$numvars]
#           if(is.null(lambda)) d1 <- rowSums(d1)
#         }
#         
#         if(any(data_plus$catvars)){
#           d2 <- data1[1, data_plus$catvars, drop = FALSE] != data2[1, data_plus$catvars, drop = FALSE]
#           d2[is.na(d2)] <- FALSE
#           if(length(lambda) > 1) d2 <- as.matrix(d2) %*% lambda[data_plus$catvars]
#           if(is.null(lambda)) d2 <- rowSums(d2)
#         }
#         
#         if(any(data_plus$ordvars)){
#           d3 <- abs(data1[1, data_plus$ordvars, drop = FALSE] - data2[1, data_plus$ordvars, drop = FALSE])
#           for(jord in 1:ncol(d3)) d3[,jord] <- d3[,jord] / data_plus$rgords[jord] 
#           d3[is.na(d3)] <- 0
#           if(length(lambda) > 1) d3 <- as.matrix(d3) %*% lambda[data_plus$ordvars]
#           if(is.null(lambda)) d3 <- rowSums(d3)
#         }
#         
#         return(d1 + d2 + d3)
#         
#       }else{#calculate distances between objects of two datasets
#         
#         # in case of no numeric / factor / ordinal variables set:
#         d1 <- d2 <- d3 <- rep(0, nrow(data1)*nrow(data2))
#         
#         if(any(data_plus$numvars)){
#           d1 <- abs(data1[rep(1:nrow(data1), each=nrow(data2)), data_plus$numvars, drop = FALSE] - 
#                       data2[rep(1:nrow(data2), times=nrow(data1)), data_plus$numvars, drop = FALSE])
#           for(jnum in 1:ncol(d1)) d1[,jnum] <- d1[,jnum] / data_plus$rgnums[jnum] 
#           #time efficient alternative for codeline before:
#           # d1 <- d1 / rep(data_plus$rgnums, rep(nrow(d1), ncol(d1)))
#           d1[is.na(d1)] <- 0
#           if(length(lambda) > 1) d1 <- as.matrix(d1) %*% lambda[data_plus$numvars]
#           if(is.null(lambda)) d1 <- rowSums(d1)
#         }
#         
#         if(any(data_plus$catvars)){
#           d2 <- data1[rep(1:nrow(data1), each=nrow(data2)), data_plus$catvars, drop = FALSE] != 
#             data2[rep(1:nrow(data2), times = nrow(data1)), data_plus$catvars, drop = FALSE]
#           d2[is.na(d2)] <- FALSE
#           if(length(lambda) > 1) d2 <- as.matrix(d2) %*% lambda[data_plus$catvars]
#           if(is.null(lambda)) d2 <- rowSums(d2)
#         }
#         
#         if(any(data_plus$ordvars)){
#           d3 <- abs(data1[rep(1:nrow(data1), each=nrow(data2)), data_plus$ordvars, drop = FALSE] - 
#                       data2[rep(1:nrow(data2), times=nrow(data1)), data_plus$ordvars, drop = FALSE])
#           for(jord in 1:ncol(d3)) d3[,jord] <- d3[,jord] / data_plus$rgords[jord] 
#           #time efficient alternative for codeline before:
#           # d3 <- d3 / rep(data_plus$rgords, rep(nrow(d3), ncol(d3)))
#           d3[is.na(d3)] <- 0
#           if(length(lambda) > 1) d3 <- as.matrix(d3) %*% lambda[data_plus$ordvars]
#           if(is.null(lambda)) d3 <- rowSums(d3)
#         }
#         
#         return(sum(d1 + d2 + d3))
#       }
#     }else{ #calculate distances between the objects of one dataset
#       
#       d1 <- d2 <- d3 <- rep(0,length(rep(1:(nrow(data1)-1),times=(nrow(data1)-1):1, each=1)))
#       
#       if(any(data_plus$numvars)){
#         d1 <- abs(data1[rep(1:(nrow(data1)-1),times=(nrow(data1)-1):1, each=1), data_plus$numvars, drop = FALSE] - 
#                     data1[unlist(lapply(2:nrow(data1), seq, to=nrow(data1))), data_plus$numvars, drop = FALSE])
#         for(jnum in 1:ncol(d1)) d1[,jnum] <- d1[,jnum] / data_plus$rgnums[jnum] 
#         #time efficient alternative for codeline before:
#         # d1 <- d1 / rep(rgnums, rep(nrow(d1), ncol(d1)))
#         d1[is.na(d1)] <- 0
#         if(length(lambda) > 1) d1 <- as.matrix(d1) %*% lambda[data_plus$numvars]
#         if(is.null(lambda)) d1 <- rowSums(d1)
#       }
#       
#       if(any(data_plus$catvars)){
#         d2 <- data1[rep(1:(nrow(data1)-1),times=(nrow(data1)-1):1, each=1), data_plus$catvars, drop = FALSE] != 
#           data1[unlist(lapply(2:nrow(data1), seq, to=nrow(data1))), data_plus$catvars, drop = FALSE]
#         d2[is.na(d2)] <- FALSE
#         if(length(lambda) > 1) d2 <- as.matrix(d2) %*% lambda[data_plus$catvars]
#         if(is.null(lambda)) d2 <- rowSums(d2)
#       }
#       
#       if(any(data_plus$ordvars)){
#         d3 <- abs(data1[rep(1:(nrow(data1)-1),times=(nrow(data1)-1):1, each=1), data_plus$ordvars, drop = FALSE] - 
#                     data1[unlist(lapply(2:nrow(data1), seq, to=nrow(data1))), data_plus$ordvars, drop = FALSE])
#         for(jord in 1:ncol(d3)) d3[,jord] <- d3[,jord] / data_plus$rgords[jord] 
#         d3[is.na(d3)] <- 0
#         if(length(lambda) > 1) d3 <- as.matrix(d3) %*% lambda[data_plus$ordvars]
#         if(is.null(lambda)) d3 <- rowSums(d3)
#       }
#       
#       if(all.dists == TRUE){ #return distance matrix
#         m <- matrix(rep(0,nrow(data1)^2), ncol = nrow(data1))
#         
#         m[lower.tri(m)] <- d1+d2+d3
#         m <- t(m)
#         m[lower.tri(m)] <- d1+d2+d3
#         return(m)
#       }else{ #return the sum of all dists between objects
#         return(sum(d1 + d2 + d3))
#       }
#       
#     }
#   }
# }






# #####################################################################
# ### determination of distances, required for initialization functions
# dists_kproto <- function(x, y = NULL, lambda = NULL, verbose = FALSE){
#   
#   if(is.null(y)){
#     dat_part1 <- x[rep(c(1:nrow(x)), each = nrow(x)),]
#     dat_part2 <- x[rep(c(1:nrow(x)), times = nrow(x)),]
#   }else{
#     if(nrow(x) == 1 & nrow(y) == 1){
#       return(cbind(x, y, dist = dist_kproto(x,y,lambda = lambda)))
#     }else{
#       dat_part1 <- x[rep(c(1:nrow(x)), each = nrow(y)),]
#       dat_part2 <- y[rep(c(1:nrow(y)), times = nrow(x)),]
#     }
#   }
#   
#   # check for numeric and factor variables
#   numvars <- sapply(x, is.numeric)
#   anynum <- any(numvars)
#   catvars <- sapply(x, is.factor)
#   anyfact <- any(catvars)
#   
#   # determination of lambda
#   if(length(lambda) > 1) {if(length(lambda) != sum(c(numvars,catvars))) stop("If lambda is a vector, its length should be the sum of numeric and factor variables in the data frame!")}
#   if(is.null(lambda)){
#     if(anynum & anyfact){
#       vnum <- mean(sapply(x[,numvars, drop = FALSE], var, na.rm = TRUE))
#       vcat <- mean(sapply(x[,catvars, drop = FALSE], function(z) return(1-sum((table(z)/sum(!is.na(z)))^2))))
#       if (vnum == 0){
#         if(verbose) warning("All numerical variables have zero variance.")
#         anynum <- FALSE
#       } 
#       if (vcat == 0){
#         if(verbose) warning("All categorical variables have zero variance.")
#         anyfact <- FALSE
#       } 
#       if(anynum & anyfact){
#         lambda <- vnum/vcat
#         if(verbose) cat("Estimated lambda:", lambda, "\n\n")
#       }else{
#         lambda <- 1
#       }
#     }
#   }
#   
#   # compute distances 
#   nrows <- nrow(x)
#   d1 <- (dat_part1[,numvars, drop = FALSE] - dat_part2[,numvars, drop = FALSE])^2
#   if(length(lambda) == 1) d1 <- rowSums(d1, na.rm = TRUE)
#   if(length(lambda) > 1) d1 <- as.matrix(d1) %*% lambda[numvars]
#   d2 <- sapply(which(catvars), function(j) return(dat_part1[,j] != dat_part2[,j]))
#   d2[is.na(d2)] <- FALSE
#   if(length(lambda) == 1) d2 <- lambda * rowSums(d2)
#   if(length(lambda) > 1) d2 <- as.matrix(d2) %*% lambda[catvars]
#   
#   return(cbind(dat_part1, dat_part2, dist = as.vector(d1 + d2)))
# }
# 
# 
# 
# dist_kproto <- function(x, y, lambda, verbose = FALSE){
#   
#   x <- rbind(x,y)
#   
#   # check for numeric and factor variables
#   numvars <- sapply(x, is.numeric)
#   anynum <- any(numvars)
#   catvars <- sapply(x, is.factor)
#   anyfact <- any(catvars)
#   
#   # compute distances 
#   nrows <- nrow(x)
#   d1 <- (x[1, numvars, drop = FALSE] - x[2, numvars, drop = FALSE])^2
#   if(length(lambda) == 1) d1 <- rowSums(d1, na.rm = TRUE)
#   if(length(lambda) > 1) d1 <- as.matrix(d1) %*% lambda[numvars]
#   d2 <- sapply(which(catvars), function(j) return(x[1,j] != x[2,j]))
#   d2[is.na(d2)] <- FALSE
#   if(length(lambda) == 1) d2 <- lambda * rowSums(matrix(d2, nrow = 1))
#   if(length(lambda) > 1) d2 <- as.matrix(d2) %*% lambda[catvars]
#   
#   return(as.numeric(d1 + d2))
# }
# 

kproto_gower <- function(x, k, lambda = NULL, iter.max = 100, na.rm = "yes", keep.data = TRUE, verbose = TRUE){
  # # enable input of tibbles #...done within kproto()
  # if(is_tibble(x) == TRUE){x <- as.data.frame(x)}
  
  # initial error checks
  if(!is.data.frame(x)) stop("x should be a data frame!")
  if(ncol(x) < 2) stop("For clustering x should contain at least two variables!")
  #  if(iter.max < 1 | nstart < 1) stop("iter.max and nstart must not be specified < 1!") # no parameter within kproto_gower()
  if(!is.null(lambda)){
    if(any(lambda < 0)) stop("lambda must be specified >= 0!")
    if(!any(lambda > 0)) stop("lambda must be specified > 0 for at least two variables!")
    if(length(lambda) != ncol(x)) {
      warning("For gower distance length(lambda) must match the # of variables. lambda will be ignored.")
      lambda <- NULL
      }
    }
  # check for numeric and factor variables
  numvars <- sapply(x, is.numeric)
  anynum <- any(numvars)
  ordvars <- sapply(x, is.ordered)
  anyord <- any(ordvars)
  catvars <- sapply(x, is.factor) & !ordvars
  anyfact <- any(catvars)
  
  # initialize lookup table to store standardized ranks / ranges of numeric variables for predict.kproto()
  lookup <- list() 
  
  # vector of ranges for normalization  
  if(any(numvars)) rgnums <- sapply(x[, numvars, drop = FALSE], function(z) diff(range(z, na.rm = TRUE)))
  if(any(ordvars)){
    xord   <- x[, ordvars, drop = FALSE] # store original variables 
    # ...and replace ordered variables by their ranks
    for(jord in which(ordvars)) x[,jord] <- rank(x[,jord], na.last = "keep")
    rgords <- sapply(x[, ordvars, drop = FALSE], function(z) diff(range(z, na.rm = TRUE)))
  }

  # initialize prototypes
  if(!is.data.frame(k)){
    if (length(k) == 1){
      if(as.integer(k) != k){k <- as.integer(k); warning(paste("k has been set to", k,"!"))}
      if(sum(complete.cases(x)) < k) stop("Data frame has less complete observations than clusters!")
      ids <- sample(row.names(x[complete.cases(x),]), k)
      protos <- x[ids,]
    }
    if (length(k) > 1){
      if(nrow(x) < length(k)) stop("Data frame has less observations than clusters!")
      ids <- k
      k <- length(ids)
      if(length(unique(ids)) != length(ids)) stop("If k is specified as a vector it should contain different indices!")
      if(any(ids<1)|any(ids>nrow(x))) stop("If k is specified as a vector all elements must be valid indices of x!")
      #check for integer
      protos <- x[ids,]
      if(any(!complete.cases(protos))) stop("Choose initial prototypes without missing values!")
    }
    rm(ids)
  }
  if(is.data.frame(k)){
    if(nrow(x) < nrow(k)) stop("Data frame has less observations than clusters!")
    if(length(names(k)) != length(names(x))) stop("k and x have different numbers of columns!")
    if(any(names(k) != names(x))) stop("k and x have different column names!")
    if(anynum) {if( any(sapply(k, is.numeric) != numvars)) stop("Numeric variables of k and x do not match!")}
    if(anyfact) {if( any(sapply(k, is.factor) != catvars)) stop("Factor variables of k and x do not match!")}

    if(anyord){
      # ...and replace ordered variables for initial prototypes by their ranks according to 
      for(jord in 1:ncol(xord)){
        levsj <- unique(xord[,jord])
        rgsj  <- unique(x[,which(ordvars)[jord]])
        names(rgsj) <- levsj
        if(!all(k[,which(ordvars)[jord]] %in% levsj)) stop(paste("Some levels of initial prototypes do not match levels in variable", names(k)[which(ordvars)[jord]],"!")) 
        k[,which(ordvars)[jord]] <- as.numeric(rgsj[k[,which(ordvars)[jord]]])
      }
    }

    protos <- k
    if(any(!complete.cases(protos))) stop("Prototypes with missing values. Choose initial prototypes without missing values!")
    k <- nrow(protos)
  }
  if(k < 1) stop("Number of clusters k must not be smaller than 1!")
  
  if(length(lambda) > 0){
    if(length(lambda) != sum(c(numvars,catvars,ordvars))) {
      warning("For gower distance if lambda is specified, its length must be the sum of numeric and factor variables in the data frame! Otherwise it will be ignored.")
      lambda <- NULL
    }
  }
  
  # initialize clusters
  clusters  <- numeric(nrow(x)) 
  tot.dists <- NULL
  moved   <- NULL
  iter <- 1
  
  # check for any equal prototypes and reduce cluster number in case of occurence
  if(k > 1){
    keep.protos <- rep(TRUE,k)
    for(l in 1:(k-1)){
      for(m in (l+1):k){
        d1 <- d2 <- d3 <- 0
        if(any(numvars)) d1 <- sum((protos[l,numvars, drop = FALSE]-protos[m,numvars, drop = FALSE])^2) # euclidean for numerics
        if(any(catvars)) d2 <- sum(protos[l,catvars, drop = FALSE] != protos[m,catvars, drop = FALSE]) # compare levels 
        if(any(ordvars)) d3 <- sum((protos[l,ordvars, drop = FALSE]-protos[m,ordvars, drop = FALSE])^2) # euclidean for ranks of ordinals
        if((d1+d2+d3) == 0) keep.protos[m] <- FALSE 
      }
    }
    if(!all(keep.protos)){
      protos <- protos[keep.protos,]
      k <- sum(keep.protos)
      if(verbose) message("Equal prototypes merged. Cluster number reduced to:", k, "\n\n")      
    }
  }

  # special case only one cluster
  if(k == 1){clusters <- rep(1, nrow(x)); size  <- table(clusters); iter <- iter.max} # REM: named vector size is needed later...
  
  # start iterations for standard case (i.e. k > 1)
  while(iter < iter.max){
    
    # compute distances 
    nrows <- nrow(x)
    dists <- matrix(NA, nrow=nrows, ncol = k)
    for(i in 1:k){
      
      # in case of no numeric / factor / ordinal variables set:
      d1 <- d2 <- d3 <- rep(0, nrows)
      
      if(any(numvars)){
        d1 <- abs(x[, numvars, drop = FALSE] - matrix(rep(as.numeric(protos[i, numvars, drop = FALSE]), nrows), nrow=nrows, byrow=T))
        for(jnum in 1:ncol(d1)) d1[,jnum] <- d1[,jnum] / rgnums[jnum] 
        d1[is.na(d1)] <- 0
        if(length(lambda) > 1) d1 <- as.matrix(d1) %*% lambda[numvars]
        if(is.null(lambda)) d1 <- rowSums(d1)
      }
      
      if(any(catvars)){
        d2 <- sapply(which(catvars), function(j) return(x[,j] != rep(protos[i,j], nrows)) )
        d2[is.na(d2)] <- FALSE
        if(length(lambda) > 1) d2 <- as.matrix(d2) %*% lambda[catvars]
        if(is.null(lambda)) d2 <- rowSums(d2)
      }

      if(any(ordvars)){
        d3 <- abs(x[, ordvars, drop = FALSE] - matrix(rep(as.numeric(protos[i, ordvars, drop = FALSE]), nrows), nrow=nrows, byrow=T))
        for(jord in 1:ncol(d3)) d3[,jord] <- d3[,jord] / rgords[jord] 
        d3[is.na(d3)] <- 0
        if(length(lambda) > 1) d3 <- as.matrix(d3) %*% lambda[ordvars]
        if(is.null(lambda)) d3 <- rowSums(d3)
      }
      
      dists[,i] <- d1 + d2 + d3
    }
    
    # assign clusters 
    old.clusters  <- clusters
    # clusters      <- apply(dists, 1, function(z) which.min(z))
    clusters      <- apply(dists, 1, function(z) {a <- which(z == min(z)); if (length(a)>1) a <- sample(a,1); return(a)}) # sample in case of multiple minima
    size          <- table(clusters)  
    min.dists     <- apply(cbind(clusters, dists), 1, function(z) z[z[1]+1])
    within        <- as.numeric(by(min.dists, clusters, sum))
    tot.within    <- sum(within)
    # prevent from empty classes
    #tot.within    <- numeric(k)
    #totw.list     <- by(min.dists, clusters, sum) 
    #tot.within[names(totw.list)] <- as.numeric(totw.list)
    
    # ...check for empty clusters and eventually reduce number of prototypes    
    if (length(size) < k){
      k <- length(size)
      protos <- protos[1:length(size),]  
      if(verbose) cat("Empty clusters occur. Cluster number reduced to:", k, "\n\n")
    }
    
    # trace
    tot.dists <- c(tot.dists, sum(tot.within))      
    moved <- c(moved, sum(clusters != old.clusters))
   
    # compute new prototypes
    remids <- as.integer(names(size))
    for(i in remids){
      some_vals <- sapply(x[clusters == i, , drop = FALSE], function(z) !all(is.na(z))) # only update variables if not all values are NA
      if(any(some_vals & numvars)){
        protos[which(remids == i), some_vals & numvars] <- sapply(x[clusters == i, some_vals & numvars, drop = FALSE], median, na.rm = TRUE)
      }
      if(any(some_vals & catvars)){
        protos[which(remids == i), some_vals & catvars] <- sapply(x[clusters == i, some_vals & catvars, drop = FALSE], function(z) levels(z)[which.max(table(z))])
      }
      if(any(some_vals & ordvars)){
        protos[which(remids == i), some_vals & ordvars] <- sapply(x[clusters == i, some_vals & ordvars, drop = FALSE], median, na.rm = TRUE)
      }
    }
    
    if(k == 1){clusters <- rep(1, length(clusters)); size <- table(clusters); iter <- iter.max; break}
    
    # check for any equal prototypes and reduce cluster number in case of occurence
    if(iter == (iter.max-1)){ # REM: for last iteration equal prototypes are allowed. otherwise less prototypes than assigned clusters.
      keep.protos <- rep(TRUE,k)
      for(l in 1:(k-1)){
        for(m in (l+1):k){        
          d1 <- d2 <- d3 <- 0
          if(any(numvars)) d1 <- sum((protos[l,numvars, drop = FALSE]-protos[m,numvars, drop = FALSE])^2) # euclidean for numerics
          if(any(catvars)) d2 <- sum(protos[l,catvars, drop = FALSE] != protos[m,catvars, drop = FALSE]) # compare levels for categorics 
          if(any(ordvars)) d3 <- sum((protos[l,ordvars, drop = FALSE]-protos[m,ordvars, drop = FALSE])^2) # euclidean for ranks of ordinals
          if((d1+d2+d3) == 0) keep.protos[m] <- FALSE 
        }
      }
      if(!all(keep.protos)){
        protos <- protos[keep.protos,]
        k <- sum(keep.protos)
        if(verbose) cat("Equal prototypes merged. Cluster number reduced to:", k, "\n\n")      
      }
    }

    # add stopping rules
    if(moved[length(moved)] ==  0) break
    
    if(k == 1){clusters <- rep(1, length(clusters)); size <- table(clusters); iter <- iter.max; break}
    
    #cat("iter", iter, "moved", moved[length(moved)], "tot.dists",tot.dists[length(tot.dists)],"\n" )      
    iter <- iter+1
  }

  ### Final update of prototypes and dists
  if(iter == iter.max){ # otherwise there have been no moves anymore and prototypes correspond to cluster assignments 
    # compute new prototypes
    remids <- as.integer(names(size))
    for(i in remids){
      some_vals <- sapply(x[clusters == i, , drop = FALSE], function(z) !all(is.na(z))) # only update variables if not all values are NA
      if(any(some_vals & numvars)){
        protos[which(remids == i), some_vals & numvars] <- sapply(x[clusters == i, some_vals & numvars, drop = FALSE], median, na.rm = TRUE)
      }
      if(any(some_vals & catvars)){
        protos[which(remids == i), some_vals & catvars] <- sapply(x[clusters == i, some_vals & catvars, drop = FALSE], function(z) levels(z)[which.max(table(z))])
      }
      if(any(some_vals & ordvars)){
        protos[which(remids == i), some_vals & ordvars] <- sapply(x[clusters == i, some_vals & ordvars, drop = FALSE], median, na.rm = TRUE)
      }
    }
    
    # compute distances 
    nrows <- nrow(x)
    dists <- matrix(NA, nrow=nrows, ncol = k)
    for(i in 1:k){
      
      # in case of no numeric / factor / ordinal variables set:
      d1 <- d2 <- d3 <- rep(0, nrows)
      
      if(any(numvars)){
        d1 <- abs(x[, numvars, drop = FALSE] - matrix(rep(as.numeric(protos[i, numvars, drop = FALSE]), nrows), nrow=nrows, byrow=T))
        for(jnum in 1:ncol(d1)) d1[,jnum] <- d1[,jnum] / rgnums[jnum] 
        d1[is.na(d1)] <- 0
        if(length(lambda) > 1) d1 <- as.matrix(d1) %*% lambda[numvars]
        if(is.null(lambda)) d1 <- rowSums(d1)
      }
      
      if(any(catvars)){
        d2 <- sapply(which(catvars), function(j) return(x[,j] != rep(protos[i,j], nrows)) )
        d2[is.na(d2)] <- FALSE
        if(length(lambda) > 1) d2 <- as.matrix(d2) %*% lambda[catvars]
        if(is.null(lambda)) d2 <- rowSums(d2)
      }
      
      if(any(ordvars)){
        d3 <- abs(x[, ordvars, drop = FALSE] - matrix(rep(as.numeric(protos[i, ordvars, drop = FALSE]), nrows), nrow=nrows, byrow=T))
        for(jord in 1:ncol(d3)) d3[,jord] <- d3[,jord] / rgords[jord] 
        d3[is.na(d3)] <- 0
        if(length(lambda) > 1) d3 <- as.matrix(d3) %*% lambda[ordvars]
        if(is.null(lambda)) d3 <- rowSums(d3)
      }
      
      dists[,i] <- d1 + d2 + d3
    }

        
    size          <- table(clusters)  
    min.dists     <- apply(cbind(clusters, dists), 1, function(z) z[z[1]+1])
    within        <- as.numeric(by(min.dists, clusters, sum))
    tot.within    <- sum(within)
  }
  
  # replace ranks of ordinal variables from prototypes by the level with the closest rank
  if(any(ordvars)){
    protos.ord <- protos[,ordvars, drop = FALSE]
    for(jord in 1:ncol(protos.ord)){
      protos[,which(ordvars)[jord]] <- sapply(protos.ord[,jord], function(z) xord[which.min(abs(rank(xord[,jord]) - z)), jord])
    }
    for(jord in which(ordvars)) protos[,jord] <- factor(protos[,jord], levels = levels(protos[,jord]), ordered = TRUE)
    
    # store standardizied ranks in lookup table for predict.kproto()
    for(jord in 1:ncol(xord)){
      df <- data.frame(unique(xord[,jord]),
                       unique(x[,which(ordvars)[jord], drop = FALSE] / rgords[jord])
                       )
      names(df) <- c("level", "value")
      lookup[[names(xord)[jord]]] <- df
    }  
  }
  # store ranges of numeric variables for predict.kproto
  if(any(numvars)) lookup$num_ranges <- rgnums

  if(na.rm == "no"){
    allNAs <- apply(x,1,function(z) all(is.na(z))) # note: not passed from kproto!  
    if(sum(allNAs) > 0){
      clusters[allNAs] <- NA
      dists[allNAs,] <- NA
    }
  }

  names(clusters) <- row.names(dists) <- row.names(x)
  rownames(protos) <- NULL
  # create result: 
  res <- list(cluster = clusters,  
              centers = protos, 
              lambda = lambda, 
              size = size,
              withinss = within,
              tot.withinss = tot.within,   
              dists = dists, 
              iter = iter,
              stdization = lookup,
              trace = list(tot.dists = tot.dists, moved = moved))
  
  # if(keep.data) res$data = x  # ...done by kproto() 
  # class(res) <- "kproto"
  return(res)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/kprototypes.R
\name{clprofiles}
\alias{clprofiles}
\title{Profiling k-Prototypes Clustering}
\usage{
clprofiles(object, x, vars = NULL, col = NULL)
}
\arguments{
\item{object}{Object resulting from a call of resulting \code{kproto}. Also other \code{kmeans} like objects with \code{object$cluster} and \code{object$size} are possible.}

\item{x}{Original data.}

\item{vars}{Optional vector of either column indices or variable names.}

\item{col}{Palette of cluster colours to be used for the plots. As a default RColorBrewer's 
\code{brewer.pal(max(unique(object$cluster)), "Set3")} is used for k > 2 clusters and lightblue and orange else.}
}
\description{
Visualization of a k-prototypes clustering result for cluster interpretation.
}
\details{
For numerical variables boxplots and for factor variables barplots of each cluster are generated.
}
\examples{
# generate toy data with factors and numerics

n   <- 100
prb <- 0.9
muk <- 1.5 
clusid <- rep(1:4, each = n)

x1 <- sample(c("A","B"), 2*n, replace = TRUE, prob = c(prb, 1-prb))
x1 <- c(x1, sample(c("A","B"), 2*n, replace = TRUE, prob = c(1-prb, prb)))
x1 <- as.factor(x1)

x2 <- sample(c("A","B"), 2*n, replace = TRUE, prob = c(prb, 1-prb))
x2 <- c(x2, sample(c("A","B"), 2*n, replace = TRUE, prob = c(1-prb, prb)))
x2 <- as.factor(x2)

x3 <- c(rnorm(n, mean = -muk), rnorm(n, mean = muk), rnorm(n, mean = -muk), rnorm(n, mean = muk))
x4 <- c(rnorm(n, mean = -muk), rnorm(n, mean = muk), rnorm(n, mean = -muk), rnorm(n, mean = muk))

x <- data.frame(x1,x2,x3,x4)

# apply k-prototyps
kpres <- kproto(x, 4)
clprofiles(kpres, x)

# in real world clusters are often not as clear cut
# by variation of lambda the emphasize is shifted towards factor / numeric variables    
kpres <- kproto(x, 2)
clprofiles(kpres, x)

kpres <- kproto(x, 2, lambda = 0.1)
clprofiles(kpres, x)

kpres <- kproto(x, 2, lambda = 25)
clprofiles(kpres, x)

}
\author{
\email{gero.szepannek@web.de}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/kproto2silhouette.R
\name{kproto2silhouette}
\alias{kproto2silhouette}
\title{Compute individual Silhouette Widths}
\usage{
kproto2silhouette(object)
}
\arguments{
\item{object}{Object of class \code{kproto}.}
}
\value{
Object of class \code{silhouette} to be further used with \code{\link[cluster]{summary.silhouette}} 
or \code{\link[cluster]{plot.silhouette}}. It consists in a matrix where each row corresponds to one observation 
and the columns contain the assigned cluster as well as the neighbor cluster as well as the individual silhouette 
width ob the observation.
}
\description{
Computes an object of class \code{silhouette} from k-prototypes clustering for mixed-type data.
}
\details{
Information on the individual silhouette width for all observations is computed. 
The k-prototypes result of class \code{kproto} is turned into an object of class \code{silhouette} 
that can be further used with the functions \code{\link[cluster]{summary.silhouette}} or \code{\link[cluster]{plot.silhouette}}.
}
\examples{
# generate toy data with factors and numerics

n   <- 50
prb <- 0.9
muk <- 1.5 
clusid <- rep(1:4, each = n)

x1 <- sample(c("A","B"), 2*n, replace = TRUE, prob = c(prb, 1-prb))
x1 <- c(x1, sample(c("A","B"), 2*n, replace = TRUE, prob = c(1-prb, prb)))
x1 <- as.factor(x1)

x2 <- sample(c("A","B"), 2*n, replace = TRUE, prob = c(prb, 1-prb))
x2 <- c(x2, sample(c("A","B"), 2*n, replace = TRUE, prob = c(1-prb, prb)))
x2 <- as.factor(x2)

x3 <- c(rnorm(n, mean = -muk), rnorm(n, mean = muk), rnorm(n, mean = -muk), rnorm(n, mean = muk))
x4 <- c(rnorm(n, mean = -muk), rnorm(n, mean = muk), rnorm(n, mean = -muk), rnorm(n, mean = muk))

x <- data.frame(x1,x2,x3,x4)

# apply k-prototypes
kpres <- kproto(x, 4)

# compute silhouette widhts
# extract object of class sil
sils <- kproto2silhouette(kpres)

# apply functions from package cluster
library(cluster)
summary(sils)
plot(sils)

# compute average silhouette width  
mean(sils[,3])

# ...alternatively, compute average silhouette width directly using validation_kproto() 
cindex_value <- validation_kproto(method = "silhouette", object = kpres)
cindex_value

}
\references{
\itemize{
    \item Szepannek, G. (2018): 
    clustMixType: User-Friendly Clustering of Mixed-Type Data in R, 
    {\emph{The R Journal 10/2}}, 200-208. 
    \doi{10.32614/RJ-2018-048}.
    
    \item Aschenbruck, R., Szepannek, G. (2020): 
    Cluster Validation for Mixed-Type Data. 
    \emph{Archives of Data Science, Series A, Vol 6, Issue 1}.
    \doi{10.5445/KSP/1000098011/02}.
    
    \item Kaufman, L., Rousseeuw, P. (1990): 
    \emph{Finding Groups in Data: An Introduction to Cluster Analysis.} 
     Wiley.
  }
}
\author{
\email{gero.szepannek@web.de}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/kprototypes.R
\name{summary.kproto}
\alias{summary.kproto}
\title{Summary Method for kproto Cluster Result}
\usage{
\method{summary}{kproto}(object, data = NULL, pct.dig = 3, ...)
}
\arguments{
\item{object}{Object of class \code{kproto}.}

\item{data}{Optional data set to be analyzed. If \code{!(is.null(data))} clusters for \code{data} are assigned by 
\code{predict(object, data)}. If not specified the clusters of the original data ara analyzed which is only possible if \code{kproto} 
has been called using \code{keep.data = TRUE}.}

\item{pct.dig}{Number of digits for rounding percentages of factor variables.}

\item{\dots}{Further arguments to be passed to internal call of \code{summary()} for numeric variables.}
}
\value{
List where each element corresponds to one variable. Each row of any element corresponds to one cluster.
}
\description{
Investigation of variances to specify lambda for k-prototypes clustering.
}
\details{
For numeric variables statistics are computed for each clusters using \code{summary()}. 
For categorical variables distribution percentages are computed.
}
\examples{
# generate toy data with factors and numerics

n   <- 100
prb <- 0.9
muk <- 1.5 
clusid <- rep(1:4, each = n)

x1 <- sample(c("A","B"), 2*n, replace = TRUE, prob = c(prb, 1-prb))
x1 <- c(x1, sample(c("A","B"), 2*n, replace = TRUE, prob = c(1-prb, prb)))
x1 <- as.factor(x1)

x2 <- sample(c("A","B"), 2*n, replace = TRUE, prob = c(prb, 1-prb))
x2 <- c(x2, sample(c("A","B"), 2*n, replace = TRUE, prob = c(1-prb, prb)))
x2 <- as.factor(x2)

x3 <- c(rnorm(n, mean = -muk), rnorm(n, mean = muk), rnorm(n, mean = -muk), rnorm(n, mean = muk))
x4 <- c(rnorm(n, mean = -muk), rnorm(n, mean = muk), rnorm(n, mean = -muk), rnorm(n, mean = muk))

x <- data.frame(x1,x2,x3,x4)

res <- kproto(x, 4)
summary(res)

}
\author{
\email{gero.szepannek@web.de}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/clustervalidation.R
\name{validation_kproto}
\alias{validation_kproto}
\title{Validating k Prototypes Clustering}
\usage{
validation_kproto(
  method = "silhouette",
  object = NULL,
  data = NULL,
  type = "huang",
  k = NULL,
  lambda = NULL,
  kp_obj = "optimal",
  verbose = FALSE,
  ...
)
}
\arguments{
\item{method}{Character specifying the validation index: \code{cindex}, \code{dunn}, \code{gamma}, \code{gplus}, \code{mcclain}, \code{ptbiserial}, \code{silhouette} (default) or \code{tau}.}

\item{object}{Object of class \code{kproto} resulting from a call with \code{kproto(..., keep.data=TRUE)}.}

\item{data}{Original data; only required if \code{object == NULL} and neglected if \code{object != NULL}.}

\item{type}{Character, to specify the distance for clustering; either \code{"huang"} or \code{"gower"}.}

\item{k}{Vector specifying the search range for optimum number of clusters; if \code{NULL} the range will set as \code{2:sqrt(n)}. Only required if \code{object == NULL} and neglected if \code{object != NULL}.}

\item{lambda}{Factor to trade off between Euclidean distance of numeric variables and simple matching coefficient between categorical variables.}

\item{kp_obj}{character either "optimal" or "all": Output of the index-optimal clustering (kp_obj == "optimal") or all computed cluster partitions (kp_obj == "all"); only required if \code{object != NULL}.}

\item{verbose}{Logical, whether additional information about process should be printed.}

\item{...}{Further arguments passed to \code{\link[clustMixType]{kproto}}, like:
\itemize{
  \item \code{nstart}: If > 1 repetitive computations of \code{kproto} with random initializations are computed.
  \item \code{na.rm}: Character, either \code{"yes"} to strip \code{NA} values for complete case analysis, \code{"no"} to keep and ignore \code{NA} values, \code{"imp.internal"} to impute the \code{NAs} within the algorithm or \code{"imp.onestep"} to apply the algorithm ignoring the \code{NAs} and impute them after the partition is determined.
}}
}
\value{
For computing the optimal number of clusters based on the choosen validation index for k-Prototype clustering the output contains:

\item{k_opt}{optimal number of clusters (sampled in case of ambiguity)}

\item{index_opt}{index value of the index optimal clustering}

\item{indices}{calculated indices for \eqn{k=2,...,k_{max}}}

\item{kp_obj}{if(kp_obj == "optimal") the kproto object of the index optimal clustering and if(kp_obj == "all") all kproto which were calculated}

For computing the index-value for a given k-Prototype clustering the output contains:

\item{index}{calculated index-value}
}
\description{
Calculating the preferred validation index for a k-Prototypes clustering with k clusters or computing the optimal number of clusters based on the choosen index for k-Prototype clustering. Possible validation indices are: \code{cindex}, \code{dunn}, \code{gamma}, \code{gplus}, \code{mcclain}, \code{ptbiserial}, \code{silhouette} and \code{tau}.
}
\details{
More information about the implemented validation indices:
  \itemize{
    \item \code{cindex} \deqn{Cindex = \frac{S_w-S_{min}}{S_{max}-S_{min}}} \cr
For \eqn{S_{min}} and \eqn{S_{max}} it is necessary to calculate the distances between all pairs of points in the entire data set (\eqn{\frac{n(n-1)}{2}}). 
\eqn{S_{min}} is the sum of the "total number of pairs of objects belonging to the same cluster" smallest distances and 
\eqn{S_{max}} is the sum of the "total number of pairs of objects belonging to the same cluster" largest distances. \eqn{S_w} is the sum of the within-cluster distances. \cr
The minimum value of the index is used to indicate the optimal number of clusters.

    \item \code{dunn} \deqn{Dunn = \frac{\min_{1 \leq i < j \leq q} d(C_i, C_j)}{\max_{1 \leq k \leq q} diam(C_k)}} \cr
The following applies: The dissimilarity between the two clusters \eqn{C_i} and \eqn{C_j} is defined as \eqn{d(C_i, C_j)=\min_{x \in C_i, y \in C_j} d(x,y)} and
the diameter of a cluster is defined as \eqn{diam(C_k)=\max_{x,y \in C} d(x,y)}. \cr
The maximum value of the index is used to indicate the optimal number of clusters.

    \item \code{gamma} \deqn{Gamma = \frac{s(+)-s(-)}{s(+)+s(-)}} \cr 
Comparisons are made between all within-cluster dissimilarities and all between-cluster dissimilarities. 
\eqn{s(+)} is the number of concordant comparisons and \eqn{s(-)} is the number of discordant comparisons.
A comparison is named concordant (resp. discordant) if a within-cluster dissimilarity is strictly less (resp. strictly greater) than a between-cluster dissimilarity.\cr
The maximum value of the index is used to indicate the optimal number of clusters.

    \item \code{gplus} \deqn{Gplus = \frac{2 \cdot s(-)}{\frac{n(n-1)}{2} \cdot (\frac{n(n-1)}{2}-1)}} \cr 
Comparisons are made between all within-cluster dissimilarities and all between-cluster dissimilarities. 
\eqn{s(-)} is the number of discordant comparisons and a comparison is named discordant if a within-cluster 
dissimilarity is strictly greater than a between-cluster dissimilarity. \cr
The minimum value of the index is used to indicate the optimal number of clusters.

    \item \code{mcclain} \deqn{McClain = \frac{\bar{S}_w}{\bar{S}_b}} \cr 
\eqn{\bar{S}_w} is the sum of within-cluster distances divided by the number of within-cluster distances and 
\eqn{\bar{S}_b} is the sum of between-cluster distances divided by the number of between-cluster distances.\cr
The minimum value of the index is used to indicate the optimal number of clusters.

    \item\code{ptbiserial} \deqn{Ptbiserial = \frac{(\bar{S}_b-\bar{S}_w) \cdot (\frac{N_w \cdot N_b}{N_t^2})^{0.5}}{s_d}} \cr 
\eqn{\bar{S}_w} is the sum of within-cluster distances divided by the number of within-cluster distances and 
\eqn{\bar{S}_b} is the sum of between-cluster distances divided by the number of between-cluster distances.\cr
\eqn{N_t} is the total number of pairs of objects in the data, \eqn{N_w} is the total number of pairs of 
objects belonging to the same cluster and \eqn{N_b} is the total number of pairs of objects belonging to different clusters.
\eqn{s_d} is the standard deviation of all distances.\cr
The maximum value of the index is used to indicate the optimal number of clusters.

    \item \code{silhouette} \deqn{Silhouette = \frac{1}{n} \sum_{i=1}^n \frac{b(i)-a(i)}{max(a(i),b(i))}} \cr 
\eqn{a(i)} is the average dissimilarity of the i\emph{th} object to all other objects of the same/own cluster.
\eqn{b(i)=min(d(i,C))}, where \eqn{d(i,C)} is the average dissimilarity of the i\emph{th} object to all the other clusters except the own/same cluster.\cr
The maximum value of the index is used to indicate the optimal number of clusters.
    
    \item \code{tau} \deqn{Tau = \frac{s(+) - s(-)}{((\frac{N_t(N_t-1)}{2}-t)\frac{N_t(N_t-1)}{2})^{0.5}}} \cr 
Comparisons are made between all within-cluster dissimilarities and all between-cluster dissimilarities. 
\eqn{s(+)} is the number of concordant comparisons and \eqn{s(-)} is the number of discordant comparisons.
A comparison is named concordant (resp. discordant) if a within-cluster dissimilarity is strictly less 
(resp. strictly greater) than a between-cluster dissimilarity.\cr
\eqn{N_t} is the total number of distances \eqn{\frac{n(n-1)}{2}} and \eqn{t} is the number of comparisons 
of two pairs of objects where both pairs represent within-cluster comparisons or both pairs are between-cluster
comparisons. \cr
The maximum value of the index is used to indicate the optimal number of clusters.
   
  }
}
\examples{
\dontrun{
# generate toy data with factors and numerics
n   <- 10
prb <- 0.99
muk <- 2.5 

x1 <- sample(c("A","B"), 2*n, replace = TRUE, prob = c(prb, 1-prb))
x1 <- c(x1, sample(c("A","B"), 2*n, replace = TRUE, prob = c(1-prb, prb)))
x1 <- as.factor(x1)
x2 <- sample(c("A","B"), 2*n, replace = TRUE, prob = c(prb, 1-prb))
x2 <- c(x2, sample(c("A","B"), 2*n, replace = TRUE, prob = c(1-prb, prb)))
x2 <- as.factor(x2)
x3 <- c(rnorm(n, mean = -muk), rnorm(n, mean = muk), rnorm(n, mean = -muk), rnorm(n, mean = muk))
x4 <- c(rnorm(n, mean = -muk), rnorm(n, mean = muk), rnorm(n, mean = -muk), rnorm(n, mean = muk))
x <- data.frame(x1,x2,x3,x4)


# calculate optimal number of cluster, index values and clusterpartition with Silhouette-index
val <- validation_kproto(method = "silhouette", data = x, k = 3:5, nstart = 5)


# apply k-prototypes
kpres <- kproto(x, 4, keep.data = TRUE)

# calculate cindex-value for the given clusterpartition
cindex_value <- validation_kproto(method = "cindex", object = kpres)
}

}
\references{
\itemize{
    \item Aschenbruck, R., Szepannek, G. (2020): 
    Cluster Validation for Mixed-Type Data. 
    \emph{Archives of Data Science, Series A, Vol 6, Issue 1}.
    \doi{10.5445/KSP/1000098011/02}.
    
    \item Charrad, M., Ghazzali, N., Boiteau, V., Niknafs, A. (2014): 
    NbClust: An R Package for Determining the Relevant Number of Clusters in a Data Set. 
    \emph{Journal of Statistical Software, Vol 61, Issue 6}.
    \doi{10.18637/jss.v061.i06}.
  }
}
\author{
Rabea Aschenbruck
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/kprototypes.R
\name{kproto}
\alias{kproto}
\alias{kproto.default}
\title{k-Prototypes Clustering}
\usage{
kproto(x, ...)

\method{kproto}{default}(
  x,
  k,
  lambda = NULL,
  type = "huang",
  iter.max = 100,
  nstart = 1,
  na.rm = "yes",
  keep.data = TRUE,
  verbose = TRUE,
  init = NULL,
  p_nstart.m = 0.9,
  ...
)
}
\arguments{
\item{x}{Data frame with both numerics and factors (also ordered factors are possible).}

\item{\dots}{Currently not used.}

\item{k}{Either the number of clusters, a vector specifying indices of initial prototypes, or a data frame of 
prototypes of the same columns as \code{x}.}

\item{lambda}{Parameter > 0 to trade off between Euclidean distance of numeric variables and simple matching 
coefficient between categorical variables (if \code{type = "huang"}). Also a vector of variable specific factors 
is possible where the order must correspond to the order of the variables in the data. In this case all variables' 
distances will be multiplied by their corresponding lambda value.}

\item{type}{Character, to specify the distance for clustering. Either \code{"huang"} or \code{"gower"} (cf. details 
below).}

\item{iter.max}{Numeric; maximum number of iterations if no convergence before.}

\item{nstart}{Numeric; If > 1 repetitive computations with random initializations are computed and the result with 
minimum \code{tot.dist} is returned.}

\item{na.rm}{Character, either \code{"yes"} to strip \code{NA} values for complete case analysis, \code{"no"} to 
keep and ignore \code{NA} values, \code{"imp.internal"} to impute the \code{NAs} within the algorithm or 
\code{"imp.onestep"} to apply the algorithm ignoring the \code{NAs} and impute them after the partition is determined.}

\item{keep.data}{Logical, whether original should be included in the returned object.}

\item{verbose}{Logical, whether additional information about process should be printed. 
Caution: For \code{verbose=FALSE}, if the number of clusters is reduced during the iterations it will not mentioned.}

\item{init}{Character, to specify the initialization strategy. Either \code{"nbh.dens"}, \code{"sel.cen"} or 
\code{"nstart.m"}. Default is \code{"NULL"}, which results in nstart repetitive algorithm computations with random 
starting prototypes. Otherwise, \code{nstart} is not used. Argument \code{k} must be a number if a specific 
initialization strategy is choosen!}

\item{p_nstart.m}{Numeric, probability(=0.9 is default) for \code{init="nstart.m"}, where the strategy assures 
that with a probability of \code{p_nstart.m} at least one of the m sets of initial prototypes contains objects 
of every cluster group (cf. Aschenbruck et al. (2023): Random-based Initialization for clustering mixed-type data 
with the k-Prototypes algorithm. In: {\emph{Cladag 2023 Book of abstracts and short spapers}}, isbn: 9788891935632.).}
}
\value{
\code{\link{kmeans}} like object of class \code{kproto}:

\item{cluster}{Vector of cluster memberships.}

\item{centers}{Data frame of cluster prototypes.}

\item{lambda}{Distance parameter lambda.}

\item{size}{Vector of cluster sizes.}

\item{withinss}{Vector of within cluster distances for each cluster, i.e. summed distances of all observations belonging to a cluster to their respective prototype.}

\item{tot.withinss}{Target function: sum of all observations' distances to their corresponding cluster prototype.}

\item{dists}{Matrix with distances of observations to all cluster prototypes.}

\item{iter}{Prespecified maximum number of iterations.}

\item{trace}{List with two elements (vectors) tracing the iteration process: 
\code{tot.dists} and \code{moved} number of observations over all iterations.}

\item{inits}{Initial prototypes determined by specified initialization strategy, if init is either 'nbh.dens' or 'sel.cen'.}

\item{nstart.m}{only for 'init = nstart_m': determined number of randomly choosen sets.}

\item{data}{if 'keep.data = TRUE' than the original data will be added to the output list.}

\item{type}{Type argument of the function call.}

\item{stdization}{Only returned for \code{type = "gower"}: List of standardized ranks for ordinal variables 
and an additional element \code{num_ranges} with ranges of all numeric variables. Used by \code{\link{predict.kproto}}.}
}
\description{
Computes k-prototypes clustering for mixed-type data.
}
\details{
Like k-means, the k-prototypes algorithm iteratively recomputes cluster prototypes and reassigns 
clusters, whereby with \code{type = "huang"} clusters are assigned using the distance
\eqn{d(x,y) =  d_{euclid}(x,y) + \lambda d_{simple\,matching}(x,y)}. Cluster prototypes are computed as 
cluster means for numeric variables and modes for factors (cf. Huang, 1998). Ordered factors variables 
are treated as categorical variables.\cr
For \code{type = "gower"} range-normalized absolute distances from the cluster median are computed for 
the numeric variables (and for the ranks of the ordered factors respectively). For factors simple matching 
distance is used as in the original k prototypes algorithm. The prototypes are given by the median for 
numeric variables, the mode for factors and the level with the closest rank to the median rank of the 
corresponding cluster (cf. Szepannek et al., 2024).\cr
In case of \code{na.rm = FALSE}: for each observation variables with missings are ignored (i.e. only the 
remaining variables are considered for distance computation). In consequence for observations with missings 
this might result in a change of variable's weighting compared to the one specified by \code{lambda}. For 
these observations distances to the prototypes will typically be smaller as they are based on fewer variables.\cr
The \code{type} argument also accepts input \code{"standard"}, but this naming convention is deprecated and 
has been renamed to \code{"huang"}. Please use \code{"huang"} instead.
}
\examples{
# generate toy data with factors and numerics

n   <- 100
prb <- 0.9
muk <- 1.5 
clusid <- rep(1:4, each = n)

x1 <- sample(c("A","B"), 2*n, replace = TRUE, prob = c(prb, 1-prb))
x1 <- c(x1, sample(c("A","B"), 2*n, replace = TRUE, prob = c(1-prb, prb)))
x1 <- as.factor(x1)

x2 <- sample(c("A","B"), 2*n, replace = TRUE, prob = c(prb, 1-prb))
x2 <- c(x2, sample(c("A","B"), 2*n, replace = TRUE, prob = c(1-prb, prb)))
x2 <- as.factor(x2)

x3 <- c(rnorm(n, mean = -muk), rnorm(n, mean = muk), rnorm(n, mean = -muk), rnorm(n, mean = muk))
x4 <- c(rnorm(n, mean = -muk), rnorm(n, mean = muk), rnorm(n, mean = -muk), rnorm(n, mean = muk))

x <- data.frame(x1,x2,x3,x4)

# apply k-prototypes
kpres <- kproto(x, 4)
clprofiles(kpres, x)

# in real world clusters are often not as clear cut
# by variation of lambda the emphasize is shifted towards factor / numeric variables    
kpres <- kproto(x, 2)
clprofiles(kpres, x)

kpres <- kproto(x, 2, lambda = 0.1)
clprofiles(kpres, x)

kpres <- kproto(x, 2, lambda = 25)
clprofiles(kpres, x)

}
\references{
\itemize{
    \item Szepannek, G. (2018): clustMixType: User-Friendly Clustering of Mixed-Type Data in R, 
    {\emph{The R Journal 10/2}}, 200-208, \doi{10.32614/RJ-2018-048}.
    \item Aschenbruck, R., Szepannek, G., Wilhelm, A. (2022): Imputation Strategies for Clustering Mixed‑Type Data with Missing Values, 
    {\emph{Journal of Classification}}, \doi{10.1007/s00357-022-09422-y}. 
    \item Szepannek, G., Aschenbruck, R., Wilhelm, A. (2024): Clustering Large Mixed-Type Data with Ordinal Variables,
    {\emph{Advances in Data Analysis and Classification}}, \doi{10.1007/s11634-024-00595-5}.
    \item Z.Huang (1998): Extensions to the k-Means Algorithm for Clustering Large Data Sets with Categorical Variables, 
          {\emph{Data Mining and Knowledge Discovery 2}}, 283-304.
  }
}
\author{
\email{gero.szepannek@web.de}
}
\keyword{classif}
\keyword{cluster}
\keyword{multivariate}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/kprototypes.R
\name{plot.kproto}
\alias{plot.kproto}
\title{Assign k-Prototypes Clusters}
\usage{
\method{plot}{kproto}(x, ...)
}
\arguments{
\item{x}{Object resulting from a call of \code{kproto}.}

\item{\dots}{Additional arguments to be passet to \code{\link{clprofiles}} such as e.g. \code{vars}.}
}
\description{
Plot distributions of the clusters across the variables.
}
\details{
Wrapper around \code{\link{clprofiles}}. Only works for \code{kproto} object created with \code{keep.data = TRUE}.
}
\examples{
# generate toy data with factors and numerics

n   <- 100
prb <- 0.9
muk <- 1.5 
clusid <- rep(1:4, each = n)

x1 <- sample(c("A","B"), 2*n, replace = TRUE, prob = c(prb, 1-prb))
x1 <- c(x1, sample(c("A","B"), 2*n, replace = TRUE, prob = c(1-prb, prb)))
x1 <- as.factor(x1)

x2 <- sample(c("A","B"), 2*n, replace = TRUE, prob = c(prb, 1-prb))
x2 <- c(x2, sample(c("A","B"), 2*n, replace = TRUE, prob = c(1-prb, prb)))
x2 <- as.factor(x2)

x3 <- c(rnorm(n, mean = -muk), rnorm(n, mean = muk), rnorm(n, mean = -muk), rnorm(n, mean = muk))
x4 <- c(rnorm(n, mean = -muk), rnorm(n, mean = muk), rnorm(n, mean = -muk), rnorm(n, mean = muk))

x <- data.frame(x1,x2,x3,x4)

# apply k-prototyps
kpres <- kproto(x, 4)
plot(kpres, vars = c("x1","x3")) 


}
\author{
\email{gero.szepannek@web.de}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/kprototypes.R
\name{lambdaest}
\alias{lambdaest}
\title{Compares Variability of Variables}
\usage{
lambdaest(
  x,
  num.method = 1,
  fac.method = 1,
  outtype = "numeric",
  verbose = TRUE
)
}
\arguments{
\item{x}{Data.frame with both numerics and factors.}

\item{num.method}{Integer 1 or 2. Specifies the heuristic used for numeric variables.}

\item{fac.method}{Integer 1 or 2. Specifies the heuristic used for factor variables.}

\item{outtype}{Specifies the desired output: either 'numeric', 'vector' or 'variation'.}

\item{verbose}{Logical whether additional information about process should be printed.}
}
\value{
\item{lambda}{Ratio of averages over all numeric/factor variables is returned. 
In case of \code{outtype = "vector"} the separate lambda for all variables is returned as the inverse of the single variables' 
variation as specified by the \code{num.method} and \code{fac.method} argument. \code{outtype = "variation"} directly returns these quantities and is not meant to be 
passed directly to \code{kproto()}.}
}
\description{
Investigation of the variables' variances/concentrations to support specification of lambda for k-prototypes clustering.
}
\details{
Variance (\code{num.method = 1}) or standard deviation (\code{num.method = 2}) of numeric variables 
and \eqn{1-\sum_i p_i^2} (\code{fac.method = 1}) or \eqn{1-\max_i p_i} (\code{fac.method = 2}) for factors is computed.
}
\examples{
# generate toy data with factors and numerics

n   <- 100
prb <- 0.9
muk <- 1.5 
clusid <- rep(1:4, each = n)

x1 <- sample(c("A","B"), 2*n, replace = TRUE, prob = c(prb, 1-prb))
x1 <- c(x1, sample(c("A","B"), 2*n, replace = TRUE, prob = c(1-prb, prb)))
x1 <- as.factor(x1)

x2 <- sample(c("A","B"), 2*n, replace = TRUE, prob = c(prb, 1-prb))
x2 <- c(x2, sample(c("A","B"), 2*n, replace = TRUE, prob = c(1-prb, prb)))
x2 <- as.factor(x2)

x3 <- c(rnorm(n, mean = -muk), rnorm(n, mean = muk), rnorm(n, mean = -muk), rnorm(n, mean = muk))
x4 <- c(rnorm(n, mean = -muk), rnorm(n, mean = muk), rnorm(n, mean = -muk), rnorm(n, mean = muk))

x <- data.frame(x1,x2,x3,x4)

lambdaest(x)
res <- kproto(x, 4, lambda = lambdaest(x))

}
\author{
\email{gero.szepannek@web.de}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/kprototypes.R
\name{predict.kproto}
\alias{predict.kproto}
\title{Assign k-Prototypes Clusters}
\usage{
\method{predict}{kproto}(object, newdata, ...)
}
\arguments{
\item{object}{Object resulting from a call of \code{kproto}.}

\item{newdata}{New data frame (of same structure) where cluster memberships are to be predicted.}

\item{\dots}{Currently not used.}
}
\value{
\code{\link{kmeans}} like object of class \code{kproto}:

\item{cluster}{Vector of cluster memberships.}

\item{dists}{Matrix with distances of observations to all cluster prototypes.}
}
\description{
Predicts k-prototypes cluster memberships and distances for new data.
}
\examples{
# generate toy data with factors and numerics

n   <- 100
prb <- 0.9
muk <- 1.5 
clusid <- rep(1:4, each = n)

x1 <- sample(c("A","B"), 2*n, replace = TRUE, prob = c(prb, 1-prb))
x1 <- c(x1, sample(c("A","B"), 2*n, replace = TRUE, prob = c(1-prb, prb)))
x1 <- as.factor(x1)

x2 <- sample(c("A","B"), 2*n, replace = TRUE, prob = c(prb, 1-prb))
x2 <- c(x2, sample(c("A","B"), 2*n, replace = TRUE, prob = c(1-prb, prb)))
x2 <- as.factor(x2)

x3 <- c(rnorm(n, mean = -muk), rnorm(n, mean = muk), rnorm(n, mean = -muk), rnorm(n, mean = muk))
x4 <- c(rnorm(n, mean = -muk), rnorm(n, mean = muk), rnorm(n, mean = -muk), rnorm(n, mean = muk))

x <- data.frame(x1,x2,x3,x4)

# apply k-prototyps
kpres <- kproto(x, 4)
predicted.clusters <- predict(kpres, x) 


}
\author{
\email{gero.szepannek@web.de}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/stability_kproto.R
\name{stability_kproto}
\alias{stability_kproto}
\title{Determination the stability of k Prototypes Clustering}
\usage{
stability_kproto(
  object,
  method = c("rand", "jaccard", "luxburg", "fowlkesmallows"),
  B = 100,
  verbose = FALSE,
  ...
)
}
\arguments{
\item{object}{Object of class \code{kproto} resulting from a call with \code{kproto(..., keep.data=TRUE)}}

\item{method}{character specifying the stability, either one or more of \code{luxburg}, \code{fowlkesmallows}, \code{rand} or/and \code{jaccard}.}

\item{B}{numeric, number of bootstrap samples}

\item{verbose}{Logical whether information about the bootstrap procedure should be given.}

\item{...}{Further arguments passed to \code{\link[clustMixType]{kproto}}, like:
\itemize{
  \item \code{nstart}: If > 1 repetitive computations of \code{kproto} with random initial prototypes are computed.
  \item \code{lambda}: Factor to trade off between Euclidean distance of numeric variables and simple matching coefficient between categorical variables.
}}
}
\value{
The output contains the stability for a given k-Prototype clustering in a list with two elements:

\item{kp_stab}{stability values for the given clustering}

\item{kp_bts_stab}{stability values for each bootstrap samples}
}
\description{
Calculating the stability for a k-Prototypes clustering with k clusters or computing the stability-based optimal number of clusters for k-Prototype clustering. Possible stability indices are: \code{Jaccard}, \code{Rand}, \code{Fowlkes \& Mallows} and \code{Luxburg}.
}
\examples{
\dontrun{
# generate toy data with factors and numerics
n   <- 10
prb <- 0.99
muk <- 2.5 

x1 <- sample(c("A","B"), 2*n, replace = TRUE, prob = c(prb, 1-prb))
x1 <- c(x1, sample(c("A","B"), 2*n, replace = TRUE, prob = c(1-prb, prb)))
x1 <- as.factor(x1)
x2 <- sample(c("A","B"), 2*n, replace = TRUE, prob = c(prb, 1-prb))
x2 <- c(x2, sample(c("A","B"), 2*n, replace = TRUE, prob = c(1-prb, prb)))
x2 <- as.factor(x2)
x3 <- c(rnorm(n, mean = -muk), rnorm(n, mean = muk), rnorm(n, mean = -muk), rnorm(n, mean = muk))
x4 <- c(rnorm(n, mean = -muk), rnorm(n, mean = muk), rnorm(n, mean = -muk), rnorm(n, mean = muk))
x <- data.frame(x1,x2,x3,x4)

#' # apply k-prototypes
kpres <- kproto(x, 4, keep.data = TRUE)

# calculate cluster stability
stab <- stability_kproto(method = c("luxburg","fowlkesmallows"), object = kpres)

}

}
\references{
\itemize{
    \item Aschenbruck, R., Szepannek, G., Wilhelm, A.F.X (2023): 
    Stability of mixed-type cluster partitions for determination of the number of clusters. 
    \emph{Submitted}.
    
    \item von Luxburg, U. (2010): 
     Clustering stability: an overview. 
    \emph{Foundations and Trends in Machine Learning, Vol 2, Issue 3}.
    \doi{10.1561/2200000008}.
    
    \item Ben-Hur, A., Elisseeff, A., Guyon, I. (2002): 
    A stability based method for discovering structure in clustered data. 
    \emph{Pacific Symposium on Biocomputing}.
    \doi{10/bhfxmf}.
  }
}
\author{
Rabea Aschenbruck
}
